/*
 * frameTrans.c
 *
 *  Created: Feb 16, 2015 10:30:37 AM
 *  Author: Antoine Pignede
 *
 *  This file defines the Frame Transformations
 */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "constants.h"
#include "timeConv.h"
#include "frameTrans.h"

/*
 *Frame Transformations
 * function frameEcef2geod
 *
 * Created: 30.01.2015 15:51:02
 * Author: Antoine Pignede
 *
 * Convert Earth-centered, Earth fixed (ECEF) coordinates X, Y, and Z to
 *   geodetic coordinates LATITUDE, LONGITUDE, and ALTITUDE. The World Geodetic
 *   System 1984 (WGS84) ellipsoid model of the Earth is assumed.
 *
 * Copied and edited from Compston (2011) Matlab File Exchange International Geomagnetic Reference Field (IGRF) Model
 *
 * Inputs:
 *   -recef: x,y,z coordinates of the point in kilometers.
 *
 * Ouputs:
 *   -rgeod: Geodetic latitude and longitude in degrees, height above the Earth in kilometers.
 *
 * Coupling:
 *   -frameLatitudeRecur: Compute latitude recursively
 *
 * See also
 *   frameGeod2ecef.m  - inverse function
 */

void frameEcef2geod(double recef[3], double rgeod[3]){
	// Split input vector in components
	double x = recef[0];
	double y = recef[1];
	double z = recef[2];

	// Longitude is easy
	double longitude = atan2(y, x)*RAD2DEG;

	// Compute latitude recursively
	double latitude;
	double Nphi;
	double rd = hypot(x, y);
	double lat_in = asin(z / hypot(x, hypot(y, z)));

	frameLatitudeRecur(lat_in, z, rd, 1, &latitude, &Nphi);

	double sinlat = sin(latitude);
	double coslat = cos(latitude);
	latitude = latitude*RAD2DEG;

	// Get altitude from latitude.
	double altitude = rd*coslat + (z + ECCSQRD*Nphi*sinlat)*sinlat - Nphi;

	// Build output vector
	rgeod[0] = latitude;
	rgeod[1] = longitude;
	rgeod[2] = altitude;
}

/*
 *Frame Transformations
 * function frameGeod2ecef
 *
 * Created: 30.01.2015 15:51:02
 * Author: Antoine Pignede
 *
 * Converts geodetic coordinates LATITUDE, LONGITUDE, and ALTITUDE to
 *   Earth-centered, Earth fixed (ECEF) coordinates X, Y, and Z. The World Geodetic
 *   System 1984 (WGS84) ellipsoid model of the Earth is assumed.
 *
 * Copied and edited from Compston (2011) Matlab File Exchange International Geomagnetic Reference Field (IGRF) Model
 *
 * Inputs:
 *   -rgeod: Geodetic latitude and longitude in degrees, height above the Earth in kilometers.
 *
 * Ouputs:
 *   -recef: x,y,z coordinates of the point in kilometers.
 *
 * See also
 *   frameEcef2geod.m  - inverse function
 */

void frameGeod2ecef(double rgeod[3], double recef[3]){
	// Split input vector in components
	double latitude = rgeod[0];
	double longitude = rgeod[1];
	double altitude = rgeod[2];

	latitude = latitude* DEG2RAD;
	longitude = longitude* DEG2RAD;

	// Conversion from: en.wikipedia.org/wiki/Geodetic_system#Conversion_calculations
	double sinlat = sin(latitude);
	double coslat = cos(latitude);
	double sinlong = sin(longitude);
	double coslong = cos(longitude);

	double Nphi = REQU / sqrt(1 - ECCSQRD*sinlat*sinlat);

	recef[0] = (Nphi + altitude)*coslat*coslong;
	recef[1] = (Nphi + altitude)*coslat*sinlong;
	recef[2] = (Nphi*(1 - ECCSQRD) + altitude)*sinlat;
}

/*
 *Frame Transformations
 * function ecef2eci
 *
 * Created: 30.01.2015 16:39:02
 * Author: Antoine Pignede
 *
 * This function transforms a vector from the earth fixed (itrf) frame, to
 *   the eci mean equator mean equinox (j2000) frame.  the results take into account
 *   the effects of precession, nutation, sidereal time, and polar motion.
 *
 * Copied and edited from Vallado (2013) Fundamentals of Astrodynamics and Applications
 *
 *  inputs          description                    range / units
 *    recef       - position vector earth fixed    km
 *    timeUTC     - struct containing:
 *      year        - year                           1900 .. 2100
 *      mon         - month                          1 .. 12
 *      day         - day                            1 .. 28,29,30,31
 *      hr          - universal time hour            0 .. 23
 *      min         - universal time min             0 .. 59
 *      sec         - universal time sec             0.0 .. 59.999
 *
 *  outputs       :
 *    reci        - position vector eci            km
 *
 *  locals        :
 *    deltapsi    - nutation angle                 rad
 *    meaneps     - mean obliquity of the ecliptic rad
 *    omega       -                                rad
 *    prec        - matrix for mod - eci
 *    nut         - matrix for tod - mod
 *    st          - matrix for pef - tod
 *    pm          - matrix for ecef - pef
 *    jdUTC       - julian date                    days from 4713 bc
 *    all UTC times are converted to terrestrial time (TT)
 *
 *  coupling      :
 *   framePrecess      - rotation for precession
 *   frameNutation     - rotation for nutation
 *   frameSidereal     - rotation for sidereal time
 *   framePolarm       - rotation for polar motion
 *   timeDatetime2jd   - convert date and time values to julian date
 *   timeUtc2tt        - convert UTC to TT
 *   timeJd2jc         - convert julian date to julian centuries
 *
 * See also
 *   frameEci2ecef.m  - inverse function
 */

void frameEcef2eci(double recef[3],Time timeUTC, double reci[3]){
	// find the julian centuries of terrestrial time at given UTC
	double jdUTC = timeDatetime2jd(timeUTC);
	Time timeTT = timeUtc2tt(timeUTC);
	double jdTT = timeDatetime2jd(timeTT);
	double ttt = timeJd2jc(jdTT);

	// get the transformation matrices
	double prec[9];
	framePrecess(ttt, prec);

	double deltapsi;
	double meaneps;
	double omega;
	double nut[9];
	frameNutation(ttt, &deltapsi, &meaneps, &omega, nut);

	double st[9];
	frameSidereal(jdUTC, deltapsi, meaneps, omega, st);

	double pm[9];
	framePolarm(jdUTC, pm);

	// transform ecef to eci
	double pmRecef[3];
	pmRecef[0] = pm[0]*recef[0] + pm[3]*recef[1] + pm[6]*recef[2];
	pmRecef[1] = pm[1]*recef[0] + pm[4]*recef[1] + pm[7]*recef[2];
	pmRecef[2] = pm[2]*recef[0] + pm[5]*recef[1] + pm[8]*recef[2];

	double stPmRecef[3];
	stPmRecef[0] = st[0]*pmRecef[0] + st[3]*pmRecef[1] + st[6]*pmRecef[2];
	stPmRecef[1] = st[1]*pmRecef[0] + st[4]*pmRecef[1] + st[7]*pmRecef[2];
	stPmRecef[2] = st[2]*pmRecef[0] + st[5]*pmRecef[1] + st[8]*pmRecef[2];

	double nutStPmRecef[3];
	nutStPmRecef[0] = nut[0]*stPmRecef[0] + nut[3]*stPmRecef[1] + nut[6]*stPmRecef[2];
	nutStPmRecef[1] = nut[1]*stPmRecef[0] + nut[4]*stPmRecef[1] + nut[7]*stPmRecef[2];
	nutStPmRecef[2] = nut[2]*stPmRecef[0] + nut[5]*stPmRecef[1] + nut[8]*stPmRecef[2];

	reci[0] = prec[0]*nutStPmRecef[0] + prec[3]*nutStPmRecef[1] + prec[6]*nutStPmRecef[2];
	reci[1] = prec[1]*nutStPmRecef[0] + prec[4]*nutStPmRecef[1] + prec[7]*nutStPmRecef[2];
	reci[2] = prec[2]*nutStPmRecef[0] + prec[5]*nutStPmRecef[1] + prec[8]*nutStPmRecef[2];
}

/*
 *Frame Transformations
 * function frameEci2ecef
 *
 * Created: 30.01.2015 16:35:59
 * Author: Antoine Pignede
 *
 * This function trsnforms a vector from the mean equator mean equniox frame
 *   (j2000), to an earth fixed (ITRF) frame.  the results take into account
 *   the effects of precession, nutation, sidereal time, and polar motion.
 *
 * Copied and edited from Vallado (2013) Fundamentals of Astrodynamics and Applications
 *
 *  inputs          description                    range / units
 *    reci        - position vector eci            km
 *    timeUTC     - struct containing:
 *      year        - year                           1900 .. 2100
 *      mon         - month                          1 .. 12
 *      day         - day                            1 .. 28,29,30,31
 *      hr          - universal time hour            0 .. 23
 *      min         - universal time min             0 .. 59
 *      sec         - universal time sec             0.0 .. 59.999
 *
 *  outputs       :
 *    recef       - position vector earth fixed    km
 *
 *  locals        :
 *    deltapsi    - nutation angle                 rad
 *    meaneps     - mean obliquity of the ecliptic rad
 *    omega       -                                rad
 *    prec        - matrix for mod - eci
 *    nut         - matrix for tod - mod
 *    st          - matrix for pef - tod
 *    pm          - matrix for ecef - pef
 *    jdUTC       - julian date                    days from 4713 bc
 *    all UTC times are converted to terrestrial time (TT)
 *
 *  coupling      :
 *   framePrecess      - rotation for precession
 *   frameNutation     - rotation for nutation
 *   frameSidereal     - rotation for sidereal time
 *   framePolarm       - rotation for polar motion
 *   timeDatetime2jd   - convert date and time values to julian date
 *   timeUtc2tt        - convert UTC to TT
 *   timeJd2jc         - convert julian date to julian centuries
 *
 * See also
 *   frameEcef2eci.m  - inverse function
 */

void frameEci2ecef(double reci[3],Time timeUTC, double recef[3]){
	// find the julian centuries of terrestrial time at given UTC
	double jdUTC = timeDatetime2jd(timeUTC);
	Time timeTT = timeUtc2tt(timeUTC);
	double jdTT = timeDatetime2jd(timeTT);
	double ttt = timeJd2jc(jdTT);

	// get the transformation matrices
	double prec[9];
	framePrecess(ttt, prec);

	double deltapsi;
	double meaneps;
	double omega;
	double nut[9];
	frameNutation(ttt, &deltapsi, &meaneps, &omega, nut);

	double st[9];
	frameSidereal(jdUTC, deltapsi, meaneps, omega, st);

	double pm[9];
	framePolarm(jdUTC, pm);

	// transform eci to ecef
	double precReci[3];
	precReci[0] = prec[0]*reci[0] + prec[1]*reci[1] + prec[2]*reci[2];
	precReci[1] = prec[3]*reci[0] + prec[4]*reci[1] + prec[5]*reci[2];
	precReci[2] = prec[6]*reci[0] + prec[7]*reci[1] + prec[8]*reci[2];

	double nutPrecReci[3];
	nutPrecReci[0] = nut[0]*precReci[0] + nut[1]*precReci[1] + nut[2]*precReci[2];
	nutPrecReci[1] = nut[3]*precReci[0] + nut[4]*precReci[1] + nut[5]*precReci[2];
	nutPrecReci[2] = nut[6]*precReci[0] + nut[7]*precReci[1] + nut[8]*precReci[2];

	double stNutPrecReci[3];
	stNutPrecReci[0] = st[0]*nutPrecReci[0] + st[1]*nutPrecReci[1] + st[2]*nutPrecReci[2];
	stNutPrecReci[1] = st[3]*nutPrecReci[0] + st[4]*nutPrecReci[1] + st[5]*nutPrecReci[2];
	stNutPrecReci[2] = st[6]*nutPrecReci[0] + st[7]*nutPrecReci[1] + st[8]*nutPrecReci[2];

	recef[0] = pm[0]*stNutPrecReci[0] + pm[1]*stNutPrecReci[1] + pm[2]*stNutPrecReci[2];
	recef[1] = pm[3]*stNutPrecReci[0] + pm[4]*stNutPrecReci[1] + pm[5]*stNutPrecReci[2];
	recef[2] = pm[6]*stNutPrecReci[0] + pm[7]*stNutPrecReci[1] + pm[8]*stNutPrecReci[2];
}

/*
 *Frame Transformations
 * function frameEci2teme
 *
 * Created: 31.01.2015 14:49:06
 * Author: Antoine Pignede
 *
 * This function transforms a vector from the mean equator mean equinox frame
 *   (j2000) to the true equator mean equinox of date (teme) frame.
 *
 * Copied and edited from Vallado (2013) Fundamentals of Astrodynamics and Applications
 *
 *  inputs          description                    range / units
 *    reci        - position vector eci            km
 *    timeUTC     - struct containing:
 *      year        - year                           1900 .. 2100
 *      mon         - month                          1 .. 12
 *      day         - day                            1 .. 28,29,30,31
 *      hr          - universal time hour            0 .. 23
 *      min         - universal time min             0 .. 59
 *      sec         - universal time sec             0.0 .. 59.999
 *
 *  outputs       :
 *    rteme       - position vector of date
 *                    true equator, mean equinox   km
 *
 *  locals        :
 *    prec        - matrix for eci - mod
 *    nutteme     - matrix for mod - teme
 *    all UTC times are converted to terrestrial time (TT)
 *
 *  coupling      :
 *   framePrecess      - rotation for precession        eci - mod
 *   frameTruemean     - rotation for truemean          mod - teme
 *   timeDatetime2jd   - convert date and time values to julian date
 *   timeUtc2tt        - convert UTC to TT
 *   timeJd2jc         - convert julian date to julian centuries
 *
 * See also
 *   frameTeme2eci.m  - inverse function
 */

void frameEci2teme(double reci[3],Time timeUTC, double rteme[3]){
	// find the julian centuries of terrestrial time at given UTC
	Time timeTT = timeUtc2tt(timeUTC);
	double jdTT = timeDatetime2jd(timeTT);
	double ttt = timeJd2jc(jdTT);

	// get the transformation matrices
	double prec[9];
	framePrecess(ttt, prec);

	double nutteme[9];
	frameTruemean(ttt, nutteme);

	// transform eci to teme
	double precReci[3];
	precReci[0] = prec[0]*reci[0] + prec[1]*reci[1] + prec[2]*reci[2];
	precReci[1] = prec[3]*reci[0] + prec[4]*reci[1] + prec[5]*reci[2];
	precReci[2] = prec[6]*reci[0] + prec[7]*reci[1] + prec[8]*reci[2];

	rteme[0] = nutteme[0]*precReci[0] + nutteme[1]*precReci[1] + nutteme[2]*precReci[2];
	rteme[1] = nutteme[3]*precReci[0] + nutteme[4]*precReci[1] + nutteme[5]*precReci[2];
	rteme[2] = nutteme[6]*precReci[0] + nutteme[7]*precReci[1] + nutteme[8]*precReci[2];
}

/*
 *Frame Transformations
 * function frameTeme2eci
 *
 * Created: 31.01.2015 15:18:33
 * Author: Antoine Pignede
 *
 * This function transforms a vector from the true equator mean equinox of date,
 *   (teme) frame to the mean equator mean equinox (j2000) frame.
 *
 * Copied and edited from Vallado (2013) Fundamentals of Astrodynamics and Applications
 *
 *  inputs          description                    range / units
 *    rteme       - position vector of date
 *                    true equator, mean equinox   km
 *    timeUTC     - struct containing:
 *      year        - year                           1900 .. 2100
 *      mon         - month                          1 .. 12
 *      day         - day                            1 .. 28,29,30,31
 *      hr          - universal time hour            0 .. 23
 *      min         - universal time min             0 .. 59
 *      sec         - universal time sec             0.0 .. 59.999
 *
 *  outputs       :
 *    reci        - position vector eci            km
 *
 *  locals        :
 *    prec        - matrix for eci - mod
 *    nutteme     - matrix for mod - teme - an approximation for nutation
 *    all UTC times are converted to terrestrial time (TT)
 *
 *  coupling      :
 *   framePrecess      - rotation for precession        eci - mod
 *   frameTruemean     - rotation for truemean          eci - teme
 *   timeDatetime2jd   - convert date and time values to julian date
 *   timeUtc2tt        - convert UTC to TT
 *   timeJd2jc         - convert julian date to julian centuries
 *
 * See also
 *   frameEci2teme.m  - inverse function
 */

void frameTeme2eci(double rteme[3],Time timeUTC, double reci[3]){
	// find the julian centuries of terrestrial time at given UTC
	Time timeTT = timeUtc2tt(timeUTC);
	double jdTT = timeDatetime2jd(timeTT);
	double ttt = timeJd2jc(jdTT);

	// get the transformation matrices
	double prec[9];
	framePrecess(ttt, prec);

	double nutteme[9];
	frameTruemean(ttt, nutteme);

	// transform teme to eci
	double nuttemeRteme[3];
	nuttemeRteme[0] = nutteme[0]*rteme[0] + nutteme[3]*rteme[1] + nutteme[6]*rteme[2];
	nuttemeRteme[1] = nutteme[1]*rteme[0] + nutteme[4]*rteme[1] + nutteme[7]*rteme[2];
	nuttemeRteme[2] = nutteme[2]*rteme[0] + nutteme[5]*rteme[1] + nutteme[8]*rteme[2];

	reci[0] = prec[0]*nuttemeRteme[0] + prec[3]*nuttemeRteme[1] + prec[6]*nuttemeRteme[2];
	reci[1] = prec[1]*nuttemeRteme[0] + prec[4]*nuttemeRteme[1] + prec[7]*nuttemeRteme[2];
	reci[2] = prec[2]*nuttemeRteme[0] + prec[5]*nuttemeRteme[1] + prec[8]*nuttemeRteme[2];
}

/*
 *Frame Transformations
 * function frameLatitudeRecur
 *
 * Created: 30.01.2015 15:51:02
 * Author: Antoine Pignede
 *
 * Compute geodetic latitude recursively.
 *
 * Copied and edited from Compston (2011) Matlab File Exchange International Geomagnetic Reference Field (IGRF) Model
 *
 * Inputs:
 *   -lat_in: current best estimate of latitude in rad
 *	 -z: ECEF z position in km
 *	 -rd: square root of the sum of the squares of ECEF x and y position in km
 *	 -iter: iteration number of recursive function, after a maximum of 20 the recursion stops
 *
 * Ouputs:
 *   -latitude: new best estimate of latitude in rad
 *	 -Nphi: Normal: the distance from the surface to the Z-axis along the ellipsoid normal. See http://en.wikipedia.org/wiki/Geodetic_datum#Conversion_calculations
 *
 * Note that there is a limit to 20 iterations if the maximum error tolerance
 *   in the latitude in radians is not reached (I never had more than 10 iterations).
 *   In contrast to the original function the tolerance is set to 1e-12 and not an input.
 *
 * See also
 *   frameEcef2geod.m  - function that needs frameLatitudeRecur
 */

void frameLatitudeRecur(double lat_in, double z, double rd, int8_t iter, double *latitude, double *Nphi){
	double sinlat_in = sin(lat_in);
    double thisNphi = REQU / sqrt(1.0 - ECCSQRD*sinlat_in*sinlat_in);
    double nextlat = atan((z + thisNphi*ECCSQRD*sinlat_in) / rd);

    if ((fabs(lat_in - nextlat) < 1e-12) || (iter > 20)){
        *latitude = nextlat;
        *Nphi = thisNphi;
    }
    else frameLatitudeRecur(nextlat, z, rd, iter + 1, latitude, Nphi);
}

/*
 *Frame Transformations
 * function frameFundarg
 *
 * Created: 31.01.2015 10:23:11
 * Author: Antoine Pignede
 *
 * This function calulates the delauany variables at given julian centuries.
 *
 * Copied and edited from Vallado (2013) Fundamentals of Astrodynamics and Applications
 *
 *  inputs          description                    range / units
 *    ttt         - julian centuries of tt
 *
 *  outputs       :
 *    l           - Delaunay argument               rad
 *    ll          - Delaunay argument               rad
 *    f           - Delaunay argument               rad
 *    d           - Delaunay argument               rad
 *    omega       - Delaunay argument               rad
 *
 *  locals        :
 *    ttt2,ttt3,ttt4  - powers of ttt
 *
 *  coupling      :
 *    none        -
 *
 * Note that the original function provides four different formulas for each
 *   Delaunay argument (1980, 1996, 2000b, 2010). The test routines use 1980
 *   for ecef2eci/eci2ecef and 2010 (but labelled as 1980) for
 *   eci2teme/teme2eci. Here just 2010 is used, the differences are small.
 *
 * See also
 *   frameNutation.m, frameTruemean.m  - function that needs frameFundarg
 */

Delaunay frameFundarg(double ttt){
	Delaunay fundarg;
	double ttt2= ttt*ttt;
	double ttt3= ttt2*ttt;
	double ttt4= ttt2*ttt2;

	// determine coefficients for iau 2010 nutation theory
	fundarg.l    =  134.96340251  + ( 1717915923.2178 *ttt + 31.8792 *ttt2 +
			0.051635 *ttt3 - 0.00024470 *ttt4 ) / 3600.0;
	fundarg.l1   =  357.52910918  + (  129596581.0481 *ttt -  0.5532 *ttt2 -
			0.000136 *ttt3 - 0.00001149 *ttt4 ) / 3600.0;
	fundarg.f    =   93.27209062  + ( 1739527262.8478 *ttt - 12.7512 *ttt2 +
			0.001037 *ttt3 + 0.00000417 *ttt4 ) / 3600.0;
	fundarg.d    =  297.85019547  + ( 1602961601.2090 *ttt -  6.3706 *ttt2 +
			0.006593 *ttt3 - 0.00003169 *ttt4 ) / 3600.0;
	fundarg.omega=  125.04455501  + (   -6962890.2665 *ttt +  7.4722 *ttt2 +
			0.007702 *ttt3 - 0.00005939 *ttt4 ) / 3600.0;

	// convert units to rad
	fundarg.l    = fmod(fundarg.l, 360.0)     * DEG2RAD;
	fundarg.l1   = fmod(fundarg.l1, 360.0)    * DEG2RAD;
	fundarg.f    = fmod(fundarg.f, 360.0)     * DEG2RAD;
	fundarg.d    = fmod(fundarg.d, 360.0)     * DEG2RAD;
	fundarg.omega= fmod(fundarg.omega, 360.0) * DEG2RAD;

	return fundarg;
}

/*
 *Frame Transformations
 * function frameNutation
 *
 * Created: 31.01.2015 10:23:11
 * Author: Antoine Pignede
 *
 * This function calulates the transformation matrix that accounts for the
 *   effects of nutation.
 *
 * Copied and edited from Vallado (2013) Fundamentals of Astrodynamics and Applications
 *
 *  inputs          description                    range / units
 *    ttt         - julian centuries of tt
 *
 *  outputs       :
 *    deltapsi    - nutation angle                 rad
 *    meaneps     - mean obliquity of the ecliptic rad
 *    omega       - Delaunay argument              rad
 *    nut         - transformation matrix for tod - mod
 *
 *  locals        :
 *    ddpsi       - delta psi correction to gcrf   rad
 *    ddeps       - delta eps correction to gcrf   rad
 *    iar80       - integers for fk5 1980
 *    rar80       - reals for fk5 1980
 *    ttt2        - ttt squared
 *    ttt3        - ttt cubed
 *    l           - Delaunay argument              rad
 *    ll          - Delaunay argument              rad
 *    f           - Delaunay argument              rad
 *    d           - Delaunay argument              rad
 *    deltaeps    - change in obliquity            rad
 *
 *  coupling      :
 *    timeJc2jd        - convert julian century to julian date
 *    frameFundarg     - find fundamental arguments
 *
 * Note that the original function takes the two correction to gcrf factors
 *   ddpsi and ddeps as input parameters. The linear interpolation of Kelso
 *   and Vallado (2013) Earth Orientation Parameter and Space Weather Data for Flight Operations
 *   is used instead.
 *
 * See also
 *   frameEcef2eci.m, frameEci2ecef.m  - transformation that need frameNutation
 */

void frameNutation(double ttt, double *deltapsi, double *meaneps, double *omega, double nut[9]){
/* Kelso and and Vallado (2013) Earth Orientation Parameter and Space Weather Data for Flight Operations */
	double jd = timeJc2jd(ttt);
	double MJD = jd - 2400000.5;
	double conv = DEG2RAD / 3600;
	double ddpsi = (-8e-6 * MJD + 0.2506) * conv;
	double ddeps = (-7e-7 * MJD + 0.022) * conv;

/* Vallado (2013) Fundamentals of Astrodynamics and Applications */
	double ttt2= ttt*ttt;
	double ttt3= ttt2*ttt;

	// get the Delaunay arguments
	Delaunay fundarg = frameFundarg(ttt);
	*omega = fundarg.omega;

	// find nutation angle and change in obliquity
	double convrt = 0.0001 * DEG2RAD / 3600.0;
	*deltapsi= 0.0;
	double deltaeps= 0.0;
	int i;
	for (i=105; i>=0; i--){ // reduce 105 (to 10, 20, ...) to speed up if needed
		double tempval= IAR80[i]*fundarg.l + IAR80[i+106]*fundarg.l1 +
				IAR80[i+212]*fundarg.f + IAR80[i+318]*fundarg.d + IAR80[i+424]*fundarg.omega;
		*deltapsi= *deltapsi + (RAR80[i]*convrt + RAR80[i+106]*convrt*ttt) * sin(tempval);
		deltaeps= deltaeps + (RAR80[i+212]*convrt + RAR80[i+318]*convrt*ttt) * cos(tempval);
	}

	// correction to gcrf
	*deltapsi = fmod(*deltapsi + ddpsi, TWOPI);
	deltaeps = fmod(deltaeps + ddeps, TWOPI);

	// mean and true obliquity of the ecliptic
	*meaneps = -46.8150 *ttt - 0.00059 *ttt2 + 0.001813 *ttt3 + 84381.448;
	*meaneps = fmod(*meaneps/3600.0, 360.0);
	*meaneps = *meaneps * DEG2RAD;
	double trueeps  = *meaneps + deltaeps;

	double cospsi  = cos(*deltapsi);
	double sinpsi  = sin(*deltapsi);
	double coseps  = cos(*meaneps);
	double sineps  = sin(*meaneps);
	double costrueeps = cos(trueeps);
	double sintrueeps = sin(trueeps);

	// build nutation matrix
	nut[0] =  cospsi;
	nut[1] = -coseps * sinpsi;
	nut[2] = -sineps * sinpsi;
	nut[3] =  costrueeps * sinpsi;
	nut[4] =  costrueeps * coseps * cospsi + sintrueeps * sineps;
	nut[5] =  costrueeps * sineps * cospsi - sintrueeps * coseps;
	nut[6] =  sintrueeps * sinpsi;
	nut[7] =  sintrueeps * coseps * cospsi - sineps * costrueeps;
	nut[8] =  sintrueeps * sineps * cospsi + costrueeps * coseps;
}

/*
 *Frame Transformation
 * function framePolarm
 *
 * Created: 31.01.2015 11:37:42
 * Author: Antoine Pignede
 *
 * This function calulates the transformation matrix that accounts for polar
 *   motion.
 *
 * Copied and edited from Vallado (2013) Fundamentals of Astrodynamics and Applications
 * Merged with IERS BULLETIN-A 12 February 2015 Vol. XXVIII No. 007 found at
 *   http://www.iers.org/IERS/EN/Publications/Bulletins/bulletins.html for
 *   the approximation of the polar motion coefficients over a long (years) time range.
 *
 *  inputs          description                    range / units
 *    jd          - julian date                    days from 4713 bc
 *
 *  outputs       :
 *    pm          - transformation matrix for ecef - pef
 *
 *  locals        :
 *    convrt      - conversion from arcsec to rad
 *    xp          - polar motion coefficient       rad
 *    yp          - polar motion coefficient       rad
 *
 *  coupling      :
 *    none.
 *
 * Note that the original function provides two different formulas (1980 and
 *   2000). The test routines use 1980 which is also the option taken here,
 *   the differences are small.
 *
 * See also
 *   frameEcef2eci.m, frameEci2ecef.m  - transformation that need framePolarm
 */

void framePolarm(double jdutc, double pm[9]){
/* IERS BULLETIN-A 12 February 2015 Vol. XXVIII No. 007 http://www.iers.org/IERS/EN/Publications/Bulletins/bulletins.html */
	double MJD = jdutc - 2400000.5;
	double A = TWOPI*(MJD-MJDPOLAR)/365.25;
	double C = TWOPI*(MJD-MJDPOLAR)/435.0;
	double sinA = sin(A);
	double cosA = cos(A);
	double sinC = sin(C);
	double cosC = cos(C);
	double xp = XPOLAR[0] + XPOLAR[1]*cosA + XPOLAR[2]*sinA +
			XPOLAR[3]*cosC + XPOLAR[4]*sinC;
	double yp = YPOLAR[0] + YPOLAR[1]*cosA + YPOLAR[2]*sinA +
			YPOLAR[3]*cosC + YPOLAR[4]*sinC;

	// convert arcsec to rad
	double conv = DEG2RAD / 3600;
	xp = xp * conv;
	yp = yp * conv;

/* Vallado (2013) Fundamentals of Astrodynamics and Applications */
	double cosxp = cos(xp);
	double sinxp = sin(xp);
	double cosyp = cos(yp);
	double sinyp = sin(yp);

	// build polar motion matrix
	pm[0] =  cosxp;
	pm[1] =  sinxp * sinyp;
	pm[2] =  sinxp * cosyp;
	pm[3] =  0.0;
	pm[4] =  cosyp;
	pm[5] = -sinyp;
	pm[6] = -sinxp;
	pm[7] =  cosxp * sinyp;
	pm[8] =  cosxp * cosyp;
}

/*
 *Frame Transformation
 * function framePrecess
 *
 * Created: 31.01.2015 10:23:11
 * Author: Antoine Pignede
 *
 * This function calulates the transformation matrix that accounts for the effects
 *   of precession.
 *
 * Copied and edited from Vallado (2013) Fundamentals of Astrodynamics and Applications
 *
 *  inputs          description                    range / units
 *    ttt         - julian centuries of tt
 *
 *  outputs       :
 *    prec        - transformation matrix for mod - j2000 (80 only)
 *
 *  locals        :
 *    convrt      - conversion from arcsec to rad
 *    ttt2        - ttt squared
 *    ttt3        - ttt cubed
 *    zeta        - precession angle               rad
 *    z           - precession angle               rad
 *    theta       - precession angle               rad
 *
 *  coupling      :
 *    none        -
 *
 * Note that the original function provides three different formulas (b1950,
 *   1980 and 03). The test routines use 1980 which is also the option taken
 *   here, the differences are small.
 *
 * See also
 *   frameEcef2eci.m, frameEci2ecef.m  - transformation that need framePrecess
 *   frameEci2teme.m, frameTeme2eci.m  - transformation that need framePrecess */

void framePrecess(double ttt, double prec[9]){
	double convrt = DEG2RAD / 3600.0;
	double ttt2 = ttt * ttt;
	double ttt3 = ttt2 * ttt;

	// find the precession angles
	double zeta = 2306.2181*ttt + 0.30188*ttt2 + 0.017998*ttt3;
	double theta= 2004.3109*ttt - 0.42665*ttt2 - 0.041833*ttt3;
	double z    = 2306.2181*ttt + 1.09468*ttt2 + 0.018203*ttt3;

	zeta = zeta  * convrt;
	theta= theta * convrt;
	z    = z     * convrt;

	double coszeta  = cos(zeta);
	double sinzeta  = sin(zeta);
	double costheta = cos(theta);
	double sintheta = sin(theta);
	double cosz     = cos(z);
	double sinz     = sin(z);

	// build precession matrix
	prec[0] =  coszeta * costheta * cosz - sinzeta * sinz;
	prec[1] = -sinzeta * costheta * cosz - coszeta * sinz;
	prec[2] = -sintheta * cosz;
	prec[3] =  coszeta * costheta * sinz + sinzeta * cosz;
	prec[4] = -sinzeta * costheta * sinz + coszeta * cosz;
	prec[5] = -sintheta * sinz;
	prec[6] =  coszeta * sintheta;
	prec[7] = -sinzeta * sintheta;
	prec[8] =  costheta;
}

/*  *Frame Transformations
 * function frameSidereal
 *
 * Created: 31.01.2015 11:17:37
 * Author: Antoine Pignede
 *
 * This function calulates the transformation matrix that accounts for the
 *   effects of sidereal time. Notice that deltaspi should not be moded to a
 *   positive number because it is multiplied rather than used in a
 *   trigonometric argument.
 *
 * Copied and edited from Vallado (2013) Fundamentals of Astrodynamics and Applications
 *
 *  inputs          description                    range / units
 *    jdutc       - julian centuries of utc        days
 *    deltapsi    - nutation angle                 rad
 *    meaneps     - mean obliquity of the ecliptic rad
 *    omega       - long of asc node of moon       rad
 *
 *  outputs       :
 *    st          - transformation matrix for pef - tod
 *
 *  locals        :
 *    gmst        - mean greenwich sidereal time   0 to 2pi rad
 *    ast         - apparent gmst                  0 to 2pi rad
 *
 *  coupling      :
 *    timeGstime      - find gmst
 *
 * Note that the original function uses ut1 instead of utc to get the gmst.
 *   For simplicity ut1 is approximated by utc here. This is legitimiate
 *   because abs(ut1 - utc) < 0.9s which is negligible.
 *
 * See also
 *   frameEcef2eci.m, frameEci2ecef.m  - transformation that need frameSidereal */

void frameSidereal(double jdutc,double deltapsi,double meaneps, double omega, double st[9]){
	// find mean ast
	double gmst= timeGstime(jdutc);
	double ast;
	if (jdutc > 2450449.5)
		ast= gmst + deltapsi * cos(meaneps) + 0.00264 * DEG2RAD/3600.0 * sin(omega) +
		0.000063 * DEG2RAD/3600.0 * sin(2.0*omega);
	else
		ast= gmst + deltapsi* cos(meaneps);
	ast = fmod(ast,TWOPI);
	double cosast = cos(ast);
	double sinast = sin(ast);

	// build sidereal time matrix
	st[0] =  cosast;
	st[1] =  sinast;
	st[2] =  0.0;
	st[3] = -sinast;
	st[4] =  cosast;
	st[5] =  0.0;
	st[6] =  0.0;
	st[7] =  0.0;
	st[8] =  1.0;
}

/*  *Frame Transformations
 * function frameTruemean
 *
 * Created: 31.01.2015 14:55:18
 * Author: Antoine Pignede
 *
 * This function forms the transformation matrix to go between the
 *   norad true equator mean equinox of date and the mean equator mean equinox
 *   of date (j2000).  the results approximate the effects of nutation and
 *   sidereal time.
 *
 * Copied and edited from Vallado (2013) Fundamentals of Astrodynamics and Applications
 *
 *  inputs          description                    range / units
 *    ttt         - julian centuries of tt
 *
 *  outputs       :
 *    nutteme     - matrix for mod - teme - an approximation for nutation
 *
 *  locals        :
 *    ttt2        - ttt squared
 *    ttt3        - ttt cubed
 *    l           - Delaunay argument               rad
 *    ll          - Delaunay argument               rad
 *    f           - Delaunay argument               rad
 *    d           - Delaunay argument               rad
 *    omega       - Delaunay argument               rad
 *    deltapsi    - nutation angle                 rad
 *    deltaeps    - change in obliquity            rad
 *    trueeps     - true obliquity of the ecliptic rad
 *    meaneps     - mean obliquity of the ecliptic rad
 *
 *  coupling      :
 *    frameFundarg     - find fundamental arguments
 *
 * Note that the original function provides the 2010 formulas for the
 *   Delaunay arguments, frameFundarg is used here.
 * The original function also gives the user the possibility of chosing
 *   between complete nutation, truncated nutation and truncated transfer
 *   matrix. The most accurate version (complete nutation) is used here.
 *   The original function also gives the user the possibility of chosing
 *   the number of terms for nutation (4, 50 and 106). As in frameNutation.m
 *   the most accurate option 106 is taken here.
 *
 * See also
 *   frameTeme2eci.m, frameEci2teme.m  - transformation that need frameTruemean */

void frameTruemean(double ttt, double nutteme[9]){
	double ttt2= ttt*ttt;
	double ttt3= ttt2*ttt;

	// get the Delaunay arguments
	Delaunay fundarg = frameFundarg(ttt);

	// find nutation angle and change in obliquity
	double convrt = 0.0001 * DEG2RAD / 3600.0;
	double deltapsi= 0.0;
	double deltaeps= 0.0;
	int i;
	for (i=105; i>=0; i--){ // reduce 105 (to 10, 20, ...) to speed up if needed
		double tempval= IAR80[i]*fundarg.l + IAR80[i+106]*fundarg.l1 +
				IAR80[i+212]*fundarg.f + IAR80[i+318]*fundarg.d + IAR80[i+424]*fundarg.omega;
		deltapsi= deltapsi + (RAR80[i]*convrt + RAR80[i+106]*convrt*ttt) * sin(tempval);
		deltaeps= deltaeps + (RAR80[i+212]*convrt + RAR80[i+318]*convrt*ttt) * cos(tempval);
	}
	deltapsi = fmod(deltapsi, 360.0) * DEG2RAD;
	deltaeps = fmod(deltaeps, 360.0) * DEG2RAD;

	// mean and true obliquity of the ecliptic
	double meaneps = -46.8150 *ttt - 0.00059 *ttt2 + 0.001813 *ttt3 + 84381.448;
	meaneps = fmod(meaneps/3600.0, 360.0);
	meaneps = meaneps * DEG2RAD;
	double trueeps  = meaneps + deltaeps;

	double cospsi  = cos(deltapsi);
	double sinpsi  = sin(deltapsi);
	double coseps  = cos(meaneps);
	double sineps  = sin(meaneps);
	double costrueeps = cos(trueeps);
	double sintrueeps = sin(trueeps);

	// small disconnect with ttt instead of ut1
	double jdttt = timeJc2jd(ttt);
	double eqe;
	if (jdttt > 2450449.5)
		eqe = deltapsi * cos(meaneps) + 0.00264*DEG2RAD/3600 * sin(fundarg.omega) +
			0.000063*DEG2RAD/3600 * sin(2.0*fundarg.omega);
    else eqe = deltapsi*cos(meaneps);
	double coseqe = cos(eqe);
	double sineqe = sin(eqe);

	// build nutation matrix
	double nut[9];
	nut[0] =  cospsi;
	nut[1] = -coseps * sinpsi;
	nut[2] = -sineps * sinpsi;
	nut[3] =  costrueeps * sinpsi;
	nut[4] =  costrueeps * coseps * cospsi + sintrueeps * sineps;
	nut[5] =  costrueeps * sineps * cospsi - sintrueeps * coseps;
	nut[6] =  sintrueeps * sinpsi;
	nut[7] =  sintrueeps * coseps * cospsi - sineps * costrueeps;
	nut[8] =  sintrueeps * sineps * cospsi + costrueeps * coseps;

	// build sidereal time matrix
	double st[9];
	st[0] =  coseqe;
	st[1] =  sineqe;
	st[2] =  0.0;
	st[3] = -sineqe;
	st[4] =  coseqe;
	st[5] =  0.0;
	st[6] =  0.0;
	st[7] =  0.0;
	st[8] =  1.0;

	// build transformation matrix
	nutteme[0] = nut[0]*st[0] + nut[3]*st[1]; // nut[6]*st[2]==0
	nutteme[1] = nut[1]*st[0] + nut[4]*st[1]; // nut[7]*st[2]==0
	nutteme[2] = nut[2]*st[0] + nut[5]*st[1]; // nut[8]*st[2]==0
	nutteme[3] = nut[0]*st[3] + nut[3]*st[4]; // nut[6]*st[5]==0
	nutteme[4] = nut[1]*st[3] + nut[4]*st[4]; // nut[7]*st[5]==0
	nutteme[5] = nut[2]*st[3] + nut[5]*st[4]; // nut[8]*st[5]==0
	nutteme[6] = nut[6]; // nut[0]*st[6]==0, nut[3]*st[7]==0, nut[6]*st[8]==nut[6]
	nutteme[7] = nut[7]; // nut[1]*st[6]==0, nut[4]*st[7]==0, nut[7]*st[8]==nut[7]
	nutteme[8] = nut[8]; // nut[2]*st[6]==0, nut[5]*st[7]==0, nut[8]*st[8]==nut[8]
}

/*  *Frame Transformations
 * function frameRotationNed2ecef
 *
 * Created: 07.02.2015 10:49:52
 * Author: Antoine Pignede
 *
 * Computes the rotation matrix to transfrom the North East Down (NED) geomagnetic
 *   vector into an Earth-centered, Earth fixed (ECEF) geomagnetic vector (X, Y, and Z)
 *   given the sin and cos of the latitude and longitude at the location where the
 *   geomagnetic vector was calculated.
 *
 * Copied and edited from Fossen (2011) Handbook of Marine Craft Hydrodynamics and Motion Control
 *
 * Inputs:
 *   -coslat: cosine of the latitude of the location
 *   -sinlat: sine of the latitude of the location
 *   -coslon: cosine of the longitude of the location
 *   -sinlon: sine of the longitude of the location
 *
 * Ouputs:
 *   -Ned2ecef: North East Down to ECEF rotation matrix
 *
 * See also
 *   magIgrf.m, magWmm.m  - functions that use frameRotationNed2ecef
 */

void frameRotationNed2ecef(double sinlat, double coslat, double sinlon, double coslon, double Ned2ecef[9]){
	// build transformation matrix
	Ned2ecef[0] = -coslon*sinlat;
	Ned2ecef[1] = -sinlon*sinlat;
	Ned2ecef[2] = coslat;
	Ned2ecef[3] = -sinlon;
	Ned2ecef[4] = coslon;
	Ned2ecef[5] = 0.0;
	Ned2ecef[6] = -coslon*coslat;
	Ned2ecef[7] = -sinlon*coslat;
	Ned2ecef[8] = -sinlat;
}
