/*
 * magPred.c
 *
 *  Created: Feb 17, 2015 3:14:05 PM
 *  Author: Antoine Pignede
 *
 *  This file defines the Geomagnetic Vector Prediction.
 *  Two different models, the International Geomagnetic Reference Field (IGRF)
 *    and the World Magnetic Model (WMM), are used.
 *    It is recommended to use the WMM as it prooved to be faster in the testing
 *    phase on the microcontroller, but both models are equally reliable.
 */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "constants.h"
#include "timeConv.h"
#include "frameTrans.h"
#include "magPred.h"

/*
 *Geomagnetic Vector Prediction
 * function magIgrf
 *
 * Created: 06.02.2015 08:26:22
 * Author: Antoine Pignede
 *
 * Calculates the components of the Earth's magnetic field using the
 *   International Geomagnetic Reference Field (IGRF) model.
 *   Note that the other parameters the IGRF gives can be computed from BX,
 *   BY, and BZ as:
 *     Horizonal intensity: hypot(BX, BY) (i.e., sqrt(BX.^2 + BY.^2) )
 *     Total intensity: hypot(BX, hypot(BY, BZ))
 *     Declination: atan2(BY, BX)
 *     Inclination: atan(BZ./hypot(BX, BY))
 *
 * Copied and edited from Compston (2011) Matlab File Exchange International Geomagnetic Reference Field (IGRF) Model
 *
 * Inputs:
 *    reci        - position vector eci            km
 *    timeUTC     - struct containing:
 *      year        - year                           1900 .. 2100
 *      mon         - month                          1 .. 12
 *      day         - day                            1 .. 28,29,30,31
 *      hr          - universal time hour            0 .. 23
 *      min         - universal time min             0 .. 59
 *      sec         - universal time sec             0.0 .. 59.999
 *
 * Outputs:
 *    Beci        - geomagnetic field vector in nanotesla (nT) in eci frame
 *
 * Coupling
 *    timeDatetime2years     - convert date and time values to fractional years
 *    frameEci2ecef     - transform eci position to ecef position
 *    frameEcef2geod    - transform ecef position to geodetic position
 *
 * Note that the original getigrfcoefs, loadigrfcoefs, frame and time
 *    transformation functions have been merged into this igrf function.
 *    The original vector input function to calculate the geomagnetic field
 *    for more than just one position has been removed.
 *
 * See also
 *    magWmm.m  - alternative geomagnetic vector predictor
 */

void magIgrf(double reci[3], Time timeUTC, double Beci[3]){
/* Edited excerpt of loadigrfcoefs.m from Compston (2011) Matlab File Exchange International Geomagnetic Reference Field (IGRF) Model */
	// Convert time to fractional years.
	double time = timeDatetime2years(timeUTC);

	// Interpolate g and h linearly using the values at epoch and secular variation coefficients
	double gh[196]; // numel(IGRFGH)==196
	int i;
	for (i=0; i<=195; i++) gh[i] = IGRFGH[i] + IGRFGHSV[i]*(time - 2015.0);

	// Highest order
	int nmax = sqrt(195 + 1) - 1; // ==13 // reduce (to 7, 8, ...) to speed up

/* Edited excerpt of igrf.m from Compston (2011) Matlab File Exchange International Geomagnetic Reference Field (IGRF) Model */
	// Transform eci input vector to geodetic coordinates
	double recef[3];
	frameEci2ecef(reci,timeUTC,recef);
	double rgeod[3];
	frameEcef2geod(recef,rgeod);
	double latitude = rgeod[0];
	double longitude = rgeod[1];
	double altitude = rgeod[2];

	/*
	 * SPHERICAL COORDINATE CONVERSION
	 * Convert from geodetic coordinates latitude, longitude, and altitude to
	 * geocentric coordinates r (radius), theta (inclination angle from +z axis),
	 * and phi (azimuth angle from +x axis).
	 * This method was adapted from igrf11syn, which was a conversion of the
	 * IGRF subroutine written in FORTRAN.
	 */
	double coslat = cos(latitude*DEG2RAD);
	double sinlat = sin(latitude*DEG2RAD);
	double costheta = cos((90.0 - latitude)*DEG2RAD);
	double sintheta = sin((90.0 - latitude)*DEG2RAD);
	double rho = hypot(REQU*sintheta, RPOL*costheta);
	double r = sqrt(altitude*altitude + 2*altitude*rho +
			(pow(REQU,4)*sintheta*sintheta + pow(RPOL,4)*costheta*costheta) / (rho*rho));
	double cd = (altitude + rho) / r;
	double sd = (REQU*REQU - RPOL*RPOL)/rho * costheta*sintheta/r;
	double oldcos = costheta;
	costheta = costheta*cd - sintheta*sd;
	sintheta = sintheta*cd + oldcos*sd;

	// Convert longitude to radians
	double phi = longitude*DEG2RAD;

	// We need cos(m*phi) and sin(m*phi) multiple times
	double cosphi[13+1]; // nmax==13
	double sinphi[13+1]; // nmax==13
	cosphi[0] = 1.0; // not needed, but helps with indices: cosphi[i] = cos(i*phi);
	sinphi[0] = 0.0; // not needed, but helps with indices: sinphi[i] = sin(i*phi);
	for (i=1; i<=nmax; i++){
		cosphi[i] = cos(i*phi);
		sinphi[i] = sin(i*phi);
	}

	/* BEGIN MAGNETIC FIELD CALCULATION */
	// Initialize variables used in for loop below
	int Pmax = (nmax+1)*(nmax+2)/2; // ==105
	double Br = 0.0;
	double Bt = 0.0;
	double Bp = 0.0;
	double P[105+1]; // Pmax==105
	P[0] = 0.0; // not needed, but helps with indices
	P[1] = 1.0;
	P[3] = sintheta;
	double dP[105+1]; // Pmax==105
	dP[0] = 0.0; // not needed, but helps with indices
	dP[1] = 0.0;
	dP[3] = costheta;
	double coef;
	int m = 1;
	int n = 0;
	int coefindex = 1;
	int last1n;
	int last2n;
	int lastn;

	// Relative radius of position with respect to mean earth radius squared
	double a_r = (RMEAN/r)*(RMEAN/r);

	/*
	 * Increment through all the n's and m's. gh will be a vector with g
	 * followed by h for incrementing through n and m except when h would be
	 * redundant (i.e., when m = 0).
	 */
	int Pindex;
	for (Pindex=2; Pindex<=Pmax; Pindex++){

		// Increment to the next n when m becomes larger than n.
		if (n < m){
			m = 0;
			n = n + 1;
			a_r = a_r*(RMEAN/r); // We need (Rearth_km/r)^(n+2)
		}

		/*
		 *  Calculate P and dP. They are given recursively according to:
		 *
		 * P(0, 0) = 1, P(1, 1) = sin(theta) <- Specified above
		 * P(n, n) = sqrt(1 - 1/(2n))*sin(theta)*P(n-1, n-1)
		 * P(n, m) = (2n - 1)/sqrt(n^2 - m^2)*cos(theta)*P(n-1, m) -
		 *     sqrt(((n-1)^2 - m^2) / (n^2 - m^2)) * P(n-2, m)
		 *
		 * dP(0, 0) = 0, dP(1, 1) = cos(theta) <- Specified above
		 * dP(n, n) = sqrt(1 - 1/(2n))*(sin(theta)*dP(n-1, n-1) +
		 *     cos(theta)*P(n-1, n-1))
		 * dP(n, m) = (2n - 1)/sqrt(n^2 - m^2)*(cos(theta)*dP(n-1, m) -
		 *     sin(theta)*P(n-1, m)) - sqrt(((n-1)^2 - m^2)/(n^2 - m^2))*
		 *     dP(n-2, m)
		 */
		if ((m < n) && (Pindex != 3)){ // (Pindex=3 is n=1, m=1, initial cond. above)
			last1n = Pindex - n;
			last2n = Pindex - 2*n + 1;
			P[Pindex] = (2*n - 1.0)/sqrt(n*n - m*m)*costheta*P[last1n] - sqrt(((n-1)*(n-1) - m*m) / (double)(n*n - m*m)) * P[last2n];
			dP[Pindex] = (2*n - 1.0)/sqrt(n*n - m*m)*(costheta*dP[last1n] - sintheta*P[last1n]) - sqrt(((n-1)*(n-1) - m*m) / (double)(n*n - m*m)) * dP[last2n];
		}
		else{
			if (Pindex != 3){
				lastn = Pindex - n - 1;
				P[Pindex] = sqrt(1.0 - 1.0/(2.0*m))*sintheta*P[lastn];
				dP[Pindex] = sqrt(1.0 - 1.0/(2.0*m))*(sintheta*dP[lastn] + costheta*P[lastn]);
			}
		}

		/*
		 * Calculate the magnetic field components as a running sum. Find
		 * explicit expressions for these in Global Earth Physics: a Handbook of
		 * Physical Constants by Thomas J. Aherns (1995), pg. 49. Link:
		 * http://books.google.com/books?id=aqjU_NHyre4C&lpg=PP1&dq=Global *20
		 * earth%20physics%3A%20a%20handbook%20of%20physical%20constants&pg=PA49
		 * #v=onepage&q&f=false
		 * (except equation 6 is missing a required 1/sin(theta) and m; correct
		 * equations on page 5 (equations 3a-3c) of:
		 * http://hanspeterschaub.info/Papers/UnderGradStudents/
		 * MagneticField.pdf)
		 */
		if (m == 0){ // Implies h = 0, so only coefficient in gh is g
			coef = a_r*gh[coefindex]; //*cos(0*phi) = 1
			Br = Br + (n+1)*coef*P[Pindex];
			Bt = Bt - coef*dP[Pindex];
			// Bp is 0 for m = 0.
			coefindex = coefindex + 1; // Only need to skip over g this time.
		}
		else{
			coef = a_r*(gh[coefindex]*cosphi[m] + gh[coefindex+1]*sinphi[m]);
			Br = Br + (n+1)*coef*P[Pindex];
			Bt = Bt - coef*dP[Pindex];
			if (fabs(sintheta) < 1e-6) // Use different formula when dividing by 0.
				Bp = Bp - costheta*a_r*(-gh[coefindex]*sinphi[m] + gh[coefindex+1]*cosphi[m])*dP[Pindex];
			else
				Bp = Bp - 1.0/sintheta*a_r*m*(-gh[coefindex]*sinphi[m] + gh[coefindex+1]*cosphi[m])*P[Pindex];
			coefindex = coefindex + 2; // Skip over g and h this time.
		}
		m = m + 1;
	}

	// Convert from spherical to (North,East,Down).
	double Bx = -Bt;
	double By = Bp;
	double Bz = -Br;

	double Bned[3];
	Bned[0] = Bx*cd + Bz*sd;
	Bned[1] = By;
	Bned[2] = Bz*cd - Bx*sd;

	// Convert to ECI from Fossen (2011) Handbook of Marine Craft Hydrodynamics and Motion Control
	double Ned2ecef[9];
	frameRotationNed2ecef(sinlat, coslat, sinphi[1], cosphi[1], Ned2ecef);
	double Becef[3];
	Becef[0] = Ned2ecef[0]*Bned[0] + Ned2ecef[3]*Bned[1] + Ned2ecef[6]*Bned[2];
	Becef[1] = Ned2ecef[1]*Bned[0] + Ned2ecef[4]*Bned[1] + Ned2ecef[7]*Bned[2];
	Becef[2] = Ned2ecef[2]*Bned[0] + Ned2ecef[5]*Bned[1] + Ned2ecef[8]*Bned[2];
	frameEcef2eci(Becef,timeUTC,Beci);
}

/*
 *Geomagnetic Vector Prediction
 * function magWmm
 *
 * Created: 07.02.2015 08:46:08
 * Author: Antoine Pignede
 *
 * Calculates the components of the Earth's magnetic field using the
 *   World Magnetic Model (WMM).
 *   Note that the other parameters the WMM gives can be computed from BX,
 *   BY, and BZ as:
 *     Horizonal intensity: hypot(BX, BY) (i.e., sqrt(BX.^2 + BY.^2) )
 *     Total intensity: hypot(BX, hypot(BY, BZ))
 *     Declination: atan2(BY, BX)
 *     Inclination: atan(BZ./hypot(BX, BY))
 *
 * Copied and edited from the official C source code available at http://www.ngdc.noaa.gov/geomag/WMM/soft.shtml
 *
 * Inputs:
 *    reci        - position vector eci            km
 *    timeUTC     - struct containing:
 *      year        - year                           1900 .. 2100
 *      mon         - month                          1 .. 12
 *      day         - day                            1 .. 28,29,30,31
 *      hr          - universal time hour            0 .. 23
 *      min         - universal time min             0 .. 59
 *      sec         - universal time sec             0.0 .. 59.999
 *
 * Outputs:
 *    Beci        - geomagnetic field vector in nanotesla (nT) in eci frame
 *
 * Coupling
 *    timeDatetime2years     - convert date and time values to fractional years
 *    frameEci2ecef     - transform eci position to ecef position
 *    frameEcef2geod    - transform ecef position to geodetic position
 *
 * Note that the original functions have been merged into this wmm function.
 *
 * See also
 *    magIgrf.m  - alternative geomagnetic vector predictor
 */

void magWmm(double reci[3], Time timeUTC, double Beci[3]){
/* Convert time to fractional years. */
	double time = timeDatetime2years(timeUTC);

	// Time change the Model coefficients from the base year of the model using secular variation coefficients
	double g[91]; // numel(WMMG)==91
	double h[91]; // numel(WMMH)==91
	int i;
	for (i=0; i<=90; i++){
		g[i] = WMMG[i] + WMMGSV[i]*(time - 2015.0);
		h[i] = WMMH[i] + WMMHSV[i]*(time - 2015.0);
	}
	int nMax = 12; // Highest order // reduce (to 7, 8, ...) to speed up
	//int NumTerms = (nMax + 1) * (nMax + 2) / 2; // ==91

/* Transform eci input vector to geodetic coordinates */
	double recef[3];
	frameEci2ecef(reci,timeUTC,recef);
	double rgeod[3];
	frameEcef2geod(recef,rgeod);
	double latitude = rgeod[0];
	double longitude = rgeod[1];
	double altitude = rgeod[2];

/* Convert Geodetic coordinates to Spherical coordinates */
	double coslat = cos(latitude*DEG2RAD);
	double sinlat = sin(latitude*DEG2RAD);
	// compute the local radius of curvature on the WGS-84 reference ellipsoid
	double rc = REQU / sqrt(1.0 - ECCSQRD * sinlat * sinlat);
	// compute ECEF Cartesian coordinates of specified point (for longitude=0). should actually not be necessary because they're in recef, tests fail however when removing these two lines
	double xp = (rc + altitude) * coslat;
	double zp = (rc * (1.0 - ECCSQRD) + altitude) * sinlat;
	// compute spherical radius and angle lambda and phi of specified point
	double r = sqrt(xp * xp + zp * zp);
	double phi = asin(zp/r) * RAD2DEG; // geocentric latitude
	double lambda = longitude; // longitude

/* For storing the Associated Legendre functions and the Spherical Harmonic variables */
	double Pcup[91+1]; // NumTerms==91
	double dPcup[91+1]; // NumTerms==91
	double RelativeRadiusPower[12+1]; // nMax==12
	double coslambda[12+1]; // nMax==12
	double sinlambda[12+1]; // nMax==12

/* Compute Spherical Harmonic variables */
	RelativeRadiusPower[0] = (RMEAN/r) * (RMEAN/r);
	for (i=1; i<=nMax; i++){
		RelativeRadiusPower[i] = RelativeRadiusPower[i-1] * (RMEAN/r); // (Radius of Earth / Spherical radius r)^(n+2)
	}
	coslambda[0] = 1.0; // not needed, but helps with indices: coslambda[i] = cos(i*lambda);
	sinlambda[0] = 0.0; // not needed, but helps with indices: sinlambda[i] = sin(i*lambda);
	coslambda[1] = cos(lambda*DEG2RAD);
	sinlambda[1] = sin(lambda*DEG2RAD);
	for (i=2; i<=nMax; i++){
		coslambda[i] = coslambda[i-1]*coslambda[1] - sinlambda[i-1]*sinlambda[1]; // cos(a+b) = cos(a)*cos(b) - sin(a)*sin(b)
		sinlambda[i] = coslambda[i-1]*sinlambda[1] + sinlambda[i-1]*coslambda[1]; // sin(a+b) = cos(a)*sin(b) + sin(a)*cos(b)
	}

/* Compute Associated Legendre functions */
	double cosphi = cos(phi*DEG2RAD);
	double sinphi = sin(phi*DEG2RAD);
	// Evaluate all of the Schmidt-semi normalized Associated Legendre functions up to degree nMax
	// as nMax<=12 MAG_PcupLow from the source code is used
	Pcup[0] = 1.0;
	dPcup[0] = 0.0;
	double z = sqrt((1.0-sinphi) * (1.0+sinphi));
	double schmidtQuasiNorm[91+1]; // NumTerms==91
	double k;
	int n;
	int m;
	int index;
	int index1;
	int index2;
	// First, Compute the Gauss-normalized Associated Legendre functions
	for (n=1; n<=nMax; n++){
		for (m=0; m<=n; m++){
			index = (n * (n + 1) / 2 + m);
			if (n == m){
				index1 = (n - 1) * n / 2 + m - 1;
				Pcup [index] = z * Pcup[index1];
				dPcup[index] = z * dPcup[index1] + sinphi * Pcup[index1];
			}
			else if (n == 1 && m == 0){
				index1 = (n - 1) * n / 2 + m;
				Pcup[index] = sinphi * Pcup[index1];
				dPcup[index] = sinphi * dPcup[index1] - z * Pcup[index1];
			}
			else if (n > 1 && n != m){
				index1 = (n - 2) * (n - 1) / 2 + m;
				index2 = (n - 1) * n / 2 + m;
				if (m > n - 2){
					Pcup[index] = sinphi * Pcup[index2];
					dPcup[index] = sinphi * dPcup[index2] - z * Pcup[index2];
				}
				else{
					k = (double) (((n - 1) * (n - 1)) - (m * m)) / (double) ((2 * n - 1) * (2 * n - 3));
					Pcup[index] = sinphi * Pcup[index2] - k * Pcup[index1];
					dPcup[index] = sinphi * dPcup[index2] - z * Pcup[index2] - k * dPcup[index1];
				}
			}
		}
	}
	// Compute the ratio between the the Schmidt quasi-normalized Associated Legendre
	// functions and the Gauss-normalized version
	schmidtQuasiNorm[0] = 1.0;
	for (n=1; n<=nMax; n++){
		index = (n * (n + 1) / 2);
		index1 = (n - 1) * n / 2;
		schmidtQuasiNorm[index] = schmidtQuasiNorm[index1] * (double) (2 * n - 1) / (double) n; // m==0
		for (m=1; m<=n; m++){
			index = (n * (n + 1) / 2 + m);
			index1 = (n * (n + 1) / 2 + m - 1);
			schmidtQuasiNorm[index] = schmidtQuasiNorm[index1] * sqrt((double) ((n - m + 1) * (m == 1 ? 2 : 1)) / (double) (n + m));
		}
	}
	// Converts the  Gauss-normalized Associated Legendre functions to the Schmidt quasi-normalized
	// version using pre-computed relation stored in the variable schmidtQuasiNorm
	for (n=1; n<=nMax; n++){
		for (m=0; m<=n; m++){
			index = (n * (n + 1) / 2 + m);
			Pcup[index] = Pcup[index] * schmidtQuasiNorm[index];
			dPcup[index] = -dPcup[index] * schmidtQuasiNorm[index]; // "-" Because latitude instead of colatitude is used
		}
	}

/* Accumulate the Spherical Harmonic coefficients */
	double Bz = 0.0;
	double By = 0.0;
	double Bx = 0.0;
	for (n=1; n<=nMax; n++){
		for(m=0; m<=n; m++){
			index = (n * (n + 1) / 2 + m);
			Bz -= RelativeRadiusPower[n] * (g[index]*coslambda[m] +	h[index]*sinlambda[m]) * (double) (n + 1)*Pcup[index];
			By += RelativeRadiusPower[n] * (g[index]*sinlambda[m] - h[index]*coslambda[m]) * (double) (m)*Pcup[index];
			Bx -= RelativeRadiusPower[n] * (g[index]*coslambda[m] +	h[index]*sinlambda[m]) * dPcup[index];
		}
	}

	if (fabs(cosphi) > 1.0e-10){
		By = By / cosphi;
	}
	else{ // Special calculation for By at Geographic poles
		double PcupS[12+1]; // nMax==12
		PcupS[0] = 1;
		double schmidtQuasiNorm1 = 1.0;
		double schmidtQuasiNorm2;
		double schmidtQuasiNorm3;
		By = 0.0;
		// Compute the ratio between the the Schmidt quasi-normalized Associated Legendre
		// functions and the Gauss-normalized version
		for (n=1; n<=nMax; n++){
			index = (n * (n + 1) / 2 + 1);
			schmidtQuasiNorm2 = schmidtQuasiNorm1 * (double) (2 * n - 1) / (double) n;
			schmidtQuasiNorm3 = schmidtQuasiNorm2 * sqrt((double) (n * 2) / (double) (n + 1));
			schmidtQuasiNorm1 = schmidtQuasiNorm2;
			if (n == 1){
				PcupS[n] = PcupS[n - 1];
			}
			else{
				k = (double) (((n - 1) * (n - 1)) - 1) / (double) ((2 * n - 1) * (2 * n - 3));
				PcupS[n] = sinphi * PcupS[n - 1] - k * PcupS[n - 2];
			}
			By += RelativeRadiusPower[n] * (g[index]*sinlambda[1] - h[index]*coslambda[1]) * PcupS[n]*schmidtQuasiNorm3;
		}
	}

/* Map the computed Magnetic fields to Geodetic coordinates */
	// Difference between the spherical and Geodetic latitudes
	double psi = DEG2RAD * (phi - latitude);
	double cospsi = cos(psi);
	double sinpsi = sin(psi);

	double Bned[3];
	Bned[0] = Bx*cospsi - Bz*sinpsi;
	Bned[1] = By;
	Bned[2] = Bx*sinpsi + Bz*cospsi;

/* Convert to ECI from Fossen (2011) Handbook of Marine Craft Hydrodynamics and Motion Control */
	double Ned2ecef[9];
	frameRotationNed2ecef(sinlat, coslat, sinlambda[1], coslambda[1], Ned2ecef);
	double Becef[3];
	Becef[0] = Ned2ecef[0]*Bned[0] + Ned2ecef[3]*Bned[1] + Ned2ecef[6]*Bned[2];
	Becef[1] = Ned2ecef[1]*Bned[0] + Ned2ecef[4]*Bned[1] + Ned2ecef[7]*Bned[2];
	Becef[2] = Ned2ecef[2]*Bned[0] + Ned2ecef[5]*Bned[1] + Ned2ecef[8]*Bned[2];
	frameEcef2eci(Becef,timeUTC,Beci);
}
