/*
 * orbitPred.c
 *
 *  Created: Feb 17, 2015 9:33:41 AM
 *  Author: Antoine Pignede
 *
 *  This file defines the Orbit Position And Velocity Prediction
 */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include "constants.h"
#include "timeConv.h"
#include "frameTrans.h"
#include "orbitPred.h"

/*
 *Orbit Position Prediction
 * function orbitTwoline2rv
 *
 * Created: 03.02.2015 11:21:25
 * Author: Antoine Pignede
 *
 * This function converts the two line element set character string data to
 *   variables and initializes the sgp4 variables. several intermediate varaibles
 *   and quantities are determined. simple checks of the input tle are done.
 *
 * Copied and edited from Vallado (2013) Fundamentals of Astrodynamics and Applications
 *
 * inputs        :
 *   tle         - TLE string in one line without any kind of space between the first and second line
 *
 * outputs       :
 *   satrec      - struct containing all the sgp4 satellite information
 *     The first satrec struct element is an error message. If everything is
 *     fine it will be empty, otherwise a short description of the error is
 *     given. In orbitTwoline2rv possible errors in the tle string are
 *     detected. In orbitSgp4 values outside the expected range are
 *     detected. However the calculations will always continue. The only
 *     error to stop a function is if the TLE input string in
 *     orbitTwoline2rv is less than 138 characters long.
 *
 * coupling      :
 *   timeDays2datetime    - conversion of days to month, day, hour, minute, second
 *   timeDatetime2jd - convert day month year hour minute second to julian date
 *   orbitSgp4init   - initialize the sgp4 variables
 *
 * Note that the original function takes the two-line element set (TLE) as
 *   two strings representing one line each. Here the input is just one
 *   string containing the TLE without any kind of separation between the
 *   two lines. As they always have 69 characters, the string of length 138
 *   is separated in the middle.
 * The original function uses a special extended TLE to also have the time
 *   points where values are to be calculated. Anything related to that was
 *   removed here. This function just converts the "normal" TLE to values
 *   for the satellite and calls orbitSgp4init to calculate more values.
 * This function is not meant to be used to track several satellites.
 *   Some easy checks for right TLE input were added.
 *
 * See also
 *   orbitSgp4.m  - calculate orbital position of satrec object
 */

Satrec orbitTwoline2rv(char tle[139]){
	Satrec satrec; // This is the only declaration of a Satrec struct, pointers to it are then used.
	sprintf(satrec.error,"");

    // check for right length of input string
	size_t lenTle = strlen(tle);
    if ((int)lenTle < 138){
        sprintf(satrec.error,"too short TLE, strlen(tle)==%d < 138",(int)lenTle);
        return satrec;
    }
/* process the two lines */
    // split input string into the two lines
    char longstr1[70];
    memcpy(longstr1, tle, 69);
    longstr1[69] = '\0';
    size_t len1 = strlen(longstr1);
    char longstr2[70];
    memcpy(longstr2, tle+69, 70);
    size_t len2 = strlen(longstr2);

    // set the implied decimal points since doing a formated read
    // fixes for bad input data values (missing, ...)
    if (longstr1[7] == ' ') longstr1[7] = 'U';
    if (longstr1[9] == ' ') longstr1[9] = '.';
    int i;
    for (i=10; i<=15; i++)
        if (longstr1[i] == ' ') longstr1[i] = '_';
    if (longstr1[44] != ' ') longstr1[43] = longstr1[44];
    longstr1[44] = '.';
    for (i=45; i<=49; i++)
        if (longstr1[i] == ' ') longstr1[i] = '0';
    if (longstr1[51] == ' ') longstr1[51] = '0';
    if (longstr1[53] != ' ') longstr1[52] = longstr1[53];
    longstr1[53] = '.';
    if (longstr1[62] == ' ') longstr1[62] = '0';
    if ((len1 < 68) || (longstr1[67] == ' ')) longstr1[67] = '0';

    longstr2[25] = '.';
    for (i=27; i<=32; i++)
        if (longstr2[i] == ' ') longstr2[i] = '0';

    // parse first line
    int linenum1;
    int nexp;
    int ibexp;
    sscanf(longstr1,"%2d %5ld %1c %10s %2d %12lf %11lf %7lf %2d %7lf %2d %2d %6ld ",
    		&linenum1,&satrec.satnum,&satrec.classification,&satrec.intldesg,&satrec.epochyr,&satrec.epochdays,
    		&satrec.ndot, &satrec.nddot, &nexp, &satrec.bstar,&ibexp,&satrec.ephemtype,&satrec.elnum);
    // remove checksum1 from elnum, (last digit in satrec.elnum)
    int checksum1 = satrec.elnum % 10;
    satrec.elnum = satrec.elnum / 10;

    // parse second line
    int linenum2;
    long int satnum2;
    if (longstr2[52] == ' ')
    	sscanf(longstr2,"%2d %5ld %9lf %9lf %8lf %9lf %9lf %10lf %6ld \n",
    			&linenum2,&satnum2,&satrec.inclo,&satrec.nodeo,&satrec.ecco,
    			&satrec.argpo,&satrec.mo,&satrec.no,&satrec.revnum);
	else
		sscanf(longstr2,"%2d %5ld %9lf %9lf %8lf %9lf %9lf %11lf %6ld \n",
				&linenum2,&satnum2,&satrec.inclo,&satrec.nodeo,&satrec.ecco,
				&satrec.argpo,&satrec.mo,&satrec.no,&satrec.revnum);
    // remove checksum2 from revnum, (last digit in satrec.revnum)
    int checksum2 = satrec.revnum % 10;
    satrec.revnum = satrec.revnum / 10;

/* six simple tle checks, failure does not stop the function */
    // check for line numbers
    if (linenum1 != 1) sprintf(satrec.error,"wrong first line number, linenum1==%d not 1",linenum1);
    if (linenum2 != 2) sprintf(satrec.error,"wrong second line number, linenum2==%d not 2",linenum2);
    // check for satellite number
    if (satrec.satnum != satnum2) sprintf(satrec.error,"two different satellite numbers, satnum1==%ld not satnum2==%ld",satrec.satnum,satnum2);
    // check for classification character
    if (satrec.classification != 'U') sprintf(satrec.error,"unknown classification, classification==%c not U",satrec.classification);
    // check the first line checksum
    char checksumChar;
    int checksum1Calc = 0;
    for (i=0; i<len1-1; i++){
    	memcpy(&checksumChar,longstr1+i,1);
        if (checksumChar == '-') checksum1Calc = checksum1Calc + 1;
        else{
        	int char2int = atoi(&checksumChar);
        	checksum1Calc = checksum1Calc + char2int;
        }
    }
    checksum1Calc = checksum1Calc % 10;
    if (checksum1Calc != checksum1) sprintf(satrec.error,"wrong first line checksum, calculated==%d not given==%d",checksum1Calc,checksum1);
    // check the second line checksum
    int checksum2Calc = 0;
    for (i=0; i<len2-1; i++){
		memcpy(&checksumChar,longstr2+i,1);
		if (checksumChar == '-') checksum2Calc = checksum2Calc + 1;
		else{
			int char2int = atoi(&checksumChar);
			checksum2Calc = checksum2Calc + char2int;
		}
	}
	checksum2Calc = checksum2Calc % 10;
	if (checksum2Calc != checksum2) sprintf(satrec.error,"wrong first second checksum, calculated==%d not given==%d",checksum2Calc,checksum2);

/* apply exponents to bases and perform unit changes */
    satrec.nddot= satrec.nddot * pow(10.0,nexp);
    satrec.bstar= satrec.bstar * pow(10.0,ibexp);

    // [rev/day]/[rad/min] = 229.1831180523293;
    double xpdotp   =  1440.0 / (TWOPI);

    satrec.no   = satrec.no / xpdotp; 					 // rad/min
    satrec.ndot = satrec.ndot  / (xpdotp*1440.0);        // rad/min^2
    satrec.nddot= satrec.nddot / (xpdotp*1440.0*1440.0); // rad/min^3

    satrec.inclo = satrec.inclo  * DEG2RAD;
    satrec.nodeo = satrec.nodeo  * DEG2RAD;
    satrec.argpo = satrec.argpo  * DEG2RAD;
    satrec.mo    = satrec.mo     * DEG2RAD;

/* temporary fix for years from 1957-2056,
 *   correct fix will occur when year is 4-digit in tle */
    int year;
    if (satrec.epochyr < 57) year= satrec.epochyr + 2000;
    else year= satrec.epochyr + 1900;
    Time time = timeDays2datetime(year,satrec.epochdays);
    satrec.jdsatepoch = timeDatetime2jd(time);

 /* initialize more satellite values and return the struct */
    orbitSgp4init(&satrec);
    return satrec;
}

/*
 *Orbit Position Prediction
 * function orbitSgp4init
 *
 * Created: 03.02.2015 12:02:50
 * Author: Antoine Pignede
 *
 * This procedure initializes variables for sgp4. It is automatically called
 *   each time a new tle has been read.
 *
 * Copied and edited from Vallado (2013) Fundamentals of Astrodynamics and Applications
 *
 *   inputs        :
 *     satrec      - struct containing all the sgp4 satellite information
 *
 *   outputs       :
 *     satrec      - struct containing all the sgp4 satellite information
 *
 *   locals        :
 *     Cc1sq  , Cc2    , Cc3
 *     Coef   , Coef1
 *     cosio4      -
 *     eeta        -
 *     etasq       -
 *     perige      - perigee
 *     pinvsq      -
 *     psisq       -
 *     qzms24      -
 *     sfour       -
 *     temp        -
 *     temp1, temp2, temp3       -
 *     tsi         -
 *     xhdot1      -
 *
 *   coupling      :
 *    none
 *
 * Note that the original function also has all calculations for deep space
 *   objects. Those were removed here because this program only will run on
 *   a satellite in low earth orbit.
 *   This function is not used directly but as subroutine of orbitTwoline2rv.m.
 *
 * See also
 *   orbitTwoline2rv.m  - convert TLE string to satrec values
 */

void orbitSgp4init(Satrec *satrec){
/* Initialization */
   // set all near earth variables to zero
   satrec->isimp   = 0;   satrec->aycof  = 0.0;
   satrec->con41   = 0.0; satrec->cc1    = 0.0; satrec->cc4      = 0.0;
   satrec->cc5     = 0.0; satrec->d2     = 0.0; satrec->d3       = 0.0;
   satrec->d4      = 0.0; satrec->delmo  = 0.0; satrec->eta      = 0.0;
   satrec->argpdot = 0.0; satrec->omgcof = 0.0; satrec->sinmao   = 0.0;
   satrec->t       = 0.0; satrec->t2cof  = 0.0; satrec->t3cof    = 0.0;
   satrec->t4cof   = 0.0; satrec->t5cof  = 0.0; satrec->x1mth2   = 0.0;
   satrec->x7thm1  = 0.0; satrec->mdot   = 0.0; satrec->nodedot = 0.0;
   satrec->xlcof   = 0.0; satrec->xmcof  = 0.0; satrec->nodecf  = 0.0;

   double ss     = 78.0 / REQU + 1.0;
   double qzms2t = pow((120.0 - 78.0) / REQU, 4.0);
   // sgp4fix divisor for divide by zero check on inclination
   // the old check used 1.0 + cos(pi-1.0e-9), but then compared it to
   // 1.5 e-12, so the threshold was changed to 1.5e-12 for consistancy
   double temp4    =   1.5e-12;

   satrec->t    = 0.0; // reserve place for time point

/*
 * Excerpt of initl.m from Vallado (2013) Fundamentals of Astrodynamics and Applications
 * This procedure initializes the spg4 propagator. all the initialization is
 *   consolidated here instead of having multiple loops inside other routines.
 */

   // calculate auxillary epoch quantities
   double eccsq  = satrec->ecco * satrec->ecco;
   double omeosq = 1.0 - eccsq;
   double rteosq = sqrt(omeosq);
   double cosio  = cos(satrec->inclo);
   double cosio2 = cosio * cosio;

   // un-kozai, the mean motion
   double ak    = pow(XKE / satrec->no, 2.0 / 3.0);
   double d1    = 0.75 * J2 * (3.0 * cosio2 - 1.0) / (rteosq * omeosq);
   double del   = d1 / (ak * ak);
   double adel  = ak * (1.0 - del * del - del *
		   (1.0 / 3.0 + 134.0 * del * del / 81.0));
   del   = d1/(adel * adel);
   satrec->no    = satrec->no / (1.0 + del);

   double ao    = pow(XKE / satrec->no, 2.0 / 3.0);
   double sinio = sin(satrec->inclo);
   double po    = ao * omeosq;
   double con42 = 1.0 - 5.0 * cosio2;
   satrec->con41 = -con42-cosio2-cosio2;
   double posq  = po * po;
   double rp    = ao * (1.0 - satrec->ecco);

/* Continue initialization for near space */
   if ((omeosq >= 0.0 ) || ( satrec->no >= 0.0)){
	   satrec->isimp = 0;
	   // For perigee < 220 km, set isimp=1, truncate some equations and drop some terms
	   if (rp < (220.0 / REQU + 1.0)) satrec->isimp = 1;
	   double sfour  = ss;
	   double qzms24 = qzms2t;
	   double perige = (rp - 1.0) * REQU;

	   // for perigees below 156 km, s and qoms2t are altered
	   if (perige < 156.0){
		   sfour = perige - 78.0;
		   if (perige < 98.0) sfour = 20.0;
		   qzms24 = pow((120.0 - sfour) / REQU, 4.0);
		   sfour  = sfour / REQU + 1.0;
	   }
	   double pinvsq = 1.0 / posq;

	   double tsi  = 1.0 / (ao - sfour);
	   satrec->eta  = ao * satrec->ecco * tsi;
	   double etasq = satrec->eta * satrec->eta;
	   double eeta  = satrec->ecco * satrec->eta;
	   double psisq = fabs(1.0 - etasq);
	   double coef  = qzms24 * pow(tsi, 4.0);
	   double coef1 = coef / pow(psisq, 3.5);
	   double cc2   = coef1 * satrec->no * (ao *
			   (1.0 + 1.5 * etasq + eeta * (4.0 + etasq)) +
			   0.375 * J2 * tsi / psisq * satrec->con41 *
			   (8.0 + 3.0 * etasq * (8.0 + etasq)));
	   satrec->cc1   = satrec->bstar * cc2;
	   double cc3   = 0.0;
	   if (satrec->ecco > 1.0e-4)
		   cc3 = -2.0 * coef * tsi * J3OJ2 * satrec->no * sinio / satrec->ecco;
	   satrec->x1mth2 = 1.0 - cosio2;
	   satrec->cc4    = 2.0* satrec->no * coef1 * ao * omeosq *
		   (satrec->eta * (2.0 + 0.5 * etasq) + satrec->ecco *
		   (0.5 + 2.0 * etasq) - J2 * tsi / (ao * psisq) *
		   (-3.0 * satrec->con41 * (1.0 - 2.0 * eeta + etasq *
		   (1.5 - 0.5 * eeta)) + 0.75 * satrec->x1mth2 *
		   (2.0 * etasq - eeta * (1.0 + etasq)) * cos(2.0 * satrec->argpo)));
	   satrec->cc5 = 2.0 * coef1 * ao * omeosq *
			   (1.0 + 2.75 * (etasq + eeta) + eeta * etasq);
	   double cosio4 = cosio2 * cosio2;
	   double temp1  = 1.5 * J2 * pinvsq * satrec->no;
	   double temp2  = 0.5 * temp1 * J2 * pinvsq;
	   double temp3  = -0.46875 * J4 * pinvsq * pinvsq * satrec->no;
	   satrec->mdot     = satrec->no + 0.5 * temp1 * rteosq * satrec->con41 +
			   0.0625 * temp2 * rteosq * (13.0 - 78.0 * cosio2 + 137.0 * cosio4);
	   satrec->argpdot  = -0.5 * temp1 * con42 + 0.0625 * temp2 *
			   (7.0 - 114.0 * cosio2 + 395.0 * cosio4) +
			   temp3 * (3.0 - 36.0 * cosio2 + 49.0 * cosio4);
	   double xhdot1   = -temp1 * cosio;
	   satrec->nodedot = xhdot1 + (0.5 * temp2 * (4.0 - 19.0 * cosio2) +
			   2.0 * temp3 * (3.0 - 7.0 * cosio2)) * cosio;
	   satrec->omgcof   = satrec->bstar * cc3 * cos(satrec->argpo);
	   satrec->xmcof    = 0.0;
	   if (satrec->ecco > 1.0e-4)
		   satrec->xmcof = -(2.0 / 3.0) * coef * satrec->bstar / eeta;
	   satrec->nodecf = 3.5 * omeosq * xhdot1 * satrec->cc1;
	   satrec->t2cof   = 1.5 * satrec->cc1;

	   // sgp4fix for divide by zero with xinco = 180 deg
	   if (fabs(cosio+1.0) > 1.5e-12)
		   satrec->xlcof = -0.25 * J3OJ2 * sinio *
		   (3.0 + 5.0 * cosio) / (1.0 + cosio);
	   else
		   satrec->xlcof = -0.25 * J3OJ2 * sinio *
		   (3.0 + 5.0 * cosio) / temp4;
	   satrec->aycof   = -0.5 * J3OJ2 * sinio;
	   double delmotemp = 1.0 + satrec->eta * cos(satrec->mo);
	   satrec->delmo = delmotemp * delmotemp * delmotemp;
	   satrec->sinmao  = sin(satrec->mo);
	   satrec->x7thm1  = 7.0 * cosio2 - 1.0;

	   // set additional variables if perigee >= 220 km
	   if (satrec->isimp != 1){
		   double cc1sq          = satrec->cc1 * satrec->cc1;
		   satrec->d2    = 4.0 * ao * tsi * cc1sq;
		   double temp           = satrec->d2 * tsi * satrec->cc1 / 3.0;
		   satrec->d3    = (17.0 * ao + sfour) * temp;
		   satrec->d4    = 0.5 * temp * ao * tsi *
				   (221.0 * ao + 31.0 * sfour) * satrec->cc1;
		   satrec->t3cof = satrec->d2 + 2.0 * cc1sq;
		   satrec->t4cof = 0.25 * (3.0 * satrec->d3 +
				   satrec->cc1 * (12.0 * satrec->d2 + 10.0 * cc1sq));
		   satrec->t5cof = 0.2 * (3.0 * satrec->d4 +
				   12.0 * satrec->cc1 * satrec->d3 +
				   6.0 * satrec->d2 * satrec->d2 +
				   15.0 * cc1sq * (2.0 * satrec->d2 + cc1sq));
	  }
   }
}

/*
 *Orbit Position Prediction
 * function orbitSgp4
 *
 * Created: 03.02.2015 15:16:27
 * Author: Antoine Pignede
 *
 * This procedure is the sgp4 prediction model from space command. this is an
 *   updated and combined version of sgp4 and sdp4, which were originally
 *   published separately in spacetrack report #3. this version follows the
 *   methodology from the aiaa paper (2006) describing the history and
 *   development of the code.
 *
 * Copied and edited from Vallado (2013) Fundamentals of Astrodynamics and Applications
 *
 *   inputs        :
 *    satrec      - initialised struct from sgp4init() call.
 *    time        - struct containing:
 *      year        - year                           1900 .. 2100
 *      mon         - month                          1 .. 12
 *      day         - day                            1 .. 28,29,30,31
 *      hr          - universal time hour            0 .. 23
 *      min         - universal time min             0 .. 59
 *      sec         - universal time sec             0.0 .. 59.999
 *
 *   outputs       :
 *     reci       - eci position vector                     km
 *     veci       - eci velocity vector                     km/s
 *
 *   locals        :
 *     am          -
 *     axnl, aynl        -
 *     betal       -
 *     COSIM   , SINIM   , COSOMM  , SINOMM  , Cnod    , Snod    , Cos2u   ,
 *     Sin2u   , Coseo1  , Sineo1  , Cosi    , Sini    , Cosip   , Sinip   ,
 *     Cosisq  , Cossu   , Sinsu   , Cosu    , Sinu
 *     Delm        -
 *     Delomg      -
 *     Ecose       -
 *     El2         -
 *     Eo1         -
 *     Esine       -
 *     Argpm       -
 *     Argpp       -
 *     Pl          -
 *     Rl          -
 *     Su          -
 *     T2  , T3   , T4    , Tc
 *     Tem5, Temp , Temp1 , Temp2  , Tempa  , Tempe  , Templ
 *     U   , Ux   , Uy    , Uz     , Vx     , Vy     , Vz
 *     inclm       - inclination
 *     mm          - mean anomaly
 *     nm          - mean motion
 *     nodem      - longi of ascending node
 *     xinc        -
 *     xincp       -
 *     xl          -
 *     xlm         -
 *     mp          -
 *     xmdf        -
 *     xmx         -
 *     xmy         -
 *     nodedf     -
 *     xnode       -
 *     nodep      -
 *
 *   coupling      :
 *     timeDatetime2jd   - convert date and time values to julian date
 *     frameTeme2eci     - convert position in teme from sgp4 to eci
 *
 * Note that the original function also has all calculations for deep space
 *   objects. Those were removed here because this program only will run on
 *   a satellite in low earth orbit.
 *   In contrast to the original function, the user has to provide the
 *   desired time and date for the position. The time since satellite epoch
 *   in minutes is calculated first in this function.
 *
 * See also
 *   orbitTwoline2rv.m, orbitSgp4init.m  - functions to initialize satrec
 */

void orbitSgp4(Satrec *satrec, Time timeUTC, double reci[3], double veci[3]){
   // Reset error string (if string not empty, the error was in the this run)
	sprintf(satrec->error,"");

   // time since epoch (minutes)
   double jd = timeDatetime2jd(timeUTC);
   satrec->t     = (jd - satrec->jdsatepoch) * 1440;

   double mrt = 0.0;

   // update for secular gravity and atmospheric drag
   double xmdf    = satrec->mo + satrec->mdot * satrec->t;
   double argpdf  = satrec->argpo + satrec->argpdot * satrec->t;
   double nodedf  = satrec->nodeo + satrec->nodedot * satrec->t;
   double argpm   = argpdf;
   double mm      = xmdf;
   double t2      = satrec->t * satrec->t;
   double nodem   = nodedf + satrec->nodecf * t2;
   double tempa   = 1.0 - satrec->cc1 * satrec->t;
   double tempe   = satrec->bstar * satrec->cc4 * satrec->t;
   double templ   = satrec->t2cof * t2;

   // if perigee > 220 km
   if (satrec->isimp != 1){
	   double delomg = satrec->omgcof * satrec->t;
	   double delm   = satrec->xmcof *
			   (pow(1.0 + satrec->eta * cos(xmdf), 3) - satrec->delmo);
	   double temp   = delomg + delm;
	   mm     = xmdf + temp;
	   argpm  = argpdf - temp;
	   double t3     = t2 * satrec->t;
	   double t4     = t3 * satrec->t;
	   tempa  = tempa - satrec->d2 * t2 - satrec->d3 * t3 - satrec->d4 * t4;
	   tempe  = tempe + satrec->bstar * satrec->cc5 * (sin(mm) - satrec->sinmao);
	   templ  = templ + satrec->t3cof * t3 +
			   t4 * (satrec->t4cof + satrec->t * satrec->t5cof);
   }

   double nm    = satrec->no;
   double em    = satrec->ecco;
   double inclm = satrec->inclo;

   // return 0 if mean motion less than 0
   if (nm <= 0.0){
	   sprintf(satrec->error,"mean motion==%f less than 0.0",nm);
	   reci[0] = 0.0;
	   reci[1] = 0.0;
	   reci[2] = 0.0;
	   veci[0] = 0.0;
	   veci[1] = 0.0;
	   veci[2] = 0.0;
	   return;
   }

   double am = pow(XKE / nm, (2.0 / 3.0)) * tempa * tempa;
   em = em - tempe;

   // fix tolerance for error recognition warning, return zeros
   if ((em >= 1.0) || (em < -0.001) || (am < 0.95)){
	   sprintf(satrec->error,"mean elements, ecc==%f >= 1.0 or ecc==%f < -0.001 or a==%f < 0.95",em,em,am);
	   reci[0] = 0.0;
	   reci[1] = 0.0;
	   reci[2] = 0.0;
	   veci[0] = 0.0;
	   veci[1] = 0.0;
	   veci[2] = 0.0;
	   return;
   }

   // sgp4fix change test condition for eccentricity
   if (em < 1.0e-6) em  = 1.0e-6;
   mm     = mm + satrec->no * templ;
   double xlm    = mm + argpm + nodem;
   nodem  = fmod(nodem, TWOPI);
   argpm  = fmod(argpm, TWOPI);
   xlm    = fmod(xlm, TWOPI);
   mm     = fmod(xlm - argpm - nodem, TWOPI);

   // compute extra mean quantities
   double sinim = sin(inclm);
   double cosim = cos(inclm);

   // add lunar-solar periodics
   double ep     = em;
   double xincp  = inclm;
   double argpp  = argpm;
   double nodep  = nodem;
   double mp     = mm;
   double sinip  = sinim;
   double cosip  = cosim;

   // long period periodics
   double axnl = ep * cos(argpp);
   double temp = 1.0 / (am * (1.0 - ep * ep));
   double aynl = ep* sin(argpp) + temp * satrec->aycof;
   double xl   = mp + argpp + nodep + temp * satrec->xlcof * axnl;

   // solve kepler's equation
   double u    = fmod(xl - nodep, TWOPI);
   double eo1  = u;
   double tem5 = 9999.9;
   int ktr = 1;
   // sgp4fix for kepler iteration
   // the following iteration needs better limits on corrections
   double sineo1;
   double coseo1;
   while ((fabs(tem5) >= 1.0e-12) && (ktr <= 10)){
	   sineo1 = sin(eo1);
	   coseo1 = cos(eo1);
	   tem5   = 1.0 - coseo1 * axnl - sineo1 * aynl;
	   tem5   = (u - aynl * coseo1 + axnl * sineo1 - eo1) / tem5;
	   if(fabs(tem5) >= 0.95){
		   if (tem5 > 0.0) tem5 = 0.95;
		   else tem5 = -0.95;
	   }
	   eo1 = eo1 + tem5;
	   ktr = ktr + 1;
   }

   // short period preliminary quantities
   double ecose = axnl*coseo1 + aynl*sineo1;
   double esine = axnl*sineo1 - aynl*coseo1;
   double el2   = axnl*axnl + aynl*aynl;
   double pl    = am*(1.0-el2);

   // no meaningful output if semi-latus rectum < 0, return zeros
   if (pl < 0.0){
	   sprintf(satrec->error,"semi-latus rectum==%f < 0.0",pl);
	   reci[0] = 0.0;
	   reci[1] = 0.0;
	   reci[2] = 0.0;
	   veci[0] = 0.0;
	   veci[1] = 0.0;
	   veci[2] = 0.0;
	   return;
   }
   else{
	   double rl     = am * (1.0 - ecose);
	   double rdotl  = sqrt(am) * esine/rl;
	   double rvdotl = sqrt(pl) / rl;
	   double betal  = sqrt(1.0 - el2);
	   temp   = esine / (1.0 + betal);
	   double sinu   = am / rl * (sineo1 - aynl - axnl * temp);
	   double cosu   = am / rl * (coseo1 - axnl + aynl * temp);
	   double su     = atan2(sinu, cosu);
	   double sin2u  = (cosu + cosu) * sinu;
	   double cos2u  = 1.0 - 2.0 * sinu * sinu;
	   temp   = 1.0 / pl;
	   double temp1  = 0.5 * J2 * temp;
	   double temp2  = temp1 * temp;

	   // update for short period periodics
	   mrt   = rl * (1.0 - 1.5 * temp2 * betal * satrec->con41) + 0.5 * temp1 * satrec->x1mth2 * cos2u;
	   su    = su - 0.25 * temp2 * satrec->x7thm1 * sin2u;
	   double xnode = nodep + 1.5 * temp2 * cosip * sin2u;
	   double xinc  = xincp + 1.5 * temp2 * cosip * sinip * cos2u;
	   double mvt   = rdotl - nm * temp1 * satrec->x1mth2 * sin2u / XKE;
	   double rvdot = rvdotl + nm * temp1 * (satrec->x1mth2 * cos2u + 1.5 * satrec->con41) / XKE;

	   // orientation vectors
	   double sinsu =  sin(su);
	   double cossu =  cos(su);
	   double snod  =  sin(xnode);
	   double cnod  =  cos(xnode);
	   double sini  =  sin(xinc);
	   double cosi  =  cos(xinc);
	   double xmx   = -snod * cosi;
	   double xmy   =  cnod * cosi;
	   double ux    =  xmx * sinsu + cnod * cossu;
	   double uy    =  xmy * sinsu + snod * cossu;
	   double uz    =  sini * sinsu;
	   double vx    =  xmx * cossu - cnod * sinsu;
	   double vy    =  xmy * cossu - snod * sinsu;
	   double vz    =  sini * cossu;

	   // position in km, velocity in km/s
	   double rteme[3];
	   rteme[0] = (mrt * ux)* REQU;
	   rteme[1] = (mrt * uy)* REQU;
	   rteme[2] = (mrt * uz)* REQU;
	   frameTeme2eci(rteme, timeUTC, reci);
	   double vteme[3];
	   vteme[0] = (mvt * ux + rvdot * vx) * VELKMPS;
	   vteme[1] = (mvt * uy + rvdot * vy) * VELKMPS;
	   vteme[2] = (mvt * uz + rvdot * vz) * VELKMPS;
	   frameTeme2eci(vteme, timeUTC, veci);
   }

   // warning if satellite has decayed, program continues (but is finished anyway)
   if (mrt < 1.0)
	   sprintf(satrec->error,"satellite has decayed, mrt==%f < 1.0",mrt);
}
