/*
 * sunPred.c
 *
 *  Created: Feb 16, 2015 7:08:35 PM
 *  Author: Antoine Pignede
 *
 *  This file defines the Sun Vector Predictor
 */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "constants.h"
#include "timeConv.h"
#include "frameTrans.h"
#include "sunPred.h"

/*
 *Sun Vector Prediction
 * function sun
 *
 * Created: 31.01.2015 15:33:09
 * Author: Antoine Pignede
 *
 * This function calculates the normed eci mean equator mean equinox (j2000) position vector
 *   of the sun given the julian date.  this is the low precision formula and
 *   is valid for years from 1950 to 2050.  accuaracy of apparent coordinates
 *   is 0.01  degrees.  notice many of the calculations are performed in
 *   degrees, and are not changed until later.  this is due to the fact that
 *   the almanac uses degrees exclusively in their formulations.
 *
 * Copied and edited from Vallado (2013) Fundamentals of Astrodynamics and Applications
 *
 *  inputs          description                    range / units
 *    timeUTC     - struct containing:
 *      year        - year                           1900 .. 2100
 *      mon         - month                          1 .. 12
 *      day         - day                            1 .. 28,29,30,31
 *      hr          - universal time hour            0 .. 23
 *      min         - universal time min             0 .. 59
 *      sec         - universal time sec             0.0 .. 59.999
 *
 *  outputs       :
 *    rsun        - eci position vector of the sun normed to magnitude one
 *
 *  locals        :
 *    meanlong    - mean longitude
 *    meananomaly - mean anomaly
 *    eclplong    - ecliptic longitude
 *    obliquity   - mean obliquity of the ecliptic
 *    tut1        - julian centuries of ut1 from
 *                  jan 1, 2000 12h
 *    ttdb        - julian centuries of tdb from
 *                  jan 1, 2000 12h
 *    all UTC times are converted to terrestrial time (TT)
 *
 *  coupling      :
 *   timeDatetime2jd   - convert date and time values to julian date
 *   timeUtc2tt        - convert UTC to TT
 *   timeJd2jc         - convert julian date to julian centuries
 *   framePrecess      - rotation for precession
 *
 * Note that the original function uses ut1 instead of utc to get the gmst.
 *   For simplicity ut1 is approximated by utc here. This is legitimiate
 *   because abs(ut1 - utc) < 0.9s which is negligible.
 * The same holds also for the barycentric dynamical time tdb which is
 *   approximated with terrestrial time tt. http://en.wikipedia.org/wiki/Barycentric_Dynamical_Time
 *   says that abs(tdb - tt) < 0.002s which is even more negligible.
 *   Note that the original function approximates tdb with ut1 which results
 *   in larger error: tt - ut = 69s in January 2015.
 *
 * See also
 */

void sun(Time timeUTC, double rsun[3]){
	// find the julian centuries of universal and terrestrial time at given UTC
	double jdUTC = timeDatetime2jd(timeUTC);
	double tut1= timeJd2jc(jdUTC);
	Time timeTT = timeUtc2tt(timeUTC);
	double jdTT = timeDatetime2jd(timeTT);
	double ttt = timeJd2jc(jdTT);

    // find mean longitude, mean anomaly, ecliptic longitude and mean obliquity
    double meanlong = 280.460 + 36000.77*tut1;
    meanlong = fmod(meanlong, 360.0);  //deg

    double ttdb= ttt;
    double meananomaly = 357.5277233 + 35999.05034*ttdb;
    meananomaly = fmod(meananomaly*DEG2RAD, TWOPI);  //rad
    if (meananomaly < 0.0) meananomaly= TWOPI + meananomaly;

    double eclplong = meanlong + 1.914666471*sin(meananomaly) +
    		0.019994643*sin(2.0*meananomaly); //deg
    eclplong = fmod(eclplong, 360.0);  //deg
    eclplong = eclplong * DEG2RAD;

    double obliquity = 23.439291 - 0.0130042*ttdb;  //deg
    obliquity= obliquity * DEG2RAD;

    // find magnitude of sun vector
    double magr= 1.000140612 - 0.016708617*cos(meananomaly) -
    		0.000139589*cos(2.0 *meananomaly);  //au

    // sun position in MOD
    double rsunMOD[3];
    rsunMOD[0] = magr*cos(eclplong);
    rsunMOD[1] = magr*cos(obliquity)*sin(eclplong);
    rsunMOD[2] = magr*sin(obliquity)*sin(eclplong);

    // transform to eci
    double prec[9];
    framePrecess(ttt, prec);
    rsun[0] = prec[0]*rsunMOD[0] + prec[3]*rsunMOD[1] + prec[6]*rsunMOD[2];
    rsun[1] = prec[1]*rsunMOD[0] + prec[4]*rsunMOD[1] + prec[7]*rsunMOD[2];
    rsun[2] = prec[2]*rsunMOD[0] + prec[5]*rsunMOD[1] + prec[8]*rsunMOD[2];

    // norm output vector
    double normRsun = hypot(hypot(rsun[0],rsun[1]), rsun[2]);
    rsun[0] = rsun[0]/normRsun;
    rsun[1] = rsun[1]/normRsun;
    rsun[2] = rsun[2]/normRsun;
}
