/*
 * timeTrans.c
 *
 *  Created: Feb 11, 2015 3:17:38 PM
 *  Author: Antoine Pignede
 *
 *  This file defines the Time Conversions
 */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "constants.h"
#include "timeConv.h"

/*
 *Time Conversions
 * function timeDatetime2jd
 *
 * Created: 30.01.2015 14:16:29
 * Author: Antoine Pignede
 *
 * This function finds the julian date given the year, month, day, and time.
 *
 * Copied and edited from Vallado (2013) Fundamentals of Astrodynamics and Applications
 *
 *  inputs          description                    range / units
 *    time        - struct containing:
 *      year        - year                           1900 .. 2100
 *      mon         - month                          1 .. 12
 *      day         - day                            1 .. 28,29,30,31
 *      hr          - universal time hour            0 .. 23
 *      min         - universal time min             0 .. 59
 *      sec         - universal time sec             0.0 .. 59.999
 *
 *  outputs       :
 *    jd          - julian date                    days from 4713 bc
 *
 *  locals        :
 *    none.
 *
 *  coupling      :
 *    none.
 *
 * See also
 *   timeJd2datetime.m  - inverse function
 */

double timeDatetime2jd(Time time){
	return (367.0 * time.year -
			floor((7 * (time.year + floor((time.mon + 9) / 12.0))) * 0.25) +
			floor(275 * time.mon / 9.0) + time.day + 1721013.5 +
			((time.sec/60.0 + time.min) / 60.0 + time.hr) / 24.0);
	//  - 0.5 * sign(100.0 * time.yr + time.mon - 190002.5) + 0.5;
}

/*
 *Time Conversions
 * function timeJd2datetime
 *
 * Created: 30.01.2015 14:23:30
 * Author: Antoine Pignede
 *
 * This function finds the year, month, day, hour, minute and second
 *   given the julian date.
 *
 * Copied and edited from Vallado (2013) Fundamentals of Astrodynamics and Applications
 *
 *  inputs          description                    range / units
 *    jd          - julian date                    days from 4713 bc
 *
 *  outputs       :
 *    time        - struct containing:
 *      year        - year                           1900 .. 2100
 *      mon         - month                          1 .. 12
 *      day         - day                            1 .. 28,29,30,31
 *      hr          - universal time hour            0 .. 23
 *      min         - universal time min             0 .. 59
 *      sec         - universal time sec             0.0 .. 59.999
 *
 *  locals        :
 *    days        - day of year plus fractional
 *                  portion of a day               days
 *    tu          - julian centuries from 0 h
 *                  jan 0, 1900
 *    temp        - temporary real values
 *    leapyrs     - number of leap years from 1900
 *
 *  coupling      :
 *    timeDays2datetime  - finds month, day, hour, minute and second given days and year
 *
 * See also
 *   timeDatetime2jd.m  - inverse function
 */

Time timeJd2datetime(double jd){
	// find year and days of the year
	double temp   = jd-2415019.5; // remove 01/01/1900 in JD
	double tu     = temp / 365.25;
	Time time;
	time.year   = 1900 + (int)floor(tu);
	int16_t leapyrs= (int)floor((time.year-1901) * 0.25);
	// nudge by 8.64x10-7 sec to get even outputs
	double days   = temp - ((time.year-1900)*365.0 + leapyrs) + 0.00000000001;

	// check for case of beginning of a year
	if (days < 1.0){
		time.year   = time.year - 1;
		leapyrs= (int)floor((time.year-1901) * 0.25);
		days   = temp - ((time.year-1900)*365.0 + leapyrs);
	}

	// find remaining data
	time = timeDays2datetime(time.year, days);
	time.sec= time.sec - 0.00000086400;
	return time;
}

/*
 *Time Conversions
 * function timeDatetime2days
 *
 * Created: 30.01.2015 14:39:31
 * Author: Antoine Pignede
 *
 * This function finds the fractional days through a year given the year,
 *   month, day, hour, minute and second.
 *
 * Copied and edited from Vallado (2013) Fundamentals of Astrodynamics and Applications
 *
 *  inputs          description                    range / units
 *    time        - struct containing:
 *      year        - year                           1900 .. 2100
 *      mon         - month                          1 .. 12
 *      day         - day                            1 .. 28,29,30,31
 *      hr          - universal time hour            0 .. 23
 *      min         - universal time min             0 .. 59
 *      sec         - universal time sec             0.0 .. 59.999
 *
 *  outputs       :
 *    days        - day of year plus fraction of a
 *                    day                          days
 *
 *  locals        :
 *    lmonth      - length of months of year
 *    i           - index
 *
 *  coupling      :
 *    none.
 *
 * See also
 *   timeDays2datetime.m  - inverse function
 */

double timeDatetime2days(Time time){
	// set up array of days in month with index as month number
	int lmonth[13] = {0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};
	// adjust February in leap years
	if ((time.year % 4) == 0){
		if (((time.year % 100) == 0) && ((time.year % 400) != 0)) lmonth[2]= 28;
		else lmonth[2]= 29;
	}

	// find day of the year
	int i   = 1;
	double days= 0.0;
	while ((i < time.mon) && ( i <= 12 )){
		days= days + lmonth[i];
		i= i + 1;
	}

	// add fractional day
	days= days + time.day + time.hr/24.0 + time.min/1440.0 + time.sec/86400.0;
	return days;
}

/*
 *Time Conversions
 * function timeDays2datetime
 *
 * Created: 30.01.2015 14:30:56
 * Author: Antoine Pignede
 *
 * This function converts the day of the year, days, to the equivalent month
 *   day, hour, minute and second.
 *
 * Copied and edited from Vallado (2013) Fundamentals of Astrodynamics and Applications
 *
 *  inputs          description                    range / units
 *    year        - year                           1900 .. 2100
 *    days        - day of year plus fraction of a
 *                    day                          days
 *
 *  outputs       :
 *    time        - struct containing:
 *      year        - year                           1900 .. 2100
 *      mon         - month                          1 .. 12
 *      day         - day                            1 .. 28,29,30,31
 *      hr          - universal time hour            0 .. 23
 *      min         - universal time min             0 .. 59
 *      sec         - universal time sec             0.0 .. 59.999
 *
 *  locals        :
 *    dayofyr     - day of year
 *    temp        - temporary extended values
 *    inttemp     - temporary integer value
 *    i           - index
 *    lmonth(12)  - integer array containing the number of days per month
 *
 *  coupling      :
 *    none.
 *
 * See also
 *   timeDatetime2days.m  - inverse function
 */

Time timeDays2datetime(int16_t year, double days){
	// set up array of days in month with index as month number
    int lmonth[13] = {0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};
    // adjust February in leap years
    if ((year % 4) == 0){
		if (((year % 100) == 0) && ((year % 400) != 0)) lmonth[2]= 28;
		else lmonth[2]= 29;
	}

    Time time;
    time.year = year;

    // find day of year, month and day of month
    int dayofyr = floor(days);
    int i = 1;
    int inttemp = 0;
    while ((dayofyr > inttemp + lmonth[i]) && (i < 12)){
      inttemp = inttemp + lmonth[i];
      i++;
    }
    time.mon = i;
    time.day = dayofyr - inttemp;

    // find hours minutes and seconds
    double temp = (days - dayofyr) * 24.0;
    time.hr   = floor(temp);
    temp = (temp - time.hr) * 60.0;
    time.min  = floor(temp);
    time.sec  = (temp - time.min) * 60.0;
	return time;
}

/*
 *Time Conversions
 * function timeDatetime2years
 *
 * Created: 30.01.2015 14:30:56
 * Author: Antoine Pignede
 *
 * This function converts the fractional year, to the equivalent month
 *   day, hour, minute and second.
 *
 * Inspired by Compston (2011) Matlab File Exchange International Geomagnetic Reference Field (IGRF) Model
 *
 *  inputs          description                    range / units
 *    time        - struct containing:
 *      year        - year                           1900 .. 2100
 *      mon         - month                          1 .. 12
 *      day         - day                            1 .. 28,29,30,31
 *      hr          - universal time hour            0 .. 23
 *      min         - universal time min             0 .. 59
 *      sec         - universal time sec             0.0 .. 59.999
 *
 *  outputs       :
 *    years       - fractional year                  1900.0 .. 2100.999
 *
 *  coupling      :
 *    timeDatetime2days  - find day of the year plus fraction
 *
 * See also
 *   timeYears2datetime.m  - inverse function
 */

double timeDatetime2years(Time time){
	// check if leap year or not
	int leap = 0;
	if ((time.year % 4) == 0){
		if (((time.year % 100) == 0) && ((time.year % 400) != 0)) leap = 0;
		else leap = 1;
	}
	
	// find day of year plus fraction
	double days = timeDatetime2days(time);
	// convert to years plus fraction
	double years = time.year + (days - 1)/(365.0 + (double)leap); // "(days - 1)" because days begins with 1
	return years;
}

/*
 *Time Conversions
 * function timeYears2datetime
 *
 * Created: 30.01.2015 14:30:56
 * Author: Antoine Pignede
 *
 * This function converts the fractional year, to the equivalent month
 *   day, hour, minute and second.
 *
 * Inspired by Compston (2011) Matlab File Exchange International Geomagnetic Reference Field (IGRF) Model
 *
 *  inputs          description                    range / units
 *    years       - fractional year                  1900.0 .. 2100.999
 *
 *  outputs       :
 *    time        - struct containing:
 *      year        - year                           1900 .. 2100
 *      mon         - month                          1 .. 12
 *      day         - day                            1 .. 28,29,30,31
 *      hr          - universal time hour            0 .. 23
 *      min         - universal time min             0 .. 59
 *      sec         - universal time sec             0.0 .. 59.999
 *
 *  locals        :
 *    dayofyr     - day of year
 *    temp        - temporary extended values
 *    inttemp     - temporary integer value
 *    i           - index
 *    lmonth(12)  - integer array containing the number of days per month
 *
 *  coupling      :
 *    timeDays2datetime     - convert fractional days to mon, day, hr, min, sec
 *
 * See also
 *   timeDatetime2years.m  - inverse function
 */

Time timeYears2datetime(double years){
	Time time;
	time.year = floor(years);

	// check if leap year or not
	int leap = 0;
	if ((time.year % 4) == 0){
		if (((time.year % 100) == 0) && ((time.year % 400) != 0)) leap = 0;
		else leap = 1;
	}

	// Convert fractional years to fractional day of the year
	double days = 1 + (years - time.year) * (365 + leap);
	// get mon, day, hr, min, sec
	time = timeDays2datetime(time.year,days);
	return time;
}

/*
 *Time Conversions
 * function timeJc2jd
 *
 * Created: 30.01.2015 14:16:29
 * Author: Antoine Pignede
 *
 * This function finds the julian day given the julian century after the
 *   algorithm given in Vallado (2013) Fundamentals of Astrodynamics and Applications.
 *
 * Inputs
 *   Julian century of the modified julian date (jd - 2451545.0  )/ 36525;
 *
 * Outputs
 *   Julian date (days from 4713 bc)
 *
 * See also
 *    timeJd2jc.m  - inverse function
 */

double timeJc2jd(double jc){
	return jc * 36525.0 + 2451545.0;
}

/*
 *Time Conversions
 * function timeJd2jc
 *
 * Created: 30.01.2015 14:16:29
 * Author: Antoine Pignede
 *
 * This function finds the julian century given the julian date after the
 *   algorithm given in Vallado (2013) Fundamentals of Astrodynamics and Applications.
 *
 * Inputs
 *   Julian date (days from 4713 bc)
 *
 * Outputs
 *   Julian century of the modified julian date (jd - 2451545.0  )/ 36525;
 *
 * See also
 *    timeJc2jd.m  - inverse function
 */

double timeJd2jc(double jd){
	return (jd - 2451545.0) / 36525.0;
}

/*
 *Time Conversions
 * function timeUtc2tt
 *
 * Created: 30.01.2015 17:14:19
 * Author: Antoine Pignede
 *
 * This function converts universal time to terrerstrial time calculating
 *   deltaT (deltaT = TT-UT) according to http://eclipse.gsfc.nasa.gov/SEcat5/deltatpoly.html
 *   assuming the year is between 2005 and 2050.
 *
 * Inputs
 *    timeUTC     - struct containing:
 *      year        - year                           1900 .. 2100
 *      mon         - month                          1 .. 12
 *      day         - day                            1 .. 28,29,30,31
 *      hr          - universal time hour            0 .. 23
 *      min         - universal time min             0 .. 59
 *      sec         - universal time sec             0.0 .. 59.999
 *
 * Outputs
 *    timeTT      - struct containing:
 *      year        - year                           1900 .. 2100
 *      mon         - month                          1 .. 12
 *      day         - day                            1 .. 28,29,30,31
 *      hr          - terrerstrial time hour         0 .. 23
 *      min         - terrerstrial time min          0 .. 59
 *      sec         - terrerstrial time sec          0.0 .. 59.999
 *
 *  coupling      :
 *    timeDays2datetime  - finds month, day, hour, minute and second given days and year
 *    timeDatetime2days  - finds days and year given month, day, hour, minute and second
 *
 * See also
 *   timeTt2utc.m  - inverse function
 */

Time timeUtc2tt(Time timeUTC){
/* http://eclipse.gsfc.nasa.gov/SEcat5/deltatpoly.html algorithm */
	double y = timeUTC.year + (timeUTC.mon - 0.5)/12.0;
	double deltaT = 62.92 + 0.32217 * (y-2000.0) + 0.005589 * (y-2000.0)*(y-2000.0);

	// Days of this year in UTC
	double daysUTC = timeDatetime2days(timeUTC);

	// Days of this year in TT
	double daysTT = daysUTC + deltaT/86400.0;

/* Convert back and adjust if TT is already in the next year */
	Time timeTT;
	if (daysTT < 366){
		// Same year no special case
		int16_t yearTT = timeUTC.year;
		timeTT = timeDays2datetime(yearTT,daysTT);
	}
	else {
		// Check if leap year
		int leap = 0;
		if ((timeUTC.year % 4) == 0){
				if (((timeUTC.year % 100) == 0) && ((timeUTC.year % 400) != 0)) leap = 0;
				else leap = 1;
		}

		if (leap == 0){
			// TT is already in the next year
			int16_t yearTT = timeUTC.year+1;
			timeTT = timeDays2datetime(yearTT,daysTT-365);
		}
		else {
			if (daysTT < 367){
				// Same year but it's a leap year
				int16_t yearTT = timeUTC.year;
				timeTT = timeDays2datetime(yearTT,daysTT);
			}
			else{
				// TT is already in the next year, the past year was a leap year
				int16_t yearTT = timeUTC.year+1;
				timeTT = timeDays2datetime(yearTT,daysTT-366);
			}
		}
	}
	return timeTT;
}

/*
 *Time Conversions
 * function timeTt2utc
 *
 * Created: 30.01.2015 17:14:19
 * Author: Antoine Pignede
 *
 * This function converts universal time to terrerstrial time calculating
 *   deltaT (deltaT = TT-UT) according to http://eclipse.gsfc.nasa.gov/SEcat5/deltatpoly.html
 *   assuming the year is between 2005 and 2050.
 *
 * Inputs
 *    timeTT      - struct containing:
 *      year        - year                           1900 .. 2100
 *      mon         - month                          1 .. 12
 *      day         - day                            1 .. 28,29,30,31
 *      hr          - terrerstrial time hour         0 .. 23
 *      min         - terrerstrial time min          0 .. 59
 *      sec         - terrerstrial time sec          0.0 .. 59.999
 *
 * Outputs
 *    timeUTC     - struct containing:
 *      year        - year                           1900 .. 2100
 *      mon         - month                          1 .. 12
 *      day         - day                            1 .. 28,29,30,31
 *      hr          - universal time hour            0 .. 23
 *      min         - universal time min             0 .. 59
 *      sec         - universal time sec             0.0 .. 59.999
 *
 *  coupling      :
 *    timeDays2datetime  - finds month, day, hour, minute and second given days and year
 *    timeDatetime2days  - finds days and year given month, day, hour, minute and second
 *
 * See also
 *   timeUtc2tt.m  - inverse function
 */

Time timeTt2utc(Time timeTT){
/* http://eclipse.gsfc.nasa.gov/SEcat5/deltatpoly.html algorithm */
	double y = timeTT.year + (timeTT.mon - 0.5)/12.0;
	double deltaT = 62.92 + 0.32217 * (y-2000.0) + 0.005589 * (y-2000.0)*(y-2000.0);
	// Days of this year in TT
	double daysTT = timeDatetime2days(timeTT);

	// Days of this year in UTC
	double daysUTC = daysTT - deltaT/86400.0;

/* Convert back and adjust if UTC is still in the past year */
	Time timeUTC;
	if (daysUTC >= 1){
		// Same year no special case
		int16_t yearUTC = timeTT.year;
		timeUTC = timeDays2datetime(yearUTC,daysUTC);
	}
	else{
		// Check if leap year
		int leap = 0;
		if ((timeUTC.year % 4) == 0){
				if (((timeUTC.year % 100) == 0) && ((timeUTC.year % 400) != 0)) leap = 0;
				else leap = 1;
		}

		if (leap == 0){
			// TT is already in the next year
			int16_t yearUTC = timeTT.year-1;
			timeUTC = timeDays2datetime(yearUTC,daysUTC+365);
		}
		else{
			// TT is already in the next year, the past year was a leap year
			int16_t yearUTC = timeTT.year-1;
			timeUTC = timeDays2datetime(yearUTC,daysUTC+366);
		}
	}
	return timeUTC;
}

/*
 *Time Conversions
 * function timeGstime
 *
 * Created: 31.01.2015 11:26:48
 * Author: Antoine Pignede
 *
 * This function finds the greenwich sidereal time (iau-82).
 *
 * Copied and edited from Vallado (2013) Fundamentals of Astrodynamics and Applications
 *
 *  inputs          description                    range / units
 *    jdut1       - julian date of ut1             days from 4713 bc
 *
 *  outputs       :
 *    gst         - greenwich sidereal time        0 to 2pi rad
 *
 *  locals        :
 *    temp        - temporary variable for reals   rad
 *    tut1        - julian centuries from the
 *                  jan 1, 2000 12 h epoch (ut1)
 *
 *  coupling      :
 *    timeJd2jc   - convert julian date to julian centuries
 *
 * See also
 */

double timeGstime(double jdut1){
	// julian centuries
	double tut1= timeJd2jc(jdut1);

	double gst = - 6.2e-6 * tut1 * tut1 * tut1 + 0.093104 * tut1 * tut1 +
			(876600.0 * 3600.0 + 8640184.812866) * tut1 + 67310.54841;

	// 360/86400 = 1/240, to deg, to rad
	gst = fmod((gst*DEG2RAD/240.0), TWOPI);

	// check quadrants
	if ( gst < 0.0 ) {
		gst = gst + TWOPI;
	}
	return gst;
}
