/*
 * nrp.c
 *
 * Created: 14-Feb-15 2:37:25 PM
 *  Author: Erlend
 */ 

#include "nrp_header.h"
#include "nrp.h"
#include "nrp_segment.h"
#include "nrp_csp_if.h"
#include "nrp_sm.h"
#include "nrp_conn.h"
#include "nrp_debug.h"
#include "nrp_ack.h"
#include "nrp_error.h"
#include "nrp_tester.h"
#include "crc.h"

#include "inttypes.h"

#include <stdint.h>

#include <stdio.h>
#include <string.h>
#include <asf.h>
#include "FreeRTOS.h"
#include "task.h"
#include "queue.h"

#include "semphr.h"//delete


//init variables
static Bool csp_initialized = false;
static Bool listener_initialized = false;

static xTaskHandle listen_handle = NULL;
static xQueueHandle port_queues[MAX_PORTS] = {NULL};

static Bool bound_ports[MAX_PORTS] = {false};

int8_t nrp_init(uint8_t module_number){
	/*Initialize CSP*/
	if (!csp_initialized){
		this_module = module_number;
		if (csp_usage_init(CSP_NR_OF_BUFFERED_PACKETS, nuts_modules[this_module].csp_address,
			configTSK_ROUTINGTASK_STACK_SIZE) != 0) {
				NRP_PRINT_ERROR_CODE(UNABLE_TO_INITIALIZE_CSP);
				return UNABLE_TO_INITIALIZE_CSP;
		}
		
		csp_initialized = true;
		
	}
	else NRP_PRINT_WARNING("\rCSP already initialized\r\n");
	
	//Start incoming NRP packet listener
	if (!listener_initialized){
		int8_t err = csp_create_listener(CSP_ANY, &nrp_receive);
		if (err != pdPASS) {
			NRP_PRINT_ERROR("UNABLE_TO_CREATE_LISTENER");
			return err;
		}
		else {
			listener_initialized = true;
			init_conn_pool();
		}
	}
	
	for (int i = 0; i < MAX_PORTS; i++){
		port_queues[i] = xQueueCreate(2, sizeof(nrp_recv_packet_t));
	}
	
	//initialize crc table (checksum)
	crcInit();
	
	//printf("csp_prio_t: %d, nrp_receive_packet: %d\r\n", sizeof(csp_prio_t), sizeof(nrp_recv_packet_t));
	//printModuleInfo();
	return pdPASS;
}

nrp_recv_packet_t * nrp_read(uint8_t port, uint32_t timeout){
	nrp_recv_packet_t *packet = pvPortMalloc(sizeof(*packet));
	uint8_t status = xQueueReceive(port_queues[port - PORT_LOW], packet, timeout);
	if (status != pdTRUE){
		vPortFree(packet);
		return NULL;
	}
	return packet;
}


void add_to_queue(const nrp_recv_packet_t *recv_packet, uint8_t port){
	xQueueSend(port_queues[port-PORT_LOW], recv_packet, 100);
}

int8_t nrp_create_port_handler(uint8_t listen_port, void (*listener_ptr)(void *arg_ptr), uint16_t size, uint8_t pri){
	if (listen_port > PORT_HIGH || listen_port < PORT_LOW) return ILLEGAL_PORT_NUMBER;
	if (is_port_bound(listen_port)) return PORT_ALREADY_BOUND;
	int status = SUCCESS;
	char task_name[12];
	snprintf(task_name, 12, "listen[%i]", listen_port);
	status = xTaskCreate(listener_ptr, task_name, size, (void*)&listen_port, pri, &listen_handle);
	if (status != pdPASS){
		NRP_PRINT_ERROR_CODE(UNABLE_TO_CREATE_LISTENER);
		return UNABLE_TO_CREATE_LISTENER;
	}
	
	//Do not remove the delay below! 
	vTaskDelay(100); //let the newly created task run for a bit to initialize correctly
	return status;
}

int8_t nrp_bind_port(uint8_t port){
	if (port > PORT_HIGH || port < PORT_LOW) return ILLEGAL_PORT_NUMBER;
	else if (bound_ports[port - PORT_LOW] == false){
		bound_ports[port - PORT_LOW] = true;
		return SUCCESS;
	}
	else return PORT_ALREADY_BOUND;
}

Bool is_port_bound(uint8_t port){
	if (port > PORT_HIGH || port < PORT_LOW) return ILLEGAL_PORT_NUMBER;
	else return bound_ports[port - PORT_LOW];
}

void * csp_buffer_get_wrapper(size_t size){
	csp_packet_t *packet = csp_buffer_get(size);
	packet->id.ext = 0;
	packet->length = 0;
	for(int i = 0; i < CSP_PADDING_BYTES; i++){
		packet->padding[i] = '\0';
	}
	return packet;
}

int8_t nrp_send(const module_socket_t *dest, const module_socket_t *source, const uint8_t *payload, uint32_t payload_length, uint8_t priority){
	
	if (payload_length == 0){
		NRP_PRINT_ERROR("Empty payload");
		return PAYLOAD_ERROR;
	}
	else if (payload_length >= 0xffffff){
		NRP_PRINT_ERROR("Payload is too big");
		return PAYLOAD_ERROR;
	}
	
	// Generate a "dummy" csp header to be able to calculate a checksum including the csp header
	uint8_t csp_dummy_header[CSP_HEADER_LENGTH];
	make_csp_dummy_header(csp_dummy_header, dest, source, priority);
	
	nrp_conn_ref conn_ref;
	conn_ref.client_addr = source->csp_address;
	conn_ref.client_port = source->port;
	conn_ref.server_addr = dest->csp_address;
	conn_ref.server_port = dest->port;
	
	//Request a connection from connection pool
	int8_t sta = nrp_connection_create(conn_ref, payload_length, false);
	if (sta != SUCCESS)	return sta;
	
	
	
	
	//determine how many packets which will be in the stream
	uint8_t stream_length = payload_length % NRP_MAX_PAYLOAD_SIZE == 0 ? (payload_length / NRP_MAX_PAYLOAD_SIZE) : (payload_length / NRP_MAX_PAYLOAD_SIZE) + 1;
	uint8_t stream[stream_length][CSP_MAX_DATA_SIZE];
	
	//split the payload into nrp packets and save them in 2D array stream
	nrp_use_segmentation((uint8_t**)stream, payload, payload_length, csp_dummy_header);
	
	//setting number of retransmissions for each packet
	uint8_t retransmissions[stream_length + 1];
	for (int i = 0; i < stream_length + 1; i++) retransmissions[i] = RETRANS_MAX;
	uint8_t fin_retrans = RETRANS_MAX;
	
	//For testing:
	#if NRP_TESTING == true
		static uint32_t corrupted_packets = 0;
		static uint32_t packets_sent = 0;
		static uint32_t total_pkts = 0;
		uint64_t start_t_stamp = xTaskGetTickCount();
	#endif
	
	int at_packet = 0;
	conn_set_total_packets(conn_ref, stream_length);
	conn_set_at_packet(conn_ref, at_packet);
	
	//Send all packets of the stream
	while (conn_get_fin_ack_received(conn_ref) == false){
		uint32_t time = xTaskGetTickCount();
		at_packet = conn_get_at_packet(conn_ref);
		
		//if all of the payload has been sent, transmit the FIN packet to indicate end of transmission
		if ((fin_retrans == RETRANS_MAX || time >= conn_get_last_transmit_timestamp(conn_ref) + RTT_MULTIPLE*conn_get_rtt(conn_ref)) && get_conn_next_expected_seq_num(conn_ref) == 0){
			if (fin_retrans == 0){
				NRP_PRINT_WARNING("FIN was not acknowledged, data may or may not have been received properly");
				break;
			}
			csp_packet_t *out_packet = csp_buffer_get_wrapper(sizeof(csp_packet_t) + sizeof(NRP_HEADER_SIZE));
			out_packet->length = NRP_HEADER_SIZE;
			
			uint8_t fin_packet[6];
			nrp_make_FIN_packet(fin_packet, csp_dummy_header);
			
			
			//For testing, add the possibility of a packet being corrupted with one or more bit-flips
			#if NRP_TESTING == true
			uint16_t rando = rand() % 1000;
			if (rando < (uint16_t)(NRP_TEST_PACKET_ERROR_PROB * 10)) {
				flip_random_bits(fin_packet, (rand() % 10) + 1);
				//corrupted = true;
				//corrupted_packets++;
				//printf("Corrupted packets: %i\r\n", corrupted_packets);
			}
			#endif
			
			memcpy(out_packet->data, fin_packet, NRP_HEADER_SIZE);
			
			NRP_PRINT("Sending FIN");
			int8_t a_status = csp_sendto(priority, dest->csp_address, dest->port, source->port, (uint32_t)NULL, out_packet, NRP_ACK_TIMEOUT_MIN);
			if (a_status == CSP_ERR_NOMEM){
				//free packet from csp_buffer
				NRP_PRINT_WARNING("csp_send() returned error code");
			}
			csp_buffer_free(out_packet);
			//vTaskDelay(2);
			conn_set_last_transmit_timestamp(conn_ref, xTaskGetTickCount());
			fin_retrans--;
			#if NRP_TESTING
				packets_sent++;
			#endif
		}
		
		else if (conn_get_SYN_NAK_recv(conn_ref)){
			//syn nak was received, stop transmission with error
			nrp_connection_close(conn_ref);
			NRP_PRINT_ERROR("Send failed");
			vTaskDelay(CONN_TIMEOUT); //wait for connection on other side to time out
			return SEND_FAILED;
		}
		
		//if window not full, send next packet of stream
		else if(conn_get_window(conn_ref) > 0 && at_packet < stream_length && !(at_packet == 1 && get_conn_packets_received(conn_ref) == 0)){
			//Make small delay if syn was previous packet
			
			uint8_t nrp_packet_length = NRP_HEADER_SIZE + stream[at_packet][NRP_PAYLOAD_LENGTH_BYTE];
			
			// Allocate memory for packet in csp_buffer
			csp_packet_t *out_packet = csp_buffer_get_wrapper(sizeof(csp_packet_t) + sizeof(NRP_HEADER_SIZE) + sizeof(stream[at_packet][NRP_PAYLOAD_LENGTH_BYTE]));
			if (out_packet == NULL){
				NRP_PRINT_ERROR("CSP buffer malloc failed");
				continue;
			}
			out_packet->length = nrp_packet_length;
			
			//need to copy the packet for testing purposes (bit flips). Cant perform bitflips to the original packet stored in stream
			uint8_t nrp_packet_to_send[nrp_packet_length];
			memcpy(nrp_packet_to_send, stream[at_packet], nrp_packet_length);
			
			
			//For testing, add the possibility of a packet being corrupted with one or more bit-flips
			#if NRP_TESTING == true
			uint16_t rando = rand() % 1000;
			if (rando < (uint16_t)(NRP_TEST_PACKET_ERROR_PROB * 10)) {
				flip_random_bits(nrp_packet_to_send, (rand() % 10) + 1);
				//corrupted = true;
				corrupted_packets++;
				//printf("Corrupted packets: %i\r\n", corrupted_packets);
			}
			#endif
			
			
			//copy nrp_packet to csp_packet structure
			memcpy(out_packet->data, nrp_packet_to_send, nrp_packet_length);
			
			if (at_packet == 0) {
				conn_set_syn_stamp(conn_ref, xTaskGetTickCount());
			}
			//send packet over CSP
			int8_t a_status = csp_sendto(priority, dest->csp_address, dest->port, source->port, (uint32_t)NULL, out_packet, (uint32_t)NULL);
			if (a_status == CSP_ERR_NOMEM){
				//free packet from csp_buffer
				NRP_PRINT_WARNING("csp_send() returned error code");
			}
			
			else conn_decrement_window(conn_ref);
			csp_buffer_free(out_packet);
			//csp_buffer_free(out_packet);
			conn_set_last_transmit_timestamp(conn_ref, xTaskGetTickCount());
	
			
			#if NRP_TESTING
				packets_sent++;
			#endif
			
			//increment at_packet to send next packet in the next iteration
			conn_set_at_packet(conn_ref, ++at_packet);	
		}
		
		//waiting for ack timed out, retransmit from last acknowledged packet
		else if (xTaskGetTickCount() >= conn_get_last_transmit_timestamp(conn_ref) + RTT_MULTIPLE*conn_get_rtt(conn_ref)){
			// go back to the packet that was not acknowledged
			uint8_t retransmit_indx = get_conn_packets_received(conn_ref);
			
			//check if the packet has been retransmitted max retransmissions already
			if (retransmissions[retransmit_indx] > 0){
				NRP_PRINT_WARNING("Retransmitting due to timeout");
				//printf("retransmit_indx: %i\r\n",retransmit_indx);
				conn_set_at_packet(conn_ref, retransmit_indx);
				retransmissions[retransmit_indx] = retransmissions[retransmit_indx] - 1;
				conn_reset_window(conn_ref);
				//printf("Done retransmitting\r\n");
			}
			//else: sending failed due to three retransmission fails of the same packet
			else {
				//printf("Sending done, closing connection\r\n");
				vTaskDelay(CONN_TIMEOUT);
				nrp_connection_close(conn_ref);
				NRP_PRINT_ERROR("Send failed because of timeout");
				#if NRP_TESTING
					//printf("Corrupted: %i, packets_sent: %i, percentage loss: %f, time:%i\r\n", corrupted_packets, packets_sent, (float)((float)corrupted_packets / (float)packets_sent)*100, (uint16_t)(xTaskGetTickCount() - start_t_stamp));
				#endif
				return SEND_FAILED;
			}
		}
		//else vTaskDelay(10);
		vTaskDelay(3); 
	}
	
	#if NRP_TESTING 
		//printf("Corrupted: %i, packets_sent: %i, percentage loss: %f, time:%i\r\n", corrupted_packets, packets_sent, (float)((float)corrupted_packets / (float)packets_sent)*100, (uint16_t)(xTaskGetTickCount() - start_t_stamp));
		//printf("%i:%i\r\n", ++total_pkts, (uint16_t)(xTaskGetTickCount() - start_t_stamp));
	#endif
	//printf("CSP_buffer remaining: %i\r\n", csp_buffer_remaining());
	//vTaskDelay(10);
	nrp_connection_close(conn_ref);
	return SUCCESS;
}


int8_t nrp_send_unreliable_datagram(const module_socket_t *dest, const module_socket_t *source, const uint8_t *payload, uint8_t payload_length, csp_prio_t priority){
	
	if (payload_length == 0){
		NRP_PRINT_ERROR("Empty payload");
		return PAYLOAD_ERROR;
	}
	else if (payload_length > NRP_MAX_PAYLOAD_SIZE){
		NRP_PRINT_ERROR("Payload too big");
		return PAYLOAD_ERROR;	
	}
	
	// Generate a "dummy" csp header to be able to calculate a checksum including the csp header
	uint8_t csp_dummy_header[CSP_HEADER_LENGTH];
	make_csp_dummy_header(csp_dummy_header, dest, source, priority);
	
	csp_packet_t *out_packet = csp_buffer_get_wrapper(sizeof(csp_packet_t) + sizeof(NRP_HEADER_SIZE) + payload_length);
	out_packet->length = NRP_HEADER_SIZE + payload_length;
	
	uint8_t packet[NRP_HEADER_SIZE + payload_length];
	
	for (int i = 0; i < NRP_HEADER_SIZE + payload_length; i++) packet[i] = '\0';
	
	nrp_make_unreliable_packet(packet, payload, payload_length, csp_dummy_header);
	
	memcpy(out_packet->data, packet, NRP_HEADER_SIZE + payload_length);
	
	int8_t a_status = csp_sendto(priority, dest->csp_address, dest->port, source->port, (uint32_t)NULL, out_packet, (uint32_t)NULL);
	if (a_status == CSP_ERR_NOMEM){
		//free packet from csp_buffer
		NRP_PRINT_ERROR("csp_send() returned error code");
		return SEND_FAILED;
	}
	csp_buffer_free(out_packet);
	
	return SUCCESS;
}



void free_nrp_packet(nrp_recv_packet_t *pckt){
	vPortFree(pckt->data);
	vPortFree(pckt);
}




