/*
 * nrp.h
 *
 * Created: 14-Feb-15 11:23:10 AM
 *  Author: Erlend Riis Jahren
 *
 * This is the API to the reliable NUTS application layer protocol (NRP), which uses 
 * the CSP protocol as the Network layer.
 */ 


#ifndef NRP_H_
#define NRP_H_

#include "csplib/csp_util.h"
//#include "serial/serial.h"
#include "nrp_tester.h"



/************************************************************************/
/* Definitions                                                          */
/************************************************************************/
/*Module definitions*/
#define ONBOARD_COMPUTER		0
#define RADIO					1
#define CAMERA					2
#define ADCS					3

// I2C addresses
#define OBC_I2C_ADDR		0x01
#define RADIO_I2C_ADDR		0x02
#define CAMERA_I2C_ADDR		0x03
#define ADCS_I2C_ADDR		0x04

// CSP addresses
#define OBC_CSP_ADDR		  OBC_I2C_ADDR
#define RADIO_CSP_ADDR		RADIO_I2C_ADDR
#define CAMERA_CSP_ADDR	   CAMERA_I2C_ADDR
#define ADCS_CSP_ADDR		 ADCS_I2C_ADDR

// NRP timeout constants
#define NRP_ACK_TIMEOUT_MAX	   ( portTickType ) 1000 * portTICK_RATE_MS //Should be set based on slowest link, the radio
#define NRP_ACK_TIMEOUT_MIN	   ( portTickType )   50 * portTICK_RATE_MS
#define NRP_READ_TIMEOUT		   ( portTickType )  100 * portTICK_RATE_MS
#define NRP_NO_TIMEOUT			   ( portTickType )    0 * portTICK_RATE_MS
#define CONN_TIMEOUT			   ( portTickType ) 7000 * portTICK_RATE_MS

#define MUTEX_BLOCK_TIME		   ( portTickType ) 1000 * portTICK_RATE_MS

#define NRP_WINDOW_SIZE				5

#define RTT_MULTIPLE				3

#define RETRANS_MAX						 2

//Do not change the ports without totally knowing what you are doing
 
#define PORT_LOW						10
#define PORT_HIGH						24  
#define MAX_PORTS							PORT_HIGH - PORT_LOW + 1


/******* DEBUG *******

To get debug info from nrp, define NRP_DEBUG to > 0
NRP_DEBUG_OFF: No output
NRP_DEBUG_E: Only errors
NRP_DEBUG_EW: Errors and warnings
NRP_DEBUG_EWS: Errors, warnings and success
NRP_DEBUG_ALL: Errors, warnings, success and general messages
*/
#define NRP_DEBUG								NRP_DEBUG_OFF
#define NRP_DEBUG_USE_ANSI						true	//set to false to turn off ANSI escape codes (colors) in output

#define NRP_TESTING								false
#define NRP_TEST_PACKET_ERROR_PROB				2    //(%) percent

/***** DEBUG END ******/


/************************************************************************/
/* Structs                                                              */
/************************************************************************/
typedef struct {
	uint8_t module_number;
	uint8_t csp_address;
	char name[10];
}  nuts_module_t;

typedef struct {
	uint8_t csp_address;
	uint8_t port;
}  module_socket_t;



typedef struct {
	uint8_t		flags;
	uint8_t		data_length;
	uint8_t		*data;
	uint16_t	checksum;
	uint32_t 	seq_ack_num;
} nrp_packet_t; //NUTS reliable packet


typedef struct {
	csp_packet_t csp_packet;
	uint8_t conn_id;
	csp_prio_t prio;
} nrp_send_params_t;

//the packet struct received by a nrp listener
typedef struct {
	uint32_t data_length;
	
	module_socket_t src;
	module_socket_t dest;
	
	uint8_t pri;
	uint8_t *data;
	
} nrp_recv_packet_t;

/************************************************************************/
/* Other                                                                */
/************************************************************************/

//IMPORTANT: Fill in all sub-modules of the satellite and ground station
// and use the "this_module" as reference to the item of the array representing
// the module you're working on.
static const nuts_module_t nuts_modules [4] = {
	{ONBOARD_COMPUTER,		OBC_CSP_ADDR,		"OBC"},
	{RADIO,					RADIO_CSP_ADDR,		"RADIO"},
	{CAMERA,				CAMERA_CSP_ADDR,	"CAMERA"},
	{ADCS,					ADCS_CSP_ADDR,		"ADCS"},
};

cubesat_module this_module;


/************************************************************************/
/* NRP API Functions                                                    */
/************************************************************************/
/* Examples of how to use the nrp library can be found in csp_example.c */

/**
 * Initializes NRP by initializing CSP for the module specified by the parameter
 * Creates a listener task that handles all incoming CSP packets and forwards it to the
 * NRP layer
 *
 * @param module_number:	The index of the nuts_modules array of the wanted module to initialize
 * 
 * @return:					SUCCESS if successful, negative NRP error code if not
 */
int8_t nrp_init(uint8_t module_number);

/**
 * The function is used to create a FreeRTOS task with intent to listen to a port for incoming messages
 * 
 * @param listen_port:		The port the task will listen to
 * @param listener_ptr:		A pointer to the task function
 * @param size:				The memory size to allocate for of the task. IMPORTANT: Keep this as low as
 *							possible, due to the internal memory being filled up quickly when assigning
 *							new tasks.
 * @param pri:				The FreeRTOS priority of the task
 *
 * @return:					SUCCESS if task is created successfully, negative NRP error code if not
 */
int8_t nrp_create_port_handler(uint8_t listen_port, void (*listener_ptr)(void *arg_ptr), uint16_t size, uint8_t pri);

/**
 * Blocking function to read from a port. Use function nrp_bind_port() before reading on the port
 * PS: remember to free the nrp_recv_packet_t which is returned from this function if a packet was received and read.
 * Use the free_nrp_packet() function to free the packet.
 *
 * @param port:				The port you want to read from
 * @param timeout:			Blocking timeout. 			
 * 
 * @return:					a pointer to a nrp_recv_packet_t struct, remember to free the packet afterwords
 */
nrp_recv_packet_t * nrp_read(uint8_t port, uint32_t timeout);

/*
 * The function used for sending data reliably with the NRP protocol using a Go-Back-N ARQ protocol. Using this function, 
 * data will be segmented into appropriate sizes and transmitted using the CSP library to a receiving node. 
 * At the receiving node, the data is put back together, and read as one message.
 *
 * (!)IMPORTANT: The priority of the task calling this function must be "tskIDLE_PRIORITY" !!!
 *
 * @param *dest:				The destination socket
 * @param *source:				The source socket
 * @param *payload:				Pointer to the data you want to send. 
 * @param payload_length:		Length of payload, up to 2^24
 * @param priority:				The csp_prio_t priority (0 - 3, 0 being critical, 3 being low)
 *
 * @return						SUCCESS if successful, negative NRP error code if not
 */
int8_t nrp_send(const module_socket_t *dest, const module_socket_t *source, const uint8_t *payload, uint32_t payload_length, uint8_t priority);

/*
 * This function transmits the message without the reliability features of NRP. The function does not handle segmentation,
 * but CRC-16 is used to assure the data is correct. Use this function if reliability and segmentation is not wanted.
 *
 * @param *dest:				The destination socket
 * @param *source:				The source socket
 * @param *payload:				Pointer to the data you want to send.
 * @param payload_length:		Length of payload, up to 2^24
 * @param priority:				The csp_prio_t priority (0 - 3, 0 being critical, 3 being low)
 *
 * @return						SUCCESS if successful, negative NRP error code if not
 */
int8_t nrp_send_unreliable_datagram(const module_socket_t *dest, const module_socket_t *source, const uint8_t *payload, uint8_t payload_length, csp_prio_t priority);


/**
 * Function to bind a task to a port. When bound, other tasks will not be able to bind to the same port.
 *
 * @param port:				Port to bind to
 * 
 * @return:					SUCCESS if successful, negative NRP error code if not
 */
int8_t nrp_bind_port(uint8_t port);

 /** 
 * Allocates the specified packet size (NB: Including sizeof(csp_packet_t)).
 * The packet is then cleaned by initializing certain members, and inserting padding.
 *
 * @param size:				the size you want to allocate
 * @return:					void*
 */
void * csp_buffer_get_wrapper(size_t size);

/**
 * Free a nrp_recv_packet_t struct when the message has been received and processed.
 *
 * @param *pckt:			The packet to free
 * 
 * @return:					void
 */
void free_nrp_packet(nrp_recv_packet_t *pckt);



/************************************************************************/
/* NON API Functions                                                    */
/************************************************************************/

/**
 * NOT PART OF NRP THE API!!
 * ---------------------
 * check if a port has a task listener
 *
 * @param port:				The port you want to check for listener
 * 
 * @return:					true if the port has a listener, false if not. 
 */
Bool is_port_bound(uint8_t port);

/**
 * NOT PART OF THE NRP API!!
 * ---------------------
 * Adds a packet to a port queue. This function is called from within the NRP library, and should not be used when 
 * developing.
 *
 * @param *recv_packet:		Pointer to a nrp_recv_packet_t struct that will be added to a port queue
 * @param port:				The port queue the struct should be added to
 * 
 * @return:					void
 */
void add_to_queue(const nrp_recv_packet_t *recv_packet, uint8_t port);

#endif /* NCSP_H_ */