/*
 * nrp_conn.c
 *
 * Created: 03-Mar-15 6:37:46 PM
 *  Author: Erlend
 */ 

#include "nrp.h"
#include "nrp_conn.h"
#include "nrp_debug.h"
#include "nrp_error.h"
#include "nrp_header.h"
#include "semphr.h"




typedef struct nrp_connection_t{
	Bool in_use;
	Bool fin_ack_recieved;
	Bool nak_sent;
	Bool syn_nak_received;
	
	uint8_t window_left;
	
	uint16_t packets_received;
	uint16_t at_packet;
	uint16_t total_packets;
	uint16_t rtt;
	module_socket_t client; //module_socket_t is 2 bytes
	module_socket_t server;
	
	uint32_t init_seq_num;
	uint32_t next_expected_seq_num;
	
	uint64_t last_updated_timestamp;
	uint64_t last_transmit_timestamp;
	uint64_t initialized_timestamp; 
	uint64_t syn_stamp;
	
	nrp_recv_packet_t *recv_packet_t;
} nrp_connection_t;

//dummy illegal socket to return if something went wrong
module_socket_t error_socket;



static int8_t nrp_remove_connection_from_pool(nrp_conn_ref conn_ref);
/*
 * Connection pool functionality:
 * An array of nrp_connection_t structs are allocated in memory
 * when a thread requests a connection, it is given the next available index of 
 * the array. When it closes its connection, the array item at that index is removed
 * and the array elements after this index is shifted back by one. 
 *
 * The connection pool is guarded by a mutex, and can only be changed from within this file
 */

//static xSemaphoreHandle conn_pool_mutex;
static nrp_connection_t nrp_conn_pool[NRP_MAX_CONN];


//Connection pool mutex
xSemaphoreHandle conn_mutex; 

int8_t conn_take_mutex(void);
int8_t conn_take_mutex(void){
	if (xSemaphoreTake(conn_mutex, MUTEX_BLOCK_TIME) != pdPASS){
		NRP_PRINT_ERROR("Taking semaphore timed out");
		return UNABLE_TO_OBTAIN_MUTEX;
	}
	return pdPASS;
}

void conn_give_mutex(void){
	xSemaphoreGive(conn_mutex);
}


Bool conn_ref_match(const nrp_conn_ref *conn_ref, uint8_t conn_pool_index){
	uint8_t i = conn_pool_index;
	if (nrp_conn_pool[i].client.port == conn_ref->client_port && nrp_conn_pool[i].server.port == conn_ref->server_port
	&& nrp_conn_pool[i].client.csp_address == conn_ref->client_addr
	&& nrp_conn_pool[i].server.csp_address == conn_ref->server_addr){
		return true;
	}
	else return false;
}

int8_t init_conn_pool(void){
	conn_mutex = xSemaphoreCreateMutex();
	if (conn_mutex == NULL){
		NRP_PRINT_ERROR("Unable to create mutex, system init failed");
		return UNABLE_TO_CREATE_MUTEX;
	}
	
	
	if (conn_take_mutex() != pdPASS) return UNABLE_TO_OBTAIN_MUTEX;
	for (int i = 0; i < NRP_MAX_CONN; i++){
		nrp_conn_pool[i].in_use = false;
	}
	conn_give_mutex();
	
	error_socket.port = 0;
	error_socket.csp_address = 0;
	
	
	return SUCCESS;
}

//function to check if a connection reference is valid. the function is used by most of the functions below
Bool valid_conn_ref(nrp_conn_ref conn_ref){
	if (conn_ref.client_port < PORT_LOW || conn_ref.client_port > PORT_HIGH || conn_ref.server_port < PORT_LOW || conn_ref.server_port > PORT_HIGH){
		NRP_PRINT_WARNING("Invalid ports passed to function");
		return false;
	}
	else return true;
}

void erase_outdated_connections(void);
void erase_outdated_connections(void){
	if (conn_take_mutex() != pdPASS) return;
	for (int i = 0; i < NRP_MAX_CONN; i++){
		if (nrp_conn_pool[i].in_use){
			if (nrp_conn_pool[i].last_updated_timestamp + CONN_TIMEOUT <= xTaskGetTickCount()){
				static nrp_conn_ref conn_ref;
				conn_ref.client_addr = nrp_conn_pool[i].client.csp_address;
				conn_ref.client_port = nrp_conn_pool[i].client.port;
				conn_ref.server_addr = nrp_conn_pool[i].server.csp_address;
				conn_ref.server_port = nrp_conn_pool[i].server.port;
				
				//free data that was never sent to receive queue
				if (nrp_conn_pool[i].recv_packet_t != NULL){
					if (nrp_conn_pool[i].recv_packet_t->data != NULL){
						vPortFree(nrp_conn_pool[i].recv_packet_t->data);
					}
				}
				
				//printf("i: %i, cli_src: %i\r\n", i, nrp_conn_pool[i].client.csp_address);
				NRP_PRINT_ERROR("Closing outdated connection");
				conn_give_mutex();
				nrp_connection_close(conn_ref);
				conn_take_mutex();
				continue;
			}
			
		}
		else {
			conn_give_mutex();
			return;
		}
		
		conn_give_mutex();
	}
}

int8_t nrp_get_next_available_index_from_connection_pool(nrp_conn_ref conn_ref){
	if (conn_take_mutex() != pdPASS) return UNABLE_TO_OBTAIN_MUTEX;
	for (int i = 0; i < NRP_MAX_CONN; i++){
		if (!nrp_conn_pool[i].in_use){
			conn_give_mutex();
			return i;
		}
		
		else if (nrp_conn_pool[i].client.port == conn_ref.client_port && nrp_conn_pool[i].server.port == conn_ref.server_port
		&& nrp_conn_pool[i].client.csp_address == conn_ref.client_addr
		&& nrp_conn_pool[i].server.csp_address == conn_ref.server_addr && nrp_conn_pool[i].in_use){
			conn_give_mutex();
			return PORT_ALREADY_CONNNECTED;
		}
		
	}
	conn_give_mutex();
	NRP_PRINT_ERROR("Connection pool is full");
	return CONN_POOL_FULL;
}





//a port can only have one connection at a time.
static int8_t nrp_remove_connection_from_pool(nrp_conn_ref conn_ref){
	Bool conn_found = false;
	//grab items mutex
	if (conn_take_mutex() != pdPASS) return UNABLE_TO_OBTAIN_MUTEX;
	for (int i = 0; i < NRP_MAX_CONN; i++){
		
		if (!nrp_conn_pool[i].in_use){
			if (conn_found)	nrp_conn_pool[i-1] = nrp_conn_pool[i];
			break;
		}
		else if (conn_found && nrp_conn_pool[i].in_use){
			nrp_conn_pool[i-1] = nrp_conn_pool[i];
		}
		else if (conn_ref_match(&conn_ref, i)){
			conn_found = true;
			
			//Only free recv packet, not the data in it, as this is later freed by receiving port task
			if (nrp_conn_pool[i].client.csp_address != nuts_modules[this_module].csp_address){
				vPortFree(nrp_conn_pool[i].recv_packet_t);
			}
			NRP_PRINT("Connection was removed from pool");
		}
	}
	conn_give_mutex();
	return SUCCESS;
}




void nrp_connection_close(nrp_conn_ref conn_ref){
	
	if (!valid_conn_ref(conn_ref)) return;
	nrp_remove_connection_from_pool(conn_ref);
}




//Returns NULL if not able to create connection
int8_t nrp_connection_create(const nrp_conn_ref conn_ref, uint32_t init_seq_num, csp_prio_t prio){
	
	if (!valid_conn_ref(conn_ref)){
		NRP_PRINT_ERROR("Invalid connection reference passed to function");
		return ILLEGAL_PORT_NUMBER;
	}
	//clean-up of connection pool, remove outdated connections that has failed to close
	erase_outdated_connections();
	
	
	if (conn_take_mutex() != pdPASS) return UNABLE_TO_OBTAIN_MUTEX;
	
	//check if there is already a connection between these end points. 
	for (int i = 0; i < NRP_MAX_CONN; i++){
		//Check if there is already a connection on this port
		if (nrp_conn_pool[i].client.port == conn_ref.client_port && nrp_conn_pool[i].server.port == conn_ref.server_port
			&& nrp_conn_pool[i].client.csp_address == conn_ref.client_addr
			&& nrp_conn_pool[i].server.csp_address == conn_ref.server_addr && nrp_conn_pool[i].in_use) {
				conn_give_mutex();
				NRP_PRINT_WARNING("Port already connected");
				return (int8_t) PORT_ALREADY_CONNNECTED;
		}
		
		else if (!nrp_conn_pool[i].in_use){
			
			//only make receiving struct if this msg is incoming (meaning you are server)
			//dont forget to free memory when packet is read.
			if (nuts_modules[this_module].csp_address == conn_ref.server_addr){
				nrp_conn_pool[i].recv_packet_t = pvPortMalloc(sizeof(nrp_recv_packet_t));
				nrp_conn_pool[i].recv_packet_t->data_length = init_seq_num;
				nrp_conn_pool[i].recv_packet_t->src.csp_address = conn_ref.client_addr;
				nrp_conn_pool[i].recv_packet_t->dest.csp_address = conn_ref.server_addr;
				nrp_conn_pool[i].recv_packet_t->src.port = conn_ref.client_port;
				nrp_conn_pool[i].recv_packet_t->dest.port = conn_ref.server_port;
				nrp_conn_pool[i].recv_packet_t->data = pvPortMalloc(sizeof(uint8_t)*(init_seq_num + 1));
				nrp_conn_pool[i].recv_packet_t->pri = prio;
				
				//check that allocation was successful
				if (nrp_conn_pool[i].recv_packet_t->data == NULL){
					if (nrp_conn_pool[i].recv_packet_t != NULL){
						vPortFree(nrp_conn_pool[i].recv_packet_t);
					}	
					//malloc of data or struct has failed, need to send syn-nak to client
					NRP_PRINT_ERROR("Not able to allocate memory for incoming stream");
					conn_give_mutex();
					return MALLOC_FAILED;
				}
					
				//freertos has no calloc function, so it must be done manually
				for (int j = 0; j < init_seq_num + 1; j++)	{
					nrp_conn_pool[i].recv_packet_t->data[j] = '\0';
				}
			}
			else nrp_conn_pool[i].recv_packet_t = NULL;
				
			nrp_conn_pool[i].server.csp_address = conn_ref.server_addr;
			nrp_conn_pool[i].server.port = conn_ref.server_port;
			nrp_conn_pool[i].client.csp_address = conn_ref.client_addr;
			nrp_conn_pool[i].client.port = conn_ref.client_port;
			nrp_conn_pool[i].init_seq_num = init_seq_num;
			nrp_conn_pool[i].next_expected_seq_num = init_seq_num;
			nrp_conn_pool[i].packets_received = 0;
			nrp_conn_pool[i].in_use = true;
			nrp_conn_pool[i].window_left = NRP_WINDOW_SIZE;
			nrp_conn_pool[i].last_updated_timestamp = xTaskGetTickCount();
			nrp_conn_pool[i].fin_ack_recieved = false;
			nrp_conn_pool[i].nak_sent = false;
			nrp_conn_pool[i].initialized_timestamp = xTaskGetTickCount();
			nrp_conn_pool[i].rtt = NRP_ACK_TIMEOUT_MAX; //transmit timeout
			nrp_conn_pool[i].syn_nak_received = false;
			
			//NRP_PRINT("Connection was taken from pool");
			conn_give_mutex();
			return SUCCESS;
		}
	}
	conn_give_mutex();
	NRP_PRINT_ERROR("Conn pool full");
	return CONN_POOL_FULL;
}

void store_data(nrp_conn_ref conn_ref, uint8_t data[], uint8_t data_length, uint32_t seq_num){
	if (conn_take_mutex() != pdPASS) return;
	for (int i = 0; i < NRP_MAX_CONN; i++){
		if (conn_ref_match(&conn_ref, i) && nrp_conn_pool[i].in_use){
			//copy data
			memcpy(&nrp_conn_pool[i].recv_packet_t->data[nrp_conn_pool[i].init_seq_num - seq_num], data, sizeof(uint8_t)*data_length);
			conn_give_mutex();
			return;
		}
		
	}
	conn_give_mutex();
}


int8_t update_connection(nrp_conn_ref conn_ref, uint32_t new_next_seq_num, uint16_t new_packets_received){
	
	if (!valid_conn_ref(conn_ref)){
		NRP_PRINT_ERROR("Invalid connection reference passed to function");
		return ILLEGAL_PORT_NUMBER;
	}
	
	if (conn_take_mutex() != pdPASS) return UNABLE_TO_OBTAIN_MUTEX;
			
	for (int i = 0; i < NRP_MAX_CONN; i++){

		
		if (conn_ref_match(&conn_ref, i) && nrp_conn_pool[i].in_use){
			
			nrp_conn_pool[i].next_expected_seq_num = new_next_seq_num;
			nrp_conn_pool[i].packets_received = new_packets_received;
			nrp_conn_pool[i].last_updated_timestamp = xTaskGetTickCount();
			
			conn_give_mutex();
			return SUCCESS;
		}
	}
	conn_give_mutex();
	NRP_PRINT_ERROR("Connection not found");
	return CONN_ID_NOT_FOUND;
}




module_socket_t get_conn_client(nrp_conn_ref conn_ref){
	module_socket_t client = error_socket;
	if (conn_take_mutex() != pdPASS) return client; //return error socket (NULL socket)
	for (int i = 0; i < NRP_MAX_CONN; i++){
		if (conn_ref_match(&conn_ref, i)){	
			
			client = nrp_conn_pool[i].client;
			break;
		}	
	}
	conn_give_mutex();	
	return client;
}

module_socket_t get_conn_server(nrp_conn_ref conn_ref){
	module_socket_t server = error_socket;
	if (conn_take_mutex() != pdPASS) return server;
	for (int i = 0; i < NRP_MAX_CONN; i++){
		if (conn_ref_match(&conn_ref, i)){
			server = nrp_conn_pool[i].server;
			break;
		}
	}
	conn_give_mutex();
	return server;
}


int64_t get_conn_next_expected_seq_num(nrp_conn_ref conn_ref){
	int64_t next_seq = -1;
	if (conn_take_mutex() != pdPASS) return UNABLE_TO_OBTAIN_MUTEX;
	for (int i = 0; i < NRP_MAX_CONN; i++){
		if (!nrp_conn_pool[i].in_use){
			break;
		}
		if (conn_ref_match(&conn_ref, i)){
			next_seq = (int64_t)nrp_conn_pool[i].next_expected_seq_num;
			break;
		}
	}
	conn_give_mutex();
	return next_seq;
}


int64_t get_conn_init_seq_num(nrp_conn_ref conn_ref){
	int32_t init_seq = -1;
	if (conn_take_mutex() != pdPASS) return UNABLE_TO_OBTAIN_MUTEX;
	for (int i = 0; i < NRP_MAX_CONN; i++){
		if (!nrp_conn_pool[i].in_use){
			break;
		}
		if (conn_ref_match(&conn_ref, i)){
			init_seq = (int64_t)nrp_conn_pool[i].init_seq_num;
			break;
		}
	}
	conn_give_mutex();
	return init_seq;
}
int16_t get_conn_packets_received(nrp_conn_ref conn_ref){
	int16_t packets_recv = -1;
	if (conn_take_mutex() != pdPASS) return UNABLE_TO_OBTAIN_MUTEX;
	for (int i = 0; i < NRP_MAX_CONN; i++){
		if (!nrp_conn_pool[i].in_use){
			conn_give_mutex();
			return;
		}
		if (conn_ref_match(&conn_ref, i)){
			packets_recv = (int16_t)nrp_conn_pool[i].packets_received;	
			break;
		}
	}
	conn_give_mutex();
	return packets_recv;
}
Bool conn_exists(nrp_conn_ref conn_ref){
	Bool in_use = false;
	if (conn_take_mutex() != pdPASS) return UNABLE_TO_OBTAIN_MUTEX;
	for (int i = 0; i < NRP_MAX_CONN; i++){
		if (!nrp_conn_pool[i].in_use){
			break;
		}
		if (conn_ref_match(&conn_ref, i)){
			in_use = nrp_conn_pool[i].in_use;
			break;
		}
	}
	conn_give_mutex();
	return in_use;
}

int8_t conn_get_window(nrp_conn_ref conn_ref){
	int8_t window = -1;
	if (conn_take_mutex() != pdPASS) return UNABLE_TO_OBTAIN_MUTEX;
	for (int i = 0; i < NRP_MAX_CONN; i++){
		if (!nrp_conn_pool[i].in_use){
			break;
		}
		if (conn_ref_match(&conn_ref, i)){
			window = nrp_conn_pool[i].window_left;
			break;
		}
	}
	conn_give_mutex();
	return window;
}

void conn_slide_window(nrp_conn_ref conn_ref){
	if (conn_take_mutex() != pdPASS) return;
	for (int i = 0; i < NRP_MAX_CONN; i++){
		if (!nrp_conn_pool[i].in_use){
			conn_give_mutex();
			return;
		}
		if (conn_ref_match(&conn_ref, i)){
			nrp_conn_pool[i].window_left++;
			break;
		}
	}
	conn_give_mutex();
}

void conn_decrement_window(nrp_conn_ref conn_ref){
	if (conn_take_mutex() != pdPASS) return;
	for (int i = 0; i < NRP_MAX_CONN; i++){
		if (!nrp_conn_pool[i].in_use){
			conn_give_mutex();
			return;
		}
		if (conn_ref_match(&conn_ref, i)){
			nrp_conn_pool[i].window_left--;
			break;
		}
	}
	conn_give_mutex();
}

void conn_reset_window(nrp_conn_ref conn_ref){
	if (conn_take_mutex() != pdPASS) return;
	for (int i = 0; i < NRP_MAX_CONN; i++){
		if (!nrp_conn_pool[i].in_use){
			conn_give_mutex();
			return;
		}
		if (conn_ref_match(&conn_ref, i)){
			nrp_conn_pool[i].window_left = NRP_WINDOW_SIZE;
			break;
		}
	}
	conn_give_mutex();
}


uint64_t conn_get_last_transmit_timestamp(nrp_conn_ref conn_ref){
	uint64_t last_transmit = 0;
	if (conn_take_mutex() != pdPASS) return UNABLE_TO_OBTAIN_MUTEX;
	for (int i = 0; i < NRP_MAX_CONN; i++){
		if (!nrp_conn_pool[i].in_use){
			break;
		}
		if (conn_ref_match(&conn_ref, i)){
			last_transmit = nrp_conn_pool[i].last_transmit_timestamp;
			break;
		}
	}
	conn_give_mutex();
	return last_transmit;
}

void conn_set_last_transmit_timestamp(nrp_conn_ref conn_ref, uint64_t timestamp){
	if (conn_take_mutex() != pdPASS) return;
	for (int i = 0; i < NRP_MAX_CONN; i++){
		if (!nrp_conn_pool[i].in_use){
			conn_give_mutex();
			return;
		}
		if (conn_ref_match(&conn_ref, i)){
			nrp_conn_pool[i].last_transmit_timestamp = timestamp;
			nrp_conn_pool[i].last_updated_timestamp = xTaskGetTickCount();
			break;
		}
	}
	conn_give_mutex();
}

void conn_set_fin_ack_received(nrp_conn_ref conn_ref){
	if (conn_take_mutex() != pdPASS) return;
	for (int i = 0; i < NRP_MAX_CONN; i++){
		if (!nrp_conn_pool[i].in_use){
			conn_give_mutex();
			return;
		}
		if (conn_ref_match(&conn_ref, i)){
			nrp_conn_pool[i].fin_ack_recieved = true;
			nrp_conn_pool[i].last_updated_timestamp = xTaskGetTickCount();
			break;
		}
	}
	conn_give_mutex();
}

Bool conn_get_fin_ack_received(nrp_conn_ref conn_ref){
	Bool fin_ack_recv = false;
	if (conn_take_mutex() != pdPASS) return UNABLE_TO_OBTAIN_MUTEX;
	for (int i = 0; i < NRP_MAX_CONN; i++){
		if (!nrp_conn_pool[i].in_use){
			break;
		}
		if (conn_ref_match(&conn_ref, i)){
			fin_ack_recv = nrp_conn_pool[i].fin_ack_recieved;
			break;
		}
	}
	conn_give_mutex();
	return fin_ack_recv;
}

void conn_set_nak_sent(nrp_conn_ref conn_ref, Bool val){
	if (conn_take_mutex() != pdPASS) return;
	for (int i = 0; i < NRP_MAX_CONN; i++){
		if (!nrp_conn_pool[i].in_use){
			conn_give_mutex();
			return;
		}
		if (conn_ref_match(&conn_ref, i)){
			nrp_conn_pool[i].nak_sent = val;
			break;
		}
	}
	conn_give_mutex();
}

Bool conn_get_nak_sent(nrp_conn_ref conn_ref){
	Bool nak_sent = false;
	if (conn_take_mutex() != pdPASS) return UNABLE_TO_OBTAIN_MUTEX;
	for (int i = 0; i < NRP_MAX_CONN; i++){
		if (!nrp_conn_pool[i].in_use){
			break;
		}
		if (conn_ref_match(&conn_ref, i)){
			nak_sent = nrp_conn_pool[i].nak_sent;
			break;
		}
	}
	conn_give_mutex();
	return nak_sent;
}

int32_t conn_get_at_packet(nrp_conn_ref conn_ref){
	uint32_t at_packet = -1;
	if (conn_take_mutex() != pdPASS) return UNABLE_TO_OBTAIN_MUTEX;
	for (int i = 0; i < NRP_MAX_CONN; i++){
		if (!nrp_conn_pool[i].in_use){
			break;
		}
		if (conn_ref_match(&conn_ref, i)){
			at_packet = nrp_conn_pool[i].at_packet;
			break;
		}
	}
	conn_give_mutex();
	return at_packet;
}


void conn_set_at_packet(nrp_conn_ref conn_ref, uint16_t at_packet){
	if (conn_take_mutex() != pdPASS) return;
	for (int i = 0; i < NRP_MAX_CONN; i++){
		if (!nrp_conn_pool[i].in_use){
			conn_give_mutex();
			return;
		}
		if (conn_ref_match(&conn_ref, i)){
			nrp_conn_pool[i].at_packet = at_packet;
			break;
		}
	}
	conn_give_mutex();
}

void conn_set_total_packets(nrp_conn_ref conn_ref, uint16_t total_packets){
	if (conn_take_mutex() != pdPASS) return;
	for (int i = 0; i < NRP_MAX_CONN; i++){
		if (!nrp_conn_pool[i].in_use){
			conn_give_mutex();
			return;
		}
		if (conn_ref_match(&conn_ref, i)){
			nrp_conn_pool[i].total_packets = total_packets;
			break;
		}
	}
	conn_give_mutex();
}

int32_t conn_get_total_packets(nrp_conn_ref conn_ref){
	int32_t total = -1;
	if (conn_take_mutex() != pdPASS) return UNABLE_TO_OBTAIN_MUTEX;
	for (int i = 0; i < NRP_MAX_CONN; i++){
		if (!nrp_conn_pool[i].in_use){
			break;
		}
		if (conn_ref_match(&conn_ref, i)){
			
			total = nrp_conn_pool[i].total_packets;
			break;
		}
	}
	conn_give_mutex();
	return total;
}


//Testing

int64_t conn_get_initial_timestamp(nrp_conn_ref conn_ref){
	int64_t init_t = -1;
	if (conn_take_mutex() != pdPASS) return UNABLE_TO_OBTAIN_MUTEX;
	for (int i = 0; i < NRP_MAX_CONN; i++){
		if (!nrp_conn_pool[i].in_use){
			break;
		}
		if (conn_ref_match(&conn_ref, i)){
			init_t = nrp_conn_pool[i].initialized_timestamp;
			break;
		}
	}
	conn_give_mutex();
	return init_t;
}

nrp_recv_packet_t * conn_get_recv_packet_ptr(nrp_conn_ref conn_ref){
	nrp_recv_packet_t *ptr = NULL;
	if (conn_take_mutex() != pdPASS) return NULL;
	for (int i = 0; i < NRP_MAX_CONN; i++){
		if (!nrp_conn_pool[i].in_use){
			break;
		}
		if (conn_ref_match(&conn_ref, i)){
			ptr = nrp_conn_pool[i].recv_packet_t;
			break;
		}
	}
	conn_give_mutex();
	return ptr;
}


uint64_t conn_get_syn_stamp(nrp_conn_ref conn_ref){
	uint64_t syn_stamp = 0;
	if (conn_take_mutex() != pdPASS) return UNABLE_TO_OBTAIN_MUTEX;
	for (int i = 0; i < NRP_MAX_CONN; i++){
		if (!nrp_conn_pool[i].in_use){
			break;
		}
		if (conn_ref_match(&conn_ref, i)){
			syn_stamp = nrp_conn_pool[i].syn_stamp;
			break;
		}
	}
	conn_give_mutex();
	return syn_stamp;
}


void conn_set_syn_stamp(nrp_conn_ref conn_ref, uint64_t stamp){
	if (conn_take_mutex() != pdPASS) return;
	for (int i = 0; i < NRP_MAX_CONN; i++){
		if (!nrp_conn_pool[i].in_use){
			conn_give_mutex();
			return;
		}
		if (conn_ref_match(&conn_ref, i)){
			nrp_conn_pool[i].syn_stamp = stamp;
			break;
		}
	}
	conn_give_mutex();
}


uint16_t conn_get_rtt(nrp_conn_ref conn_ref){
	uint64_t rtt = 0;
	if (conn_take_mutex() != pdPASS) return UNABLE_TO_OBTAIN_MUTEX;
	for (int i = 0; i < NRP_MAX_CONN; i++){
		if (!nrp_conn_pool[i].in_use){
			break;
		}
		if (conn_ref_match(&conn_ref, i)){
			rtt = nrp_conn_pool[i].rtt;
			break;
		}
	}
	conn_give_mutex();
	return rtt;
}

void conn_set_rtt(nrp_conn_ref conn_ref, uint16_t rtt){
	if (conn_take_mutex() != pdPASS) return;
	for (int i = 0; i < NRP_MAX_CONN; i++){
		if (!nrp_conn_pool[i].in_use){
			conn_give_mutex();
			return;
		}
		if (conn_ref_match(&conn_ref, i)){
			if (rtt >= NRP_ACK_TIMEOUT_MAX) nrp_conn_pool[i].rtt = NRP_ACK_TIMEOUT_MAX;
			else if (rtt <= NRP_ACK_TIMEOUT_MIN) nrp_conn_pool[i].rtt = NRP_ACK_TIMEOUT_MIN;
			else nrp_conn_pool[i].rtt = rtt;
			break;
		}
	}
	conn_give_mutex();
}


Bool conn_get_SYN_NAK_recv(nrp_conn_ref conn_ref){
	Bool syn_nak_recv = false;
	if (conn_take_mutex() != pdPASS) return UNABLE_TO_OBTAIN_MUTEX;
	for (int i = 0; i < NRP_MAX_CONN; i++){
		if (!nrp_conn_pool[i].in_use){
			break;
		}
		if (conn_ref_match(&conn_ref, i)){
			syn_nak_recv = nrp_conn_pool[i].syn_nak_received;
			break;
		}
	}
	conn_give_mutex();
	return syn_nak_recv;
}

void conn_set_SYN_NAK_recv(nrp_conn_ref conn_ref, Bool syn_nak_recv){
	if (conn_take_mutex() != pdPASS) return;
	for (int i = 0; i < NRP_MAX_CONN; i++){
		if (!nrp_conn_pool[i].in_use){
			conn_give_mutex();
			return;
		}
		if (conn_ref_match(&conn_ref, i)){
			nrp_conn_pool[i].syn_nak_received = syn_nak_recv;
			break;
		}
	}
	conn_give_mutex();
}