/*
 * nrp_conn.h
 *
 * Created: 03-Mar-15 6:37:57 PM
 *  Author: Erlend
 */ 


#ifndef NRP_CONN_H_
#define NRP_CONN_H_

#include <stdint.h>


#define NRP_MAX_CONN		  		 10



typedef struct nrp_conn_ref{
	uint8_t client_addr;
	uint8_t client_port;
	uint8_t server_addr;
	uint8_t server_port;
}nrp_conn_ref;



/************************************************************************/
/* FUNCTIONS                                                            */
/************************************************************************/


/**
 * Initializes the connection pool
 *
 * @ return:	SUCCESS if successful, CONN_POOL_FULL if 
 *				connection pool is full
 */
int8_t init_conn_pool(void);

/**
 * Allocates a place in the connection pool. This needs to be freed later by 
 * nrp_connection_close() to make room for new connections
 *
 * @param *server:			pointer to the server side socket struct
 * @param *client:			pointer to the client side socket struct
 * @param payload length:	the length of the payload to be sent over the connection 
 * @param init_seq_num:		the initial sequence number of the stream, aka the full payload length of the stream
 *
 * @return:					a pointer to the connection if successful, NULL if not
 */
int8_t nrp_connection_create(const nrp_conn_ref conn_ref, uint32_t init_seq_num, csp_prio_t prio);



/**
 * Frees a connection from the connection pool and makes room for new connections
 *
 * @param *conn:	pointer to the connection to be closed
 */
void nrp_connection_close(nrp_conn_ref conn_ref);



/**
 * Returns a pointer to the connection on a specific port if it is in use.
 * A port can only have one connection at a time.
 *
 * @param port:		the port (on "your" side of the connection) you want the connection struct to
 *
 * @return:			a pointer the the connection on the port if it exist, NULL if not
 */
int8_t update_connection(nrp_conn_ref conn_ref, uint32_t new_next_seq_num, uint16_t new_packets_received);

/**
 * Store parts data of the stream at a location in memory
 *
 * @param conn_ref:				the connection reference
 * @param data[]:				array containing the data you want to store
 * @param data_length:			data length
 * @seq_num:					The sequence number of the packet the data was taken from
 *
 * @return						void
 */
void store_data(nrp_conn_ref conn_ref, uint8_t data[], uint8_t data_length, uint32_t seq_num);


//Get and set methods, all implemented with mutual exclusion
module_socket_t get_conn_client(nrp_conn_ref conn_ref);
module_socket_t get_conn_server(nrp_conn_ref conn_ref);
int64_t get_conn_next_expected_seq_num(nrp_conn_ref conn_ref);
int64_t get_conn_init_seq_num(nrp_conn_ref conn_ref);
int16_t get_conn_packets_received(nrp_conn_ref conn_ref);
Bool conn_exists(nrp_conn_ref conn_ref);
int8_t conn_get_window(nrp_conn_ref conn_ref);
void conn_slide_window(nrp_conn_ref conn_ref);
void conn_decrement_window(nrp_conn_ref conn_ref);
void conn_reset_window(nrp_conn_ref conn_ref);
uint64_t conn_get_last_transmit_timestamp(nrp_conn_ref conn_ref);
void conn_set_last_transmit_timestamp(nrp_conn_ref conn_ref, uint64_t timestamp);
void conn_set_fin_ack_received(nrp_conn_ref conn_ref);
Bool conn_get_fin_ack_received(nrp_conn_ref conn_ref);
int32_t conn_get_at_packet(nrp_conn_ref conn_ref);
void conn_set_at_packet(nrp_conn_ref conn_ref, uint16_t at_packet);
void conn_set_total_packets(nrp_conn_ref conn_ref, uint16_t total_packets);
int32_t conn_get_total_packets(nrp_conn_ref conn_ref);
void conn_set_nak_sent(nrp_conn_ref conn_ref, Bool val);
Bool conn_get_nak_sent(nrp_conn_ref conn_ref);
int64_t conn_get_initial_timestamp(nrp_conn_ref conn_ref);
nrp_recv_packet_t * conn_get_recv_packet_ptr(nrp_conn_ref conn_ref);
void conn_set_rtt(nrp_conn_ref conn_ref, uint16_t rtt);
uint16_t conn_get_rtt(nrp_conn_ref conn_ref);
void conn_set_syn_stamp(nrp_conn_ref conn_ref, uint64_t stamp);
uint64_t conn_get_syn_stamp(nrp_conn_ref conn_ref);
void conn_set_SYN_NAK_recv(nrp_conn_ref conn_ref, Bool syn_nak_recv);
Bool conn_get_SYN_NAK_recv(nrp_conn_ref conn_ref);

#endif /* NRP_CONN_H_ */