/*
 * nrp_header.h
 *
 * Created: 16-Feb-15 3:06:32 PM
 *  Author: Erlend
 */ 


#ifndef NCSP_HEADER_H_
#define NCSP_HEADER_H_

#include "conf_csp.h"
#include "nrp.h"
#include <stdint.h>


// NRP header v.1.0
#define NRP_HEADER_SIZE					7
#define NRP_HEADER_SIZE_WITHOUT_CRC		5
#define NRP_NO_PAYLOAD					0
#define NRP_MAX_PAYLOAD_SIZE			(CSP_MAX_DATA_SIZE - NRP_HEADER_SIZE)


//Protocol flags
#define NRP_NO_FLAGS		0x00
#define NRP_ARQ				0x80
#define	NRP_ACK				0x40 | NRP_ARQ //only used when arq is enabled
#define NRP_NAK				0x20 | NRP_ARQ
#define NRP_SYN				0x10 | NRP_ARQ
#define NRP_FIN				0x08 | NRP_ARQ  

#define NRP_FIN_ACK			(NRP_FIN | NRP_ACK)
#define NRP_SYN_ACK			(NRP_SYN | NRP_ACK)
#define NRP_SYN_NAK			(NRP_SYN | NRP_NAK)



//Header bytes
#define NRP_FLAG_BYTE				0x00
#define NRP_SEQ_BYTE_FIRST			0x01	
#define NRP_SEQ_BYTE_SECOND			0x02
#define NRP_SEQ_BYTE_THIRD			0x03
#define NRP_PAYLOAD_LENGTH_BYTE		0x04 
#define NRP_PAYLOAD_START_BYTE		0x05

#define NRP_SEQ_BYTE_FIRST_SHIFT		16
#define NRP_SEQ_BYTE_SECOND_SHIFT		8
#define NRP_SEQ_BYTE_THIRD_SHIFT		0


/**
 * The function generates a NRP content packet based on the input parameters. A "ready-to-send" NRP packet is stored in the 
 * rel_packet[] parameter
 *
 * @param rel_packet:		the array where the NRP packet is stored
 * @param flags:			the NRP flags of the packet, see protocol
 * @param seq_ack_num:		the sequence number or the acknowledgement number of the packet
 * @param payload:			the payload of the packet
 * @param payload_length:	the length of the payload
 * @param csp_header:		the "dummy" header of the CSP packet (to calculate checksum)
 *
 * @return:					void
 */
void nrp_make_reliable_packet(uint8_t rel_packet[], uint8_t flags, uint32_t seq_ack_num, const uint8_t payload[], uint8_t payload_length, uint8_t csp_header[CSP_HEADER_LENGTH]);

/**
 * The function generates a packet with empty NRP header fields (except CRC-16, which is used)
 *
 * @param unrel_packet:		the array where the packet is stored
 * @param payload:			the payload of the packet
 * @param payload_length:	the length of the payload
 * @param csp_header:		the "dummy" header of the CSP packet (to calculate checksum)
 *
 * @return:					void
 */
void nrp_make_unreliable_packet(uint8_t unrel_packet[], const uint8_t payload[], uint8_t payload_length, uint8_t csp_header[CSP_HEADER_LENGTH]);


/**
 * The function generates a NRP FIN packet based on input parameters. 
 *
 * @param rel_packet:		the array where the NRP packet is stored
 * @param flags:			the NRP flags of the packet, see protocol
 * @param seq_ack_num:		the sequence number or the acknowledgement number of the packet
 * @param payload:			the payload of the packet
 * @param payload_length:	the length of the payload
 * @param csp_header:		the "dummy" header of the CSP packet (to calculate checksum)
 *
 * @return:					void
 */
void nrp_make_FIN_packet(uint8_t rel_packet[NRP_HEADER_SIZE], uint8_t csp_header[CSP_HEADER_LENGTH]);

/**
 * The function generates a ACK packet based on input parameters. 
 *
 * @param rel_packet:		the array where the NRP packet is stored
 * @param flags:			the NRP flags of the packet, see protocol
 * @param seq_ack_num:		the sequence number or the acknowledgement number of the packet
 * @param payload:			the payload of the packet
 * @param payload_length:	the length of the payload
 * @param csp_header:		the "dummy" header of the CSP packet (to calculate checksum)
 *
 * @return:					void
 */
void nrp_make_ACK_packet(uint8_t rel_packet[NRP_HEADER_SIZE], uint8_t csp_header[CSP_HEADER_LENGTH], uint32_t ack_num, uint8_t flags);

/**
 * The function generates a SYN_NAK packet with an error code specified by the error_byte parameter 
 *
 * @param rel_packet:		the array where the NRP packet is stored
 * @param csp_header:		the "dummy" header of the CSP packet (to calculate checksum)
 * @param ack_num:			the acknowledgement number of the packet
 * @param error_byte:		the error byte which is sent as payload
 *
 * @return:					void
 */
void nrp_make_SYN_NAK_packet(uint8_t rel_packet[NRP_HEADER_SIZE], uint8_t csp_header[CSP_HEADER_LENGTH], uint32_t ack_num, int8_t error_byte);

/**
 * This function generates a "dummy" csp header. The header will contain exactly the same bit pattern as 
 * the header generated by the CSP library when sending the packet. The dummy header is used to generate a 
 * checksum that includes the entire CSP packet, as this is not implemented by CSP itself. (As of CSP version 1.0+)
 * due to compatability issues: https://github.com/GomSpace/libcsp/issues/46
 *
 * @param csp_header:	the array the header will be stored in
 * @param *dest:		pointer to the destination socket struct
 * @param *source:		pointer to the source socket struct	
 * @param priority:		priority of the packet			
 */
void make_csp_dummy_header(uint8_t csp_header[CSP_HEADER_LENGTH], const module_socket_t *dest, const module_socket_t *source, csp_prio_t priority);

#endif /* NCSP_HEADER_H_ */