/*
 * nrp_sm.c
 *
 * Created: 20-Feb-15 11:08:50 AM
 *  Author: Erlend
 *
 *	Received packets at CSP sockets are handled in this file.
 */ 
#include "nrp.h"
#include "nrp_csp_if.h"
#include "nrp_header.h"
#include "nrp_conn.h"
#include "nrp_sm.h"
#include "csplib/include/csp/csp.h"
#include "nrp_ack.h"
#include "nrp_debug.h"
#include "nrp_error.h"
#include "semphr.h"
#include "crc.h"


#define TICKS_10_MIN		(TickType_t) 600000	 //10 mins in ms




//extern xSemaphoreHandle ack_signals[NRP_MAX_ACK_SIGNALS];
void extract_nrp_packet(csp_packet_t *csp_packet, nrp_packet_t *nrp_packet);
void extract_nrp_packet(csp_packet_t *csp_packet, nrp_packet_t *nrp_packet){
	/*Flags*/
	nrp_packet->flags = csp_packet->data[NRP_FLAG_BYTE];
	
	/*Sequence/Acknowledgement number*/
	nrp_packet->seq_ack_num = ((csp_packet->data[NRP_SEQ_BYTE_FIRST] << NRP_SEQ_BYTE_FIRST_SHIFT) & 0x00ff0000)
		| ((csp_packet->data[NRP_SEQ_BYTE_SECOND] << NRP_SEQ_BYTE_SECOND_SHIFT) & 0x0000ff00)
		| (csp_packet->data[NRP_SEQ_BYTE_THIRD] << NRP_SEQ_BYTE_THIRD_SHIFT);
	
	/*Payload length*/
	uint8_t length = csp_packet->data[NRP_PAYLOAD_LENGTH_BYTE];
	nrp_packet->data_length = length;
	
	
	/*copy data to other array*/
	nrp_packet->data = pvPortMalloc(sizeof(uint8_t) * length);
	memcpy(nrp_packet->data, &csp_packet->data[NRP_PAYLOAD_START_BYTE], nrp_packet->data_length);
	
	/*Checksum*/
	nrp_packet->checksum = csp_packet->data[NRP_HEADER_SIZE + nrp_packet->data_length -1] << 8;
	nrp_packet->checksum |= csp_packet->data[NRP_HEADER_SIZE + nrp_packet->data_length];
}

void make_conn_ref(nrp_conn_ref *conn_ref, uint8_t caddr, uint8_t cport, uint8_t saddr, uint8_t sport);
void make_conn_ref(nrp_conn_ref *conn_ref, uint8_t caddr, uint8_t cport, uint8_t saddr, uint8_t sport){
	conn_ref->client_addr = caddr;
	conn_ref->client_port = cport;
	conn_ref->server_addr = saddr;
	conn_ref->server_port = sport;
}

Bool correct_checksum(const csp_packet_t *packet);
Bool correct_checksum(const csp_packet_t *packet){
	
	module_socket_t *src_module = pvPortMalloc(sizeof(*src_module));
	src_module->port = packet->id.sport;
	src_module->csp_address = packet->id.src;
	
	module_socket_t *dst_module = pvPortMalloc(sizeof(*dst_module));
	dst_module->port = packet->id.dport;
	dst_module->csp_address = packet->id.dst;
	
	uint8_t nrp_payload_length = packet->data[NRP_PAYLOAD_LENGTH_BYTE];
	
	uint8_t full_packet_length = CSP_HEADER_LENGTH + NRP_HEADER_SIZE_WITHOUT_CRC + nrp_payload_length; //everything minus the 2 crc bytes
	uint8_t full_packet[full_packet_length];
	uint8_t csp_dummy[CSP_HEADER_LENGTH];
	make_csp_dummy_header(csp_dummy, dst_module, src_module, (csp_prio_t) packet->id.pri);
	
	uint8_t crc_first_byte = NRP_HEADER_SIZE_WITHOUT_CRC + packet->data[NRP_PAYLOAD_LENGTH_BYTE];
	uint8_t crc_byte_one = packet->data[crc_first_byte];
	uint8_t crc_byte_two = packet->data[crc_first_byte + 1];
	
	unsigned short packet_crc = ((crc_byte_one << 8) & 0xff00) | (crc_byte_two & 0xff); 
	
	memcpy(full_packet, csp_dummy, CSP_HEADER_LENGTH * sizeof(uint8_t));
	memcpy(&full_packet[CSP_HEADER_LENGTH], packet->data, (NRP_HEADER_SIZE_WITHOUT_CRC + nrp_payload_length) * sizeof(uint8_t));
	
	unsigned short crc_bytes = crcFast(full_packet, full_packet_length);
	vPortFree(dst_module);
	vPortFree(src_module);
	
	
	if (packet_crc == crc_bytes) return true;
	else return false;
	
}


//Handles all incoming CSP packets
void nrp_receive(void){
	csp_conn_t *csp_conn;
	csp_packet_t *packet;
	uint64_t last_crc_update = xTaskGetTickCount();
	
	csp_socket_t *sock = create_listen_socket(CSP_ANY, CSP_SO_NONE);

	if (sock == NULL) {
		NRP_PRINT_ERROR("Unable to run receive task");
		vTaskDelete(NULL); //Delete myself
		while(1);
	}
	csp_listen(sock, NRP_WINDOW_SIZE * 2);
	
	
	while(true) {
		
		//rebuild the FAST-CRC-16 lookup table every 10 mins in case of radiation bit-flips
		//if (xTaskGetTickCount() >= last_crc_update + TICKS_10_MIN){
			//crcInit();
			//last_crc_update = xTaskGetTickCount();
		//}
		
		
		// Wait for CSP "connection", 100 ms timeout 
		if ((csp_conn = csp_accept(sock, 100)) == NULL){
			continue;
		}
		 
		// Read packets. Timout is 100 ms
		while ((packet = csp_read(csp_conn, 100)) != NULL) {
			
			//NRP_PRINT("Packet received");
			
			//Check the CRC checksum of the packet to assure its not corrupted
			if (!correct_checksum(packet)) {
				//discard packet
				
				NRP_PRINT_WARNING("Corrupted packet detected and discarded");
				csp_buffer_free(packet);
				vTaskDelay(5);
				break;
			}
			
			//get nrp packet
			nrp_packet_t *nrp_packet = pvPortMalloc(sizeof(*nrp_packet));
			extract_nrp_packet(packet, nrp_packet);
			
			//Create / get connection
			module_socket_t *src_module = pvPortMalloc(sizeof(*src_module));
			src_module->port = packet->id.sport;
			src_module->csp_address = packet->id.src;
			
			nrp_conn_ref conn_ref;
			
			//printf("packet received, seq_num: %i\r\n", nrp_packet->seq_ack_num);
			/* Switch state based on flags*/
			switch (nrp_packet->flags){
				case NRP_SYN:
				{
					//SYN received, I am server
					make_conn_ref(&conn_ref, packet->id.src, packet->id.sport, packet->id.dst, packet->id.dport);
					uint32_t seq_num = nrp_packet->seq_ack_num;
					uint32_t next_expected_seq_num = 0;
					
					//check if there is a listener to this port, if not, nak the syn
					if (!is_port_bound(conn_ref.server_port)){
						nrp_acknowledge_packet(packet, next_expected_seq_num, NRP_SYN_NAK, PORT_NOT_BOUND);
						vPortFree(src_module);
						vPortFree(nrp_packet->data);
						vPortFree(nrp_packet);
						csp_buffer_free(packet);
						continue;	
					}
					
					//create connection if it is available
					int8_t status = nrp_connection_create(conn_ref, seq_num, true);
					if (status != SUCCESS){
						NRP_PRINT_WARNING("Unable to create connection");
						switch (status){
							case PORT_ALREADY_CONNNECTED:
								//acknowledge packet
								NRP_PRINT_WARNING("Got connection request from someone, but we're already connected. Sending SYN_ACK");
								next_expected_seq_num = get_conn_next_expected_seq_num(conn_ref) - nrp_packet->data_length;
								
								//Acknowledge packet
								nrp_acknowledge_packet(packet, next_expected_seq_num, NRP_SYN_ACK, 0);
							break;
							
							default: // all errors are sent as syn-nak to client
								
								//nak packet
								NRP_PRINT_ERROR("Sending SYN/NAK");
								nrp_acknowledge_packet(packet, seq_num, NRP_SYN_NAK, status);
								
								vPortFree(src_module);
								vPortFree(nrp_packet->data);
								vPortFree(nrp_packet);
								csp_buffer_free(packet);
								continue;
						}
					}
					else {
						//connection created, make buffer for message
						print_packet(packet->data);
						
						next_expected_seq_num = get_conn_next_expected_seq_num(conn_ref) - nrp_packet->data_length;;
						
						printf("Conn s: %i\r\n", packet->id.src);
						store_data(conn_ref, nrp_packet->data, nrp_packet->data_length, nrp_packet->seq_ack_num);
						update_connection(conn_ref, next_expected_seq_num, get_conn_packets_received(conn_ref) + 1);
						nrp_acknowledge_packet(packet, get_conn_next_expected_seq_num(conn_ref), NRP_SYN_ACK, 0);
					}
				
					break;
				}
				
				case NRP_SYN_ACK:
				{	
					uint32_t seq_num = nrp_packet->seq_ack_num;
					//SYN_ACK received, i am CLIENT
					make_conn_ref(&conn_ref, packet->id.dst, packet->id.dport, packet->id.src, packet->id.sport);
					//do client stuff (slide window)
					if (conn_exists(conn_ref)){
						NRP_PRINT("SYN_ACK received");
						uint16_t rtts = xTaskGetTickCount() - conn_get_syn_stamp(conn_ref);
						conn_set_rtt(conn_ref, rtts);
						//printf("\rrtt: %i\r\n", conn_get_rtt(conn_ref));
						conn_slide_window(conn_ref);
						update_connection(conn_ref, seq_num - nrp_packet->data_length, 1);
					}
					else NRP_PRINT_WARNING("Connection not found");
					break;
					
				}
					
				case NRP_SYN_NAK:
				{
					//connection request was refused, do something
					NRP_PRINT_WARNING("NRP_SYN_NAK");
					#if NRP_DEBUG >= NRP_DEBUG_EW
						char buf[40];
						sprintf(buf, "The stream was cancelled by receiver with NRP error : %i\r\n", (int8_t)nrp_packet->data[0]);
						NRP_PRINT_WARNING(buf);
					#endif
					//i am client
					make_conn_ref(&conn_ref, packet->id.dst, packet->id.dport, packet->id.src, packet->id.sport);
					conn_set_SYN_NAK_recv(conn_ref, true);
					
					//Find a way to shut the stream down through the connection
					break;
				}
				
				case NRP_ARQ:  // Content packet received at server side
				{
					//NO flags received, I am server
					make_conn_ref(&conn_ref, packet->id.src, packet->id.sport, packet->id.dst, packet->id.dport);
					
					//printf("Expected seq: %i, seq of packet: %i\r\n", (int32_t)get_conn_next_expected_seq_num(conn_ref), nrp_packet->seq_ack_num);
					// check that it is from an open connection
					if (conn_exists(conn_ref)){
						print_packet(packet->data);
						uint32_t recv_seq_num = nrp_packet->seq_ack_num;
						
						//Check if it is the next expected packet of the connection
						if (recv_seq_num == get_conn_next_expected_seq_num(conn_ref)){
							
							store_data(conn_ref, nrp_packet->data, nrp_packet->data_length, nrp_packet->seq_ack_num);
							update_connection(conn_ref, get_conn_next_expected_seq_num(conn_ref) - nrp_packet->data_length, get_conn_packets_received(conn_ref) + 1);
							nrp_acknowledge_packet(packet, get_conn_next_expected_seq_num(conn_ref), NRP_ACK, 0);
							
							//couldnt find a better way to reset the nak_sent flag than to do it each time a correct ordered packet arrives
							conn_set_nak_sent(conn_ref, false);
							
						}
						//or if you already have the content of the packet stored
						else if (recv_seq_num > get_conn_next_expected_seq_num(conn_ref)) {
							nrp_acknowledge_packet(packet, get_conn_next_expected_seq_num(conn_ref), NRP_ACK, 0);
						}
						//if none of the above, one of the packets has been lost
						else {
							//Packet has been lost, send NAK with next expected seq number
							if (!conn_get_nak_sent(conn_ref)){
								NRP_PRINT_WARNING("Lost packet, sending NAK");
								nrp_acknowledge_packet(packet, get_conn_next_expected_seq_num(conn_ref), NRP_NAK, 0);
								conn_set_nak_sent(conn_ref, true);
							}					
						}
					}
					else NRP_PRINT_WARNING("Connection not found");
					
					break;
				}
					
				case NRP_ACK:
				{
					//ack received, i am CLIENT
					make_conn_ref(&conn_ref, packet->id.dst, packet->id.dport, packet->id.src, packet->id.sport);
					
					if (conn_exists(conn_ref)){
						// figure out which packet the acknowledgement number belongs to
						uint16_t ack_number = (get_conn_init_seq_num(conn_ref) - nrp_packet->seq_ack_num) / NRP_MAX_PAYLOAD_SIZE;
						
						if (nrp_packet->seq_ack_num == 0) {
							float ack_n = (float)((get_conn_init_seq_num(conn_ref) - nrp_packet->seq_ack_num) / (float)NRP_MAX_PAYLOAD_SIZE);
							float rest = (float)(ack_n - (float)ack_number);
							if (rest > 0) ack_number++;
						}
						
						uint16_t expected_ack = get_conn_packets_received(conn_ref) + 1;
						
						if (expected_ack < ack_number) {
							NRP_PRINT_WARNING("One or more ACK's were lost, but thats fine, i'll just continue");
							//printf("Expected: %i, got: %i, init_seq_num: %" PRIu64 ", packet_ack: %i\r\n", expected_ack, ack_number, get_conn_init_seq_num(conn_ref),nrp_packet->seq_ack_num);
						}
						else if (expected_ack > ack_number){
							uint8_t buf[40];
							sprintf(buf, "Duplicate ACK was discarded, ack:%i", ack_number);
							NRP_PRINT_WARNING(buf);
						}
						//update if it is the newest ack
						if (ack_number >= expected_ack){
							update_connection(conn_ref, nrp_packet->seq_ack_num, ack_number);
							
							//do client stuff (slide window)
							for (int i = 0; i < (ack_number - expected_ack) + 1; i++) conn_slide_window(conn_ref);
						}
					}
					else NRP_PRINT_WARNING("Connection not found");
					break;
				}
				case NRP_FIN:
				{
					// transmit FIN/ACK regardless
					nrp_acknowledge_packet(packet, 0, NRP_FIN_ACK, 0);
					
					// i am server, close connection if it exists
					make_conn_ref(&conn_ref, packet->id.src, packet->id.sport, packet->id.dst, packet->id.dport);
					if (conn_exists(conn_ref)){
						print_packet(packet->data);
						
						//add packet to queue
						nrp_recv_packet_t *recv_packet_ptr = conn_get_recv_packet_ptr(conn_ref);
						add_to_queue(recv_packet_ptr, conn_ref.server_port);
						
						//vTaskDelay(10);
						nrp_connection_close(conn_ref);
						printf("Conn t: %i\r\n", packet->id.src);
					}
					else NRP_PRINT_WARNING("Connection not found");
					break;
				}
				case NRP_FIN_ACK:
				{
					//I am client
					make_conn_ref(&conn_ref, packet->id.dst, packet->id.dport, packet->id.src, packet->id.sport);
					
					if (conn_exists(conn_ref)) {
						//connection is closed by client sender task
						conn_set_fin_ack_received(conn_ref);
					}
					
					break;
				}
				
				case NRP_NAK:
				{
					//I am client
					make_conn_ref(&conn_ref, packet->id.dst, packet->id.dport, packet->id.src, packet->id.sport);
					if (conn_exists(conn_ref)){
						//print_packet(packet->data);
						//start sending from last unacknowledged packet
						conn_set_at_packet(conn_ref, get_conn_packets_received(conn_ref));
						NRP_PRINT_WARNING("NAK received");
						conn_reset_window(conn_ref);
					}
					break;
				}
				
				case NRP_NO_FLAGS:
					NRP_PRINT("Non reliable packet");
					
					//no connection, just pass the data to the port queue with a nrp_recv_packet_t
					nrp_recv_packet_t recv_packet;
					recv_packet.src.csp_address = packet->id.src;
					recv_packet.src.port = packet->id.sport;
					recv_packet.dest.csp_address = packet->id.dst;
					recv_packet.dest.port = packet->id.dport;
					recv_packet.pri = packet->id.pri;
					recv_packet.data_length = nrp_packet->data_length;
					recv_packet.data = pvPortMalloc(sizeof(uint8_t) * nrp_packet->data_length);
					memcpy(recv_packet.data, nrp_packet->data, nrp_packet->data_length);
					
					add_to_queue(&recv_packet, packet->id.dport);
					
					break;
				
				
				default:
				NRP_PRINT_WARNING("Flags in received packet not recognized");
				
				
			}
			
			
			//Free memory
			vPortFree(src_module);
			vPortFree(nrp_packet->data);
			vPortFree(nrp_packet);
			csp_buffer_free(packet);
			
		}
		csp_close(csp_conn);
		vTaskDelay(3); // let another task of lower priority run for a little while
		
	}
	 
}


