/*
 * i2c_csp_glue.h
 *
 *  Author: andrasg
 */ 
#ifndef I2C_CSP_GLUE_H_
#define I2C_CSP_GLUE_H_

#include <FreeRTOS.h>
#include "csplib/include/csp/csp.h"
#include "status_codes.h"

//! Extern counter of successful receives
extern int rx_success;
//! Extern counter of failed receives
extern int rx_fail;
//! Extern counter of successful transmission
extern int tx_success;
//! Extern counter of failed transmission.
extern int tx_fail;

/**
 * \brief Struct we can send on the i2c bus, made so it overlaps and can be cast to and from csp_packet_t
 * \todo this can be reduced in size, both in padding and used fields.
 */
typedef struct __attribute__((__packed__)){
	uint8_t padding[CSP_PADDING_BYTES - 4 * sizeof(uint8_t)];
	uint8_t dest;
	uint8_t len;
	uint8_t len_rx;
	uint8_t retries;
	uint16_t length; // length of data
	csp_id_t id;
	union {
		uint8_t data[0];
		uint16_t data16[0];
		uint32_t data32[0];
	};
}i2c_frame_t; 

//! Callback function to the CSP_rx function.
typedef void (*callback_csp_lib_rx)(i2c_frame_t* frame, void* pxTaskWoken);

/**
 * \brief Init i2c function called from csp library when csp_i2c_init is called.
 * Function that does some inital setup on the device to be enable to use the i2c bus
 * for both incomming and outgoing trafic.
 * \param addr Address of this device on the i2c network
 * \param csp_rx the callback function that places received packets in a routing queue.
 */
int i2c_init(int handle, int master, unsigned int addr, int speed,int unknown1, int unknown2, callback_csp_lib_rx csp_rx);

/**
 * \brief Function that's called from the csp library for every csp_packet to be sent
 * Functuion that's called from the csp library for every csp_packet to be sent. Will
 * free the packet if it's sent correctly.
 * \param frame packet that's going to be sent.
 * \param timeout a timeout for the sending of the entire packet.
 * \todo timeout not yet implemented.
 * \return a statuscode on how the transfer went. (0 if ok)
 */
int i2c_send(int handle, i2c_frame_t * frame, unsigned int timeout);

/**
 * \brief Function that retrieves a buffer from csp and makes recieve point to it.
 * Function that retrieves a buffer from csp and makes recieve point to it. The function was
 * created to make the handling of input and getting of buffer two separate elements. Each time
 * you read a byte from rhr you accept that byte, even if you dont have space to store it. 
 * By splitting this in two, the slave will send a NAK if it doesn't have room.
 * \return return STATUS_OK or ERR_PACKET_ALLOCATION
 */
status_code_t i2c_create_recieve_buffer(void);

/**
 * \brief Function that handles the placement of received bytes.
 * Function that places received bytes in either the csp header, or
 * in the data. The data is placed in an iterative pattern where the
 * overhead_size first elements are csp header and the next ones are data.
 * \param recv_byte The byte thats been received.
 * \return return STATUS_OK or ERR_PACKET_ALLOCATION
 */
status_code_t i2c_receive(uint8_t recv_byte);

/**
 * \brief Function for finishing a the receiveing of a packet.
 * When a packet has finished receiving its entire packet, the packet needs to be shipped
 * to the csp routing task throguh the callback function. This function handles the finished 
 * packet based on the way the slave stopped listening to the master (successfull or not successfull). Not successfull;
 * the packet is freed and recieve pointer is set to NULL. Successfull; length of packet is set,
 * packet is sent to csp routing task, and the recieve pointer is set to NULL, but packet is not
 * freed. 
 * \param stop_type Tells if the transfer was successfull(STATUS_OK) or not(Other)
 * \return pdTrue/pdFalse if a higher priority task has awoken. this comes from csp callback.
 */
portBASE_TYPE i2c_clear_receive(status_code_t stop_type);


/**
 * Function for clearing all the counters of failed
 * and successful packets and I2C transfers
 */
void clear_counters(void);

#endif /* I2C_CSP_GLUE_H_ */