/*
 * i2c_handler.c
 *
 *  Author: andrasg
 */ 
#include "FreeRTOS.h"
#include "task.h"
#include "timers.h"
#include "semphr.h"
#include "queue.h"

#include "conf_csp.h"
#include "i2c_handler.h"
#include "csplib/i2c_csp_glue.h"
#include "i2c_master_interupt.h"
#include "i2c_slave_interupt.h"

#include <avr32/io.h>
//#include "io/uart/uart_write.h"

#include "gpio.h"
#include "pm.h"

#include <time.h>
#include <stdio.h>


/************************************************************************/
/* Prototypes                                                           */
/************************************************************************/

/**
 * \brief Function for default setup of a TWIM module
 * Setup of a TWIM module, registers the IRQ and sets the speed the bus wil operate on
 */
uint8_t i2c_master_setup(void);

/**
 * \brief Function for default setup of a TWIS module
 * Setup of a TWIS master module, registers the IRQ and sets the speed the bus will operate on
 * also registers the address of this module, so the receiver knows what address to listen for.
 */
uint8_t i2c_slave_setup(uint8_t address);

/**
 * \brief Function for disabling Slave address match
 * Function that disables Slave address match and all other slave interrupts.
 * Also sets the slave module to recognize ack by pulling data high (NAK).
 */
__attribute__((__always_inline__))void _disable_receive(void);

/**
 * \brief Function for enabling Slave address match
 * Function that enables Slave address match.
 * Also sets the slave module to recognize ack by pulling data low.
 */
__attribute__((__always_inline__))void _enable_receive(void);

/**
 * \brief Function that enables slave functionality, then releases busaccess semaphore.
 */
__attribute__((__always_inline__))void _give_secure_access_to_bus(void);

/**
 * \brief Function that waits until it can take busaccess sem, then disables the slave functionality.
 */
__attribute__((__always_inline__))void _take_secure_access_to_bus(void);


/************************************************************************/
/* shared variables                                                     */
/************************************************************************/
//! Binary semaphore, makes sure that only trancieve, or recieve can access bus.
volatile xSemaphoreHandle xsem_busaccess; 
//! Value to keep track if slave should give semaphore when stop is called.
volatile portBASE_TYPE semaphore_status;


/************************************************************************/
/* I2C Setup functions                                                  */
/************************************************************************/
int i2c_enable_incomming(uint8_t address){
	if (STATUS_OK != i2c_slave_setup(address)){
		printf("Slave setup Failed\r\n");
		return ERR_IO_ERROR;
	}
	vSemaphoreCreateBinary(xsem_busaccess);
	_enable_receive();
	return STATUS_OK;
}

int i2c_enable_outgoing(void){
	if (STATUS_OK != i2c_master_setup()){
		printf("Master setup Failed\r\n");
		return ERR_IO_ERROR;
	}
	else return STATUS_OK;
}

int i2c_enable_hardware(void){
	//select pins to be used, both as master and slave pins
	const gpio_map_t twim_gpio_map = {
		{I2C_CLK_PIN, I2C_CLK_FUNCTION},
		{I2C_DATA_PIN, I2C_DATA_FUNCTION}
	};
	//Enable the selected pins
	if (GPIO_SUCCESS != gpio_enable_module(twim_gpio_map, sizeof(twim_gpio_map)/sizeof(twim_gpio_map[0]))){
		printf("Enable pins for I2C Failed\r\n");
		return ERR_IO_ERROR;
	}
	
	//enable the i2c_pba clock/ocilator in the powermanager
	if (PASS != pm_enable_module(&AVR32_PM, I2C_PBA_CLK)){
		printf("Enable I2C PBA CLK Failed\r\n");
		return ERR_IO_ERROR;
	}
	return STATUS_OK;
		
}

uint8_t i2c_master_setup(void)
{
	status_code_t status = twim_register_interupts(&I2C_MASTER_MODULE, I2C_MASTER_IRQ);
	if (STATUS_OK != status) return status;
	status = twim_set_speed(&I2C_MASTER_MODULE, I2C_BAUDRATE, I2C_PBA_SPEED);
	return status;
}


uint8_t i2c_slave_setup(uint8_t address){
	
	const twis_options_t twis_opt = {
		.pba_hz = I2C_PBA_SPEED,
		.speed = I2C_BAUDRATE,
		.chip = address,
	};
	
	twis_register_interupt(&I2C_SLAVE_MODULE, I2C_SLAVE_IRQ);
	twis_enable_settings(&I2C_SLAVE_MODULE, &twis_opt);
	
	return STATUS_OK;
}

/************************************************************************/
/* Slave callback functions                                             */
/************************************************************************/
portBASE_TYPE slave_start( portBASE_TYPE * higher_task_woken ){
	char buffer;
	semaphore_status = xQueueReceiveFromISR((xQueueHandle)xsem_busaccess, &buffer, higher_task_woken);	
	return semaphore_status;
}

status_code_t slave_rx( volatile avr32_twis_t * twis_inst)
{
	if (STATUS_OK == i2c_create_recieve_buffer()){
		return i2c_receive(twis_inst->rhr);
	}
	return ERR_PACKET_ALLOCATION;
}

portBASE_TYPE slave_stop( status_code_t stop_type )
{
	portBASE_TYPE higher_priority_task_woken = i2c_clear_receive(stop_type);
	portENTER_CRITICAL();
	_enable_receive();
	portEXIT_CRITICAL();
	if (pdTRUE == semaphore_status){
		semaphore_status = pdFALSE;
		xSemaphoreGiveFromISR(xsem_busaccess, 0);
	}		
	return higher_priority_task_woken;
}


/************************************************************************/
/* State machine enforcement functions                                  */
/************************************************************************/
void _enable_receive(void){
	I2C_SLAVE_MODULE.CR.ack = 0;
	I2C_SLAVE_MODULE.CR.smatch = 1;
	
	I2C_SLAVE_MODULE.idr = ~0UL;
	I2C_SLAVE_MODULE.scr = ~0UL;
	
	I2C_SLAVE_MODULE.ier = AVR32_TWIS_IER_SAM_MASK;
}

void _disable_receive(void){
	I2C_SLAVE_MODULE.CR.smatch = 0;
	I2C_SLAVE_MODULE.CR.ack = 1;
	I2C_SLAVE_MODULE.idr = ~0UL;
	I2C_SLAVE_MODULE.scr = ~0UL;
}

void _take_secure_access_to_bus(void){
	while( pdFALSE == xSemaphoreAltTake(xsem_busaccess, 0) ){};
	portENTER_CRITICAL();
	_disable_receive();
	portEXIT_CRITICAL();
}

void _give_secure_access_to_bus(void){
	portENTER_CRITICAL();
	_enable_receive();
	portEXIT_CRITICAL();
	xSemaphoreAltGive(xsem_busaccess);
}	

/************************************************************************/
/* I2C Send functions                                                   */
/************************************************************************/

status_code_t i2c_send_csp_packet(i2c_frame_t * frame, unsigned int timeout)
{	
	status_code_t status = ERR_TIMEOUT;
	_take_secure_access_to_bus();
	//is not clear from this end to send packet, so go forward
	status = twim_write_csp_packet(&I2C_MASTER_MODULE, frame, timeout);
	_give_secure_access_to_bus();
	return status;
}	



status_code_t i2c_read_from_slave(uint8_t module_addr, uint8_t reg_addr, uint8_t reg_addr_length, uint8_t * rx_buffer, uint8_t rx_buffer_length){
	status_code_t status = ERR_TIMEOUT;
	_take_secure_access_to_bus();
	//is not clear from this end to send packet, so go forward
	status = twim_read_from_dum_unit(&I2C_MASTER_MODULE, rx_buffer, rx_buffer_length, reg_addr, reg_addr_length, module_addr);
	_give_secure_access_to_bus();
	return status;
}
