/*
 * i2c_handler.h
 *
 *  Author: andrasg
 */ 


#ifndef I2C_HANDLER_H_
#define I2C_HANDLER_H_
#include "csplib/i2c_csp_glue.h"
#include "status_codes.h"

/************************************************************************/
/* Variables                                                            */
/************************************************************************/
//! semaphore status is made external so that the slave isr can access it.
extern volatile portBASE_TYPE semaphore_status;

/************************************************************************/
/* Setup functions                                                      */
/************************************************************************/
/**
 * \brief Function for enabling of the physical hardware for the i2c bus
 * Enables the pins, and the clock for the i2c bus.
 */
int i2c_enable_hardware(void);

/**
 * \brief Function that makes device able to receive data over i2c. 
 * Remember to enable hardware as well.
 * \param address the address of this device
 */
int i2c_enable_incomming(uint8_t address);

/**
 * \brief Function that makes device able to transmit data over i2c. 
 * Remember to enable hardware as well.
 */
int i2c_enable_outgoing(void);


/************************************************************************/
/* Send functions                                                       */
/************************************************************************/
/**
 * \brief Function for sending a i2c_frame on the bus.
 * Function that sends a i2c_frame on the bus. Tries to grab bus-semaphore
 * and immediately after getting semaphore, enter critical phase for disabling 
 * Slave address match. There might be a area here where an interrupt can sneak in, 
 * and we can end up with slave ack by pulling data high (a nak). so the enable sam
 * will make sure we reset this after sending packet.
 * \param frame the frame to send.
 * \todo not yet implemented timeout properly.
 * \param timeout timeout for the transaction. 
 * \return The status of the transfer.
 */
status_code_t i2c_send_csp_packet(i2c_frame_t * frame, unsigned int timeout);
/**
 * \brief Function for reading from a slave sensor over the bus.
 * Tries to grab bus-semaphore and immediately after getting semaphore, enter critical phase for disabling 
 * Slave address match. There might be a area here where an interrupt can sneak in, 
 * and we can end up with slave ack by pulling data high (a nak). so the enable sam
 * will make sure we reset this after sending packet.
 * \param frame the frame to send.
 * \return The status of the transfer.
 * \todo might need to implent a timeout parameter here as well
 */
status_code_t i2c_read_from_slave(uint8_t module_addr, uint8_t reg_addr, uint8_t reg_addr_length, uint8_t * rx_buffer, uint8_t rx_buffer_length);


/************************************************************************/
/* Slave callback functions                                             */
/************************************************************************/
/**
 * \brief Function callback from TWIS ISR for handling of start of receive
 * Function that grabs the bus semaphore to make sure that we are allowed to 
 * receive on the bus. Called when a slave receives a Slave Address Match.
 * \return pdTRUE or pdFALSE if we got the semaphore.
 */
portBASE_TYPE slave_start( portBASE_TYPE * higher_task_woken );

/**
 * \brief Function callback from TWIS ISR for handling receiving a byte.
 * Function that makes sure that we have room storage room for the byte, and 
 * calls the glue function to store the byte on its correct CSP possition.
 * \param twis_inst pointer to the TWIS module.
 */
status_code_t slave_rx( volatile avr32_twis_t * twis_inst);

/**
 * \brief Function callback from TWIS ISR for handling the end of a I2C transfer
 * Function that's called when a slave has to stop a transfer. 
 * Calls the glue function for handling handling CSP stop correctly based on the 
 * stop_type. Then enables slave address match and releases bus semaphore.
 * \param stop_type Value that says if we are stopping because of success or failure.
 * \return Whether or not a task has been awoken. 
 */
portBASE_TYPE slave_stop( status_code_t stop_type );

#endif /* I2C_HANDLER_H_ */