/*
 * i2c_master_interupt.c
 *
 *  Author: andrasg
 */ 
#include <stdio.h>
#include <time.h>

#include "gpio.h"
//#include "io/uart/uart_write.h"

#include "csplib/include/csp/csp.h"

#include "i2c_master_interupt.h"
#include "conf_csp.h"

#include "task.h"
#include "status_codes.h"
#include "timers.h"
#include "i2c_slave_interupt.h"

/************************************************************************/
/* Variables used in csp                                                */
/************************************************************************/

//! Flag that sais if the current send is csp or not.
static volatile bool is_csp_packet = true;
//! Counter to make the csp header bytes possition correctly
static volatile uint8_t csp_header_poss = 1;
//! Flag that sais if we have reached a timeout or not. Used with timeout implementation.
static uint8_t timeout_reached = 0;
//! Handle for the timer used to wait for timeouts.
static volatile xTimerHandle timer_byte_transfer;
//! A variable containing the size of the csp transfer overhead.
static volatile uint8_t overhead_size = 4;
//! a pointer to the i2c_frame that is being transfered
static volatile i2c_frame_t * current_frame;
//! \internal Pointer to the instance 1 of the TWIM registers for IT.
static volatile avr32_twim_t *twim_inst;
//! \internal Status of the bus transfer
static volatile twim_transfer_status_t transfer_status;
//! \internal Pointer to the applicative TWI transmit buffer.
static const volatile uint8_t  *twim_tx_buffer = NULL;
//! \internal Remaining number of bytes to transmit.
static volatile uint32_t twim_tx_nbytes = 0;
//! \internal Pointer to the applicative TWI receive buffer.
static volatile uint8_t *twim_rx_buffer;
//! \internal Remaining number of bytes to receive.
static volatile uint8_t twim_rx_nbytes = 0;
//! \internal IT mask.
static volatile uint32_t twim_it_mask;


/************************************************************************/
/* Prototypes                                                           */
/************************************************************************/
//! function that listens to the bus to detect if it is idle or not.
status_code_t twi_bus_is_idle(void);


/************************************************************************/
/* ISR Functions                                                        */
/************************************************************************/

/**
 * Call function form the Switching ISR, which performs the actual ISR content.
 * is marked noinline as a requirement from the switching isr, and also returns 
 * portBASE_TYPE because of this. Each time the function interacts with a freeRTOS
 * queue, it will get a value that sais if a higher priority task has awoken or not.
 * This is returned to be used in switching ISR
 * \return pdFALSE or pdTRUE, but is automatically used by switching isr.
 */
//Alternative header if we want to remove switching __attribute__((__interrupt__)) static void twim_isr_content(void)
__attribute__((__noinline__)) static portBASE_TYPE twim_isr_content(void)
{
	// get masked status register value
	uint32_t status = twim_inst->sr & twim_it_mask;
	portBASE_TYPE higher_priority_task_woken = pdFALSE;
	// this is a NACK
	if (status & (AVR32_TWIM_SR_ANAK_MASK
		| AVR32_TWIM_SR_DNAK_MASK
		| AVR32_TWIM_SR_ARBLST_MASK )) {
		//if we get a nak, clear the valid bit in cmdr,
		//otherwise the command will be resent.		
		transfer_status = (status &  AVR32_TWIM_SR_ARBLST_MASK) ? TWI_ARBITRATION_LOST : TWI_RECEIVE_NACK;

		twim_inst->CMDR.valid = 0;
		twim_inst->scr = ~0UL;
		twim_inst->idr = ~0UL;
	} //This is RX
	else if (status & AVR32_TWIM_SR_RXRDY_MASK) {
		// get data from Receive Holding Register
		*twim_rx_buffer = twim_inst->rhr;
		twim_rx_buffer++;
		
		// decrease recieved bytes number
		twim_rx_nbytes --;
		// receive complete
		if (twim_rx_nbytes == 0) {
			// finish the receive operation
			twim_inst->idr = AVR32_TWIM_IDR_RXRDY_MASK;
		}
	}
	// this is a TXRDY
	else if (status & AVR32_TWIM_SR_TXRDY_MASK) {
		//TODO remove the variable, and replace with direct inputs.
		uint8_t transfer_byte = 0;
		//Transfered all the bytes
		if (twim_tx_nbytes == 0) {
			// finish the receive operation
			twim_inst->idr = AVR32_TWIM_IDR_TXRDY_MASK;
		//Is a csp packet and on header bytes
		} else if (is_csp_packet && csp_header_poss <= overhead_size){
			
			transfer_byte = (uint8_t)(current_frame->id.ext >> ((overhead_size - csp_header_poss) * 8));
			twim_inst->thr = transfer_byte;
			//Iterate and Decrement counters
			csp_header_poss ++;
			twim_tx_nbytes--;
			
		//Is normal data transfer
		}else {
			// put the byte in the Transmit Holding Register
			transfer_byte = *twim_tx_buffer++;		
			twim_inst->thr = transfer_byte;
			twim_tx_nbytes--;

		}
	}
	//remove return if we want to remove switching
	return higher_priority_task_woken;
}

/**
 * \internal
 * \brief TWI interrupt handler.
 * The ISR used in a switching context. The naked param is to make sure a stackframe is created correctly
 * while the port enter switching isr function needs a function between them that returns a portBaseType.
 * The port exit switching isr uses the return value of the middle function to decide if it should yield 
 * to a task or not.
 */
__attribute__((__naked__)) static void twim_isr(void){
	portENTER_SWITCHING_ISR();
	twim_isr_content();
	portEXIT_SWITCHING_ISR();
}

/************************************************************************/
/* Functions for initial setup for I2C master                           */
/************************************************************************/

/* *
 * \brief Set the twim bus speed in cojunction with the clock frequency
 *
 * \param twim              Base address of the TWIM (i.e. &AVR32_TWIM).
 * \param speed             The desired twim bus speed
 * \param pba_hz            The current running PBA clock frequency
 * \return STATUS_OK        Transaction is successful
 * \return ERR_INVALID_ARG  Invalid arg resulting in wrong CWGR Exponential
 */
status_code_t twim_set_speed (volatile avr32_twim_t *twim, uint32_t speed,
		uint32_t pba_hz)
{
	uint32_t f_prescaled;
	uint8_t cwgr_exp = 0;
	f_prescaled = (pba_hz / speed / 2);
	// f_prescaled must fit in 8 bits, cwgr_exp must fit in 3 bits
	while ((f_prescaled > 0xFF) && (cwgr_exp <= 0x7)) {
		// increase clock divider
		cwgr_exp++;
		// divide f_prescaled value
		f_prescaled /= 2;
	}
	if (cwgr_exp > 0x7) {
		return ERR_INVALID_ARG;
	}
	// set clock waveform generator register
	twim->cwgr = ((f_prescaled/2) << AVR32_TWIM_CWGR_LOW_OFFSET)
			| ((f_prescaled - f_prescaled/2) << AVR32_TWIM_CWGR_HIGH_OFFSET)
			| (cwgr_exp << AVR32_TWIM_CWGR_EXP_OFFSET)
			| (0     << AVR32_TWIM_CWGR_DATA_OFFSET)
			| (f_prescaled << AVR32_TWIM_CWGR_STASTO_OFFSET);
	return STATUS_OK;
}

status_code_t twim_register_interupts(volatile avr32_twim_t *twim, unsigned int irq){
	
		portENTER_CRITICAL();
		twim->idr = ~0UL;
		// Enable master transfer
		twim->cr = AVR32_TWIM_CR_MEN_MASK;
		// Reset TWI
		twim->cr = AVR32_TWIM_CR_SWRST_MASK;
		portEXIT_CRITICAL();

		// Clear SR
		twim->scr = ~0UL;

		portENTER_CRITICAL();
		INTC_register_interrupt(twim_isr_content, irq, AVR32_INTC_INTLEVEL_INT1);
		//To implement isr without switching, use the one below, and remove/add some code in twis_alt_isr
		//INTC_register_interrupt(twim_isr, irq, AVR32_INTC_INTLEVEL_INT1);
		portEXIT_CRITICAL();
		
		twim->cr = AVR32_TWIM_CR_MDIS_MASK;
		return STATUS_OK;
}

/************************************************************************/
/* Functions for initiationg transfer on the I2C bus                    */
/************************************************************************/
status_code_t twim_read_from_dum_unit(volatile avr32_twim_t *twim, uint8_t * rx_buffer, uint8_t rx_buffer_length, uint8_t reg_addr, uint8_t reg_addr_length, uint8_t to_module_addr){
	twim->cr = AVR32_TWIM_CR_MDIS_MASK;
	
	twim_inst = twim;
	portENTER_CRITICAL();
	is_csp_packet = false;
	transfer_status = TWI_SUCCESS;
	
	twim_rx_buffer = rx_buffer;
	twim_rx_nbytes = rx_buffer_length;
	
	twim_it_mask =	AVR32_TWIM_IER_DNAK_MASK 
			| AVR32_TWIM_IER_ANAK_MASK 
			| AVR32_TWIM_IER_ARBLST_MASK;
	
	if (reg_addr_length){
		
		twim_tx_buffer = (uint8_t*) &reg_addr;
		twim_tx_nbytes = reg_addr_length;
		
		twim_it_mask |= (AVR32_TWIM_IER_TXRDY_MASK | AVR32_TWIM_IER_RXRDY_MASK);
				
		twim_inst->cmdr = (to_module_addr << AVR32_TWIM_CMDR_SADR_OFFSET)
				| (twim_tx_nbytes << AVR32_TWIM_CMDR_NBYTES_OFFSET)
				| (AVR32_TWIM_CMDR_VALID_MASK)
				| (AVR32_TWIM_CMDR_START_MASK)
				| (0 << AVR32_TWIM_CMDR_READ_OFFSET);
				
		twim_inst->ncmdr = (to_module_addr << AVR32_TWIM_CMDR_SADR_OFFSET)
				| (twim_rx_nbytes << AVR32_TWIM_CMDR_NBYTES_OFFSET)
				| (AVR32_TWIM_CMDR_VALID_MASK)
				| (AVR32_TWIM_CMDR_START_MASK)
				| (AVR32_TWIM_CMDR_STOP_MASK)
				| (AVR32_TWIM_CMDR_READ_MASK);
	}else{
		twim_tx_nbytes = 0;
		twim_it_mask |= AVR32_TWIM_IER_RXRDY_MASK;
		// Set the command register to initiate the transfer
		twim_inst->cmdr = (to_module_addr << AVR32_TWIM_CMDR_SADR_OFFSET)
				| (twim_rx_nbytes << AVR32_TWIM_CMDR_NBYTES_OFFSET)
				| (AVR32_TWIM_CMDR_VALID_MASK)
				| (AVR32_TWIM_CMDR_START_MASK)
				| (AVR32_TWIM_CMDR_STOP_MASK)
				| (AVR32_TWIM_CMDR_READ_MASK);
	}
	
	twim_inst->ier = twim_it_mask;
	//wait until bus is free
	if (STATUS_OK != twi_bus_is_idle()){
		portEXIT_CRITICAL();
		return ERR_BUSY;
	}
	
	twim_inst->cr = AVR32_TWIM_CR_MEN_MASK | AVR32_TWIM_CR_SMDIS_MASK;
	portEXIT_CRITICAL();
	
	while (!(transfer_status) && !(twim_status ())) {
		cpu_relax();
	}
	
	// Disable master transfer
	twim->cr = AVR32_TWIM_CR_MDIS_MASK;
	if (transfer_status == TWI_RECEIVE_NACK
			|| transfer_status == TWI_ARBITRATION_LOST) {
		return ERR_IO_ERROR;
	}
	return STATUS_OK;
	
	
}

status_code_t twim_write_csp_packet (volatile avr32_twim_t *twim, i2c_frame_t * frame, unsigned int timeout)
{
	// Reset the TWIM module to clear the THR register
	twim->cr = AVR32_TWIM_CR_MEN_MASK;
	twim->cr = AVR32_TWIM_CR_SWRST_MASK;
	twim->cr = AVR32_TWIM_CR_MDIS_MASK;
	// Set pointer to TWIM instance for IT
	twim_inst = twim;
	// Disable the TWIM interrupts
	portENTER_CRITICAL();
	// Clear the interrupt flags
	twim->idr = ~0UL;
	// Clear the status flags
	twim->scr = ~0UL;
	// get a pointer to applicative data
	current_frame = frame;
	twim_tx_buffer = frame->data;
	
	// set the number of bytes to transmit
	twim_tx_nbytes = frame->len + frame->length;//data + csp_id_t
	
	is_csp_packet = true;
	csp_header_poss = 1;
	
	// Initialize bus transfer status
	transfer_status = TWI_SUCCESS;
	
	// set the command to start the transfer
	twim_inst->cmdr = (frame->id.dst << AVR32_TWIM_CMDR_SADR_OFFSET)
	| (twim_tx_nbytes << AVR32_TWIM_CMDR_NBYTES_OFFSET)
	| (AVR32_TWIM_CMDR_VALID_MASK)
	| (AVR32_TWIM_CMDR_START_MASK)
	| (AVR32_TWIM_CMDR_STOP_MASK)
	| (0 << AVR32_TWIM_CMDR_TENBIT_OFFSET)
	| (0 << AVR32_TWIM_CMDR_READ_OFFSET);
	
	// Set a mask for the interrupts to be enables in transfer
	twim_it_mask =	AVR32_TWIM_IER_DNAK_MASK 
	| AVR32_TWIM_IER_ANAK_MASK 
	| AVR32_TWIM_IER_ARBLST_MASK
	| AVR32_TWIM_IER_TXRDY_MASK ;
	// update IMR through IER
	twim_inst->ier = twim_it_mask;
	
	if (STATUS_OK != twi_bus_is_idle()){
		portEXIT_CRITICAL();
		return ERR_BUSY;
	}
	
	// Enable master transfer
	twim_inst->cr = AVR32_TWIM_CR_MEN_MASK | AVR32_TWIM_CR_SMDIS_MASK;
	// Enable all interrupts
	portEXIT_CRITICAL();
	
	
	int timeout_id = 50;
	timeout_reached = 0;
	
	//TODO to enable timeout remove these comments.
	//timer_byte_transfer = xTimerCreate("Send_Byte_Timeout", timeout + CSP_MINIMUM_TRANSFER_TIMEOUT, pdFALSE, (void *)timeout_id, &send_byte_timeout_callback);
	//xTimerStart(timer_byte_transfer, 0);
	
	// send data	
	while (!(transfer_status) && !(twim_status()) && !timeout_reached) {
		cpu_relax();
	}
	
	//xTimerStop(timer_byte_transfer, 0);
	//xTimerDelete(timer_byte_transfer, 0);
	
	#if AVR32_TWIM_H_VERSION > 101	// Removed in twim100 module due to IC bug
	// Disable master transfer
	twim->cr = AVR32_TWIM_CR_MDIS_MASK;
	#endif
	
	if (transfer_status == TWI_RECEIVE_NACK
	|| transfer_status == TWI_ARBITRATION_LOST) {
		return ERR_IO_ERROR;
	}else if (transfer_status == TWI_SUCCESS_BUT_BUGGY_IRQ){
		return ERR_PROTOCOL;
	} else if (timeout_reached){			
		twim->cr = AVR32_TWIM_CR_MEN_MASK;
		twim->cr = AVR32_TWIM_CR_SWRST_MASK;
		twim->cr = AVR32_TWIM_CR_MDIS_MASK;

		return ERR_TIMEOUT;	
	}
	return STATUS_OK;
}

/************************************************************************/
/* Internal functions                                                   */
/************************************************************************/

/**
 * \brief Function that listens to the TWI pins for an idle bus, and returns if the bus idle or not.
 * \return ERR_BUSY if busy, STATUS_OK if idle
 */
status_code_t twi_bus_is_idle(void){
	int idle = 0;
	int busy = 0;
	int idle_limit = I2C_IDLE_LIMIT;
	int busy_limit = I2C_BUSY_LIMIT;
	
	while (idle <= idle_limit){//make sure bus is free
		if (gpio_pin_is_high(I2C_DATA_PIN) && gpio_pin_is_high(I2C_CLK_PIN)){
			idle++;
		}else {
			idle = 0;
			if (++ busy > busy_limit){
				//reset
				twim_inst->CMDR.valid = 0;
				twim_inst->cr = AVR32_TWIM_CR_MEN_MASK;
				twim_inst->cr = AVR32_TWIM_CR_SWRST_MASK;
				twim_inst->cr = AVR32_TWIM_CR_MDIS_MASK;
				return ERR_BUSY;
			}
		}			
	}
	return STATUS_OK;
}

/**
 * \brief Information about the current status of the TWI Bus
 */
uint8_t twim_status ( void )
{
	uint32_t status = twim_inst->sr;
	if ((status & AVR32_TWIM_SR_IDLE_MASK)
#if AVR32_TWIM_H_VERSION > 101 ||(status&AVR32_TWIM_SR_BUSFREE_MASK)
#endif
		) {
		return 1;
	} else {
		return 0;
	}
}
/**
 * \brief callback function for the timer function for the idle loop to detect timeout.
 */
void send_byte_timeout_callback(xTimerHandle * pxTimer){
	timeout_reached = 1;
}