/*
 * i2c_master_interupt.h
 *
 *  Author: andrasg
 */ 


#ifndef I2C_MASTER_INTERUPT_H_
#define I2C_MASTER_INTERUPT_H_


#include <avr32/io.h>
#include <stdint.h>
#include "compiler.h"
#include "status_codes.h"
#include "config/conf_twim.h"
#include "csplib/i2c_csp_glue.h"

/************************************************************************/
/* Structs for functionality                                            */
/************************************************************************/
//! Status Register Mask for No Acknowledgements
#define AVR32_TWIM_SR_NAK_MASK  (AVR32_TWIM_SR_ANAK_MASK  \
| AVR32_TWIM_SR_DNAK_MASK)

#define AVR32_TWIM_SR_STD_MASK  (AVR32_TWIM_SR_NAK_MASK  \
| AVR32_TWIM_SR_ARBLST_MASK)

/**
 * \brief Status Codes for TWI Transfer
 */
enum twim_transfer_status{
	TWI_SUCCESS = 0,            //!< \brief TWI Transaction Success
	TWI_INVALID_ARGUMENT = -1,  //!< \brief Invalid Argument Passed
	TWI_ARBITRATION_LOST = -2,  //!< \brief Bus Arbitration Lost
	TWI_NO_CHIP_FOUND = -3,     //!< \brief Slave Not Found
	TWI_RECEIVE_NACK = -4,      //!< \brief Data No Acknowledgement Received
	TWI_SEND_NACK = -5,         //!< \brief Data No Acknowledgement Send
	TWI_INVALID_CLOCK_DIV = -6,  //!< \brief Invalid Clock Divider Value
	TWI_SUCCESS_BUT_BUGGY_IRQ = -7,
	TWI_SERVICE_BUSY = -8
};
typedef enum twim_transfer_status twim_transfer_status_t;

/**
 * \brief Input parameters when initializing the twim module mode
 */
typedef struct
{
	//! The PBA clock frequency.
	uint32_t pba_hz;
	//! The baudrate of the TWI bus.
	uint32_t speed;
	//! The desired address.
	uint32_t chip;
}
twim_options_t;


/************************************************************************/
/* Functions for initial setup of I2C master                            */
/************************************************************************/

/**
 * \brief Register the interrupts for master
 *
 * \param twim              Base address of the TWIM (i.e. &AVR32_TWIM)
 * \param *opt              IRQ of the interrupt handler to register 
 * \return STATUS_OK        Transaction is successful
 */
status_code_t twim_register_interupts(volatile avr32_twim_t *twim, unsigned int irq);

/**
 * \brief Set the twim bus speed in cojunction with the clock frequency
 *
 * \param twim              Base address of the TWIM (i.e. &AVR32_TWIM).
 * \param speed             The desired twim bus speed
 * \param pba_hz            The current running PBA clock frequency
 * \return STATUS_OK        Transaction is successful
 * \return ERR_INVALID_ARG  Invalid arg resulting in wrong CWGR Exponential
 */
extern status_code_t twim_set_speed (volatile avr32_twim_t *twim,
		uint32_t speed, uint32_t pba_hz);

/************************************************************************/
/* Functions for initiating transfers on the I2C-bus                    */
/************************************************************************/

/**
 * \brief Read registers from a dum slave unit
 *
 * \param twim            Base address of the TWIM (i.e. &AVR32_TWIM)
 * \param rx_buffer		  Buffer pointing to a place to store the recieved bytes
 * \param rx_buffer_length Number of bytes to rx
 * \param reg_addr		  Addres of the register you want to read
 * \param reg_addr_length Number of bytes in the register addr
 * \param to_module_add   The i2c address of the module to read from
 * \return STATUS_OK      If all bytes were send successfully
 * \return ERR_IO_ERROR   NACK received or Bus Arbitration lost
 */
status_code_t twim_read_from_dum_unit(volatile avr32_twim_t *twim, uint8_t * rx_buffer, uint8_t rx_buffer_length, uint8_t reg_addr, uint8_t reg_addr_length, uint8_t to_module_addr);

/**
 * \brief Write multiple bytes to a TWI compatible slave device
 *
 * \param twim            Base address of the TWIM (i.e. &AVR32_TWIM)
 * \param *frame		  Frame to be sent.
 *                        (see \ref i2c_frame_t)
 * \param timeout			Not used atm.
 * \return STATUS_OK      If all bytes were send successfully
 * \return ERR_IO_ERROR   NACK received or Bus Arbitration lost
 */
status_code_t twim_write_csp_packet (volatile avr32_twim_t *twim, i2c_frame_t * frame, unsigned int timeout);

/*!
 * \brief Information about the current status of the TWI Bus
 */
extern uint8_t twim_status ( void );

#endif /* I2C_MASTER_INTERUPT_H_ */