/*
 * i2c_slave_interupt.c
 *
 *  Author: andrasg
 */ 

#include <avr32/io.h>
#include <math.h>

#include "FreeRTOS.h"
#include "task.h"

#include "compiler.h"
#include "status_codes.h"

#include "i2c_slave_interupt.h"
#include "i2c/i2c_handler.h"
//#include "io/uart/uart_write.h"

#include "conf_csp.h"

/************************************************************************/
/* Defines to make isr more organized                                   */
/************************************************************************/
//!	Define the mask of all errors the slave will detect in the status register
#define I2C_SLAVE_ERROR_SR_MASK (AVR32_TWIS_SR_BUSERR_MASK     \
| AVR32_TWIS_SR_SMBPECERR_MASK                  \
| AVR32_TWIS_SR_SMBTOUT_MASK                    \
| AVR32_TWIS_SR_ORUN_MASK                       \
| AVR32_TWIS_SR_URUN_MASK)
//! Define the mask of all errors a slave will enable with interrupt enable register
#define I2C_SLAVE_ERROR_IER_MASK (AVR32_TWIS_IER_BUSERR_MASK     \
| AVR32_TWIS_IER_SMBPECERR_MASK                  \
| AVR32_TWIS_IER_SMBTOUT_MASK                    \
| AVR32_TWIS_IER_ORUN_MASK                       \
| AVR32_TWIS_IER_URUN_MASK)
//! Define the mask of the conditions for a finished slave that must be able to enables in IER
#define I2C_SLAVE_FINISH_IER_MASK (AVR32_TWIS_IER_TCOMP_MASK | AVR32_TWIS_IER_REP_MASK)
//! Define the mask of the conditions for a finished slave that has happened in SR
#define I2C_SLAVE_FINISH_SR_MASK (AVR32_TWIS_SR_TCOMP_MASK | AVR32_TWIS_SR_REP_MASK)
//! Define the mask of the conditions for a finished slave that has to be cleared in SCR
#define I2C_SLAVE_FINISH_SCR_MASK (AVR32_TWIS_SCR_TCOMP_MASK | AVR32_TWIS_SCR_REP_MASK)
//! Define the mask for the conditions on which a receive must react with IER
#define I2C_SLAVE_RECIEVE_IER_MASK (AVR32_TWIS_IER_RXRDY_MASK | AVR32_TWIS_IER_TCOMP_MASK | AVR32_TWIS_IER_REP_MASK)

/************************************************************************/
/* Variables                                                            */
/************************************************************************/
//! Pointer to the TWIS instance physical address could be exchanged with the define
volatile avr32_twis_t *twis_inst;
//! Value to indicate how the receive is going, which will be sendt to stop function
static volatile status_code_t return_status = 0;

/************************************************************************/
/* ISR Functions                                                        */
/************************************************************************/
/**
 * \internal
 * \brief Functionallity of the TWIS ISR
 * Defined as noinline, since switching isr is based of this. Returns a pdTRUE/pdFALSE
 * depending on if a higher priority task has awoken. 
 * Uses functions start, stop and rx from the i2c handler.
 */
//Alternative header if we want to remove switching  __attribute__((__interrupt__)) twis_alt_isr(void){
__attribute__((__noinline__)) static portBASE_TYPE twis_alt_isr(void){
//__attribute__((__interrupt__)) twis_alt_isr(void){
	// if no sem taken, try take sem. slave start
	unsigned long status = twis_inst->sr;
	unsigned long pending = status & twis_inst->imr;
	portBASE_TYPE higher_task = pdFALSE;
	if (!semaphore_status){
		if (!slave_start(&higher_task)){
			return_status = ERR_SEMAPHORE_FAIL;
			twis_inst->CR.ack = 1;
		}else{
			return_status = STATUS_OK;
		}
	}
	
	// if error dissable interrupts and run slave stop.
	if (status & I2C_SLAVE_ERROR_SR_MASK){
		higher_task = slave_stop(ERR_IO_ERROR);
	}
	
	// if SAM, enable recieve and errors. clear comp and rep
	if (pending & AVR32_TWIS_IER_SAM_MASK){
		twis_inst->scr = I2C_SLAVE_FINISH_SCR_MASK;
		twis_inst->ier = I2C_SLAVE_ERROR_IER_MASK | I2C_SLAVE_RECIEVE_IER_MASK;
	}
	
	// if read, read. slave rx.
	if (pending & AVR32_TWIS_IER_RXRDY_MASK){
		if (STATUS_OK != slave_rx(&I2C_SLAVE_MODULE)){
			return_status = ERR_PACKET_ALLOCATION;
			twis_inst->CR.ack = 1;
			twis_inst->rhr;
		}
	}
	
	// finish, dissable interrupts clear com rep nak slave stop
	if (pending & I2C_SLAVE_FINISH_IER_MASK){
		higher_task = slave_stop(return_status);
	}
	twis_inst->scr = pending;
	
	//remove return if we want to remove switching
	return higher_task;
}

/**
 * \brief TWIS ISR, implemented with switching isr.
 * The interrupt routine. Defined with __naked__ for saving of context. Switching ISR in case higher
 * priority task is called. The yeild will only be called from portEXIT_SWITCHING_ISR();.
 * Defined also as critical, since we dont want to deal with nesting interrupts. 
 */
__attribute__((__naked__)) static void twis_isr(void){
	portENTER_SWITCHING_ISR();
		twis_alt_isr();
	portEXIT_SWITCHING_ISR();
}

/************************************************************************/
/* Functions for initial setup for I2C slave                            */
/************************************************************************/
void twis_register_interupt(volatile avr32_twis_t *twis, unsigned int irq){

	portENTER_CRITICAL();
	twis->idr = ~0UL;
	twis->scr = ~0UL;
	
	// Reset the TWIS module
	twis->cr = AVR32_TWIS_CR_SWRST_MASK;
	portEXIT_CRITICAL();

	// Clear the TWIS status register
	twis->scr = ~0UL;;

	// Disable all interrupts
	portENTER_CRITICAL();
	//Register ISR
	INTC_register_interrupt(twis_isr, irq, AVR32_INTC_INTLEVEL_INT1);
	//To implement isr without switching, use the one below, and remove/add some code in twis_alt_isr
	//INTC_register_interrupt(twis_alt_isr, irq, AVR32_INTC_INTLEVEL_INT1);
	
	twis_inst = twis;
	//Enable all interrupts
	portEXIT_CRITICAL();
}

void twis_enable_settings(volatile avr32_twis_t *twis, const twis_options_t *opt){
		// Enable the TWI Slave Module and allow for clock stretching
	twis->cr = AVR32_TWIS_CR_SEN_MASK | AVR32_TWIS_CR_SMATCH_MASK | AVR32_TWIS_CR_STREN_MASK;

	/*
	 * Bug on TWIS up to version 120 included which incorrectly handles
	 * clock stretch on address match.
	 */
#if AVR32_TWIS_H_VERSION > 120
	twis->cr |= AVR32_TWIS_CR_SOAM_MASK;
#endif

	// Set slave address
	twis->cr |= (opt->chip << AVR32_TWIS_CR_ADR_OFFSET);

	// Calculate the wait time from clk falling edge to
	//let the slave control the bus
	uint8_t setup_time = (opt->pba_hz / opt->speed)/7;
	twis->tr = (setup_time << AVR32_TWIS_TR_SUDAT_OFFSET);

	// Slave Address Match Interrupt Enable
	twis->ier = AVR32_TWIS_IER_SAM_MASK;
}
