/*
 * main.c
 *
 * Created: 25.04.2015 16:33:26
 *  Author: mariuwes
 */ 

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <avr32/io.h>
#include "data.h"

/* Environment header files. */
#include "compiler.h"
#include "board.h"
#include "power_clocks_lib.h"
#include "print_funcs.h"
#include "flashc.h"
#include "gpio.h"

/* FreeRTOS header files. */
#include "FreeRTOS.h"
#include "task.h"
#include "queue.h"
#include "semphr.h"

/* ADCS header files. */
#include "logging.h"
#include "sun_sensor.h"
#include "magnetometer.h"
#include "gyroscope.h"
#include "predictor.h"
#include "estimator.h"
#include "detumbling_controller.h"
#include "tumbling_controller.h"
#include "pointing_controller.h"
#include "coil_driver.h"
#include "sensor_processing.h"

/* Flashc UserPage setup. */
#if defined (__GNUC__)
__attribute__((__section__(".userpage")))
#elif defined(__ICCAVR32__)
__no_init
#endif
;

static PersistentVariables persistent;

/* Data Structures. */
Time current_time;
Data current_data;
int mode = MODE_IDLE;


/* Synchronization variables, only need to use one. */
xSemaphoreHandle xSemaphore;
xSemaphoreHandle xMutex;

int initialize(void) {
	
	// Clock
	pcl_switch_to_osc(PCL_OSC0, FOSC0, OSC0_STARTUP);
	
	// USART, for debug purposes
	init_dbg_rs232(FOSC0);
	
	// LED, for debug purposes
	gpio_configure_pin(AVR32_PIN_PB03, GPIO_DIR_OUTPUT); 
		
	return 42;
}

static portTASK_FUNCTION(blinkerTask, pvParameters) {
	portTickType xLastWakeTime;
	xLastWakeTime = xTaskGetTickCount();
	
	while(1) {
		gpio_toggle_pin(AVR32_PIN_PB03);
		print_dbg("LED \r \n");
		vTaskDelayUntil(&xLastWakeTime, (1000/portTICK_RATE_MS));
	}
}

static portTASK_FUNCTION(Sensor_Task, pvParameters) {
	portTickType xLastWakeTime;
	xLastWakeTime = xTaskGetTickCount();
	
	while(1) {
		if(xSemaphoreTake(xSemaphore, 10) == pdTRUE) {
			print_dbg("Sensor_Task \r \n");
			if (mode != MODE_IDLE) {
				gyroscope_poll(&current_data.gyroscope_data[0]);
				magnetometer_poll(&current_data.magnetometer_data[0]);
				sun_sensor_poll(&current_data.sun_sensor_data[0]);
				median_voter(&current_data.gyroscope_data[0]);
				median_voter(&current_data.magnetometer_data[0]);
				// FUNCTION TO MAKE THE SUN VECTOR FROM SUN SENSOR DATA IS NEEDED
			}
		}
		
		xSemaphoreGive(xSemaphore);
		vTaskDelayUntil(&xLastWakeTime, (1000/portTICK_RATE_MS));
	}
}

static portTASK_FUNCTION(Predictor_Task, pvParameters) {
	portTickType xLastWakeTime;
	xLastWakeTime = xTaskGetTickCount();
	
	while(1) {
		if(xSemaphoreTake(xSemaphore, 10) == pdTRUE) {
			print_dbg("Predictor_Task \r \n");
			if(mode == MODE_ESTIMATION || mode == MODE_POINTING || mode == MODE_TUMBLING) {
				predictor_get_data(current_data.time, &current_data.predictor_data);
			}
		}
		
		xSemaphoreGive(xSemaphore);
		vTaskDelayUntil(&xLastWakeTime, (1000/portTICK_RATE_MS));	// Predictor can wait longer than the rest of the system
	}
}

static portTASK_FUNCTION(Estimator_Task, pvParameters) {
	portTickType xLastWakeTime;
	xLastWakeTime = xTaskGetTickCount();
	
	while(1) {
		if(xSemaphoreTake(xSemaphore, 10) == pdTRUE) {
			print_dbg("Estimator_Task \r \n");
			if(mode == MODE_ESTIMATION || mode == MODE_POINTING || mode == MODE_TUMBLING) {
				// UPPDATED ESTIMATOR NEEDED
			}
		}
	
		xSemaphoreGive(xSemaphore);
		vTaskDelayUntil(&xLastWakeTime, (1000/portTICK_RATE_MS));
	}	
}

static portTASK_FUNCTION(Controller_Task, pvParameters) {
	portTickType xLastWakeTime;
	xLastWakeTime = xTaskGetTickCount();
	
	while(1) {
		if(xSemaphoreTake(xSemaphore, 10) == pdTRUE) {
			print_dbg("Controller_Task \r \n");
			if(mode == MODE_DETUMBLING) {
				detumbling_controller_get_output(&current_data.magnetometer_data[0], &current_data.output_vector[0]);
				// SET COIL OUTPUT FUNCTION NEEDED
			}
			else if(mode == MODE_TUMBLING) {
				tumbling_controller_get_output(&current_data.magnetometer_data[0], &current_data.estimator_data[0], current_data.output_vector[0]);
				// SET COIL OUTPUT FUNCTION NEEDED
			}
			else if(mode == MODE_POINTING) {
				pointing_controller_get_output(&current_data.magnetometer_data[0], &current_data.estimator_data[0], current_data.output_vector[0]);
				// SET COIL OUTPUT FUNCTION NEEDED
			}
		}
		
		xSemaphoreGive(xSemaphore);
		vTaskDelayUntil(&xLastWakeTime, (1000/portTICK_RATE_MS));
	}
}

static portTASK_FUNCTION(ADCS_State_Machine_Task, pvParameters) {
	// Task timing variables
	portTickType xLastWakeTime;
	xLastWakeTime = xTaskGetTickCount();
		
	while(1) {	
		
		print_dbg("ADCS_State_Machine_Task \r \n");
		
		switch(mode) {
			case MODE_IDLE:
				// Do nothing
			break;
			case MODE_DETUMBLING:
				gyroscope_poll(&current_data.gyroscope_data[0]);
				magnetometer_poll(&current_data.magnetometer_data[0]);
				sun_sensor_poll(&current_data.sun_sensor_data[0]);
				median_voter(&current_data.gyroscope_data[0]);
				median_voter(&current_data.magnetometer_data[0]);
				// FUNCTION TO MAKE THE SUN VECTOR FROM SUN SENSOR DATA IS NEEDED
			
				detumbling_controller_get_output(&current_data.magnetometer_data[0], &current_data.output_vector[0]);
				// SET COIL OUTPUT FUNCTION NEEDED
			break;
			case MODE_TUMBLING:
				gyroscope_poll(&current_data.gyroscope_data[0]);
				magnetometer_poll(&current_data.magnetometer_data[0]);
				sun_sensor_poll(&current_data.sun_sensor_data[0]);
				median_voter(&current_data.gyroscope_data[0]);
				median_voter(&current_data.magnetometer_data[0]);
				// FUNCTION TO MAKE THE SUN VECTOR FROM SUN SENSOR DATA IS NEEDED
			
				predictor_get_data(current_data.time, &current_data.predictor_data);
				// UPPDATED ESTIMATOR NEEDED
			
				tumbling_controller_get_output(&current_data.magnetometer_data[0], &current_data.estimator_data[0], current_data.output_vector[0]);
				// SET COIL OUTPUT FUNCTION NEEDED
			break;
			case MODE_ESTIMATION:
				gyroscope_poll(&current_data.gyroscope_data[0]);
				magnetometer_poll(&current_data.magnetometer_data[0]);
				sun_sensor_poll(&current_data.sun_sensor_data[0]);
				median_voter(&current_data.gyroscope_data[0]);
				median_voter(&current_data.magnetometer_data[0]);
				// FUNCTION TO MAKE THE SUN VECTOR FROM SUN SENSOR DATA IS NEEDED
			
				predictor_get_data(current_data.time, &current_data.predictor_data);
				// UPPDATED ESTIMATOR NEEDED
			break;
			case MODE_POINTING:
				gyroscope_poll(&current_data.gyroscope_data[0]);
				magnetometer_poll(&current_data.magnetometer_data[0]);
				sun_sensor_poll(&current_data.sun_sensor_data[0]);
				median_voter(&current_data.gyroscope_data[0]);
				median_voter(&current_data.magnetometer_data[0]);
				// FUNCTION TO MAKE THE SUN VECTOR FROM SUN SENSOR DATA IS NEEDED
			
				predictor_get_data(current_data.time, &current_data.predictor_data);
				// UPPDATED ESTIMATOR NEEDED
				pointing_controller_get_output(&current_data.magnetometer_data[0], &current_data.estimator_data[0], current_data.output_vector[0]);
				// SET COIL OUTPUT FUNCTION NEEDED
			break;
		}

		vTaskDelayUntil(&xLastWakeTime, (1000/portTICK_RATE_MS));
	}
}

static void flash_write_mode(PersistentVariables *persistent_variables, int *new_mode) {
	flashc_memcpy((void *)&persistent_variables->mode, new_mode, sizeof(persistent_variables->mode),   true);
}

static void flash_write_tle(PersistentVariables *persistent_variables, char *TLE[136]) {
	flashc_memcpy((void *)&persistent_variables->TLE, TLE, sizeof(persistent_variables->TLE),   true);
}

static void flash_init_read(PersistentVariables *persistent_variables) {
	mode = persistent_variables->mode;
	predictor_update_tle(persistent_variables->TLE);
	// R needs to be set when the estimator is updated
	// Q needs to be set when the estimator is updated
	
	// Debug code for flash
	//print_dbg("Mode: \r \n");
	//print_dbg_char_hex(persistent_variables->mode);
	//print_dbg("\r\n");
	
	//print_dbg("TLE: \r \n");
	//for(int i = 0; i < 136; i++) {
		//print_dbg_char(persistent_variables->TLE[i]);
	//}
	//print_dbg("\r\n");
}

int main(void) {
	
	
	
	// Initalize board
	initialize();		
	
	// Initialize time, code needs to be made for the interfacing of the RTC
	current_time.year = 2015;
	current_time.mon = 3;
	current_time.day = 31;
	current_time.hr = 13;
	current_time.min = 0;
	current_time.sec = 0;
	
	// Initialize data set
	current_data.mode = mode;
	current_data.time = current_time;
	memset(current_data.gyroscope_data, 0, sizeof(current_data.gyroscope_data));
	memset(current_data.magnetometer_data, 0, sizeof(current_data.magnetometer_data));
	memset(current_data.sun_sensor_data, 0, sizeof(current_data.sun_sensor_data));
	memset(current_data.estimator_data, 0.F, sizeof(current_data.estimator_data));
	memset(current_data.predictor_data, 0, sizeof(current_data.predictor_data));
	memset(current_data.output_vector, 0, sizeof(current_data.output_vector));
	
	// Initialize synchronization variables
	vSemaphoreCreateBinary(xSemaphore);	
	
	// Create Tasks
	//xTaskCreate(blinkerTask, (signed char *) "LEDx", configMINIMAL_STACK_SIZE + 1024, NULL, tskIDLE_PRIORITY+2, ( xTaskHandle * ) NULL );
	//xTaskCreate(ADCS_State_Machine_Task, (signed char *) "ADCS_State_Machine_Task", configMINIMAL_STACK_SIZE + 4096, NULL, tskIDLE_PRIORITY+1, ( xTaskHandle * ) NULL );
	//xTaskCreate(Controller_Task, (signed char *) "Controller_Task", configMINIMAL_STACK_SIZE + 2048, NULL, tskIDLE_PRIORITY+1, ( xTaskHandle * ) NULL );
	//xTaskCreate(Estimator_Task, (signed char *) "Estimator_Task", configMINIMAL_STACK_SIZE + 2048, NULL, tskIDLE_PRIORITY+2, ( xTaskHandle * ) NULL );
	//xTaskCreate(Predictor_Task, (signed char *) "Predictor_Task", configMINIMAL_STACK_SIZE + 2048, NULL, tskIDLE_PRIORITY+3, ( xTaskHandle * ) NULL );
	//xTaskCreate(Sensor_Task, (signed char *) "Sensor_Task", configMINIMAL_STACK_SIZE + 2048, NULL, tskIDLE_PRIORITY+4, ( xTaskHandle * ) NULL );

	//vTaskStartScheduler();

	while (true);
}
