%**************************************************************************
% File main.m
% Main script. Contains the following sections:
%
% - Simulate (integrate) system
%
% Written by Zdenko Tudor, 2011. Edited by Marius Westgaard, 2014
%**************************************************************************

clear all; close all; clc;
addpath('myIGRF');

% Initialize storage variables 
global VAR;
global simulation_time;
global control_output_found;
global control_output;
global B_dot_estimate_prev;
global B_B_prev;


% Create parameters struct
P = parameters();

% Load neccessary parameters
orbitPeriod = P.orbitPeriod;        % For defining simulation length
scalePlot = P.scalePlot;            % For x-axes scaling when plotting
w_o = P.w_o;
ix_max = P.ix_max;
iy_max = P.iy_max;
iz_max = P.iz_max;
Nx = P.Nx;
Ny = P.Ny;
Nz = P.Nz;
Ax = P.Ax;
Ay = P.Ay;
Az = P.Az;

m_B_x_max = Nx*Ax*ix_max;
m_B_y_max = Ny*Ay*iy_max;
m_B_z_max = Nz*Az*iz_max;

% Simulation parameters
numOfOrbits = 16;                    % Simulation time in number of orbits
tSpan = [0,orbitPeriod*numOfOrbits];% Time span for ODE45
simulation_time = 0;
control_output_found = 0; 
control_output = [0; 0; 0];
B_dot_estimate_prev = 0;
B_B_prev = 0;

% Initial parameters
w_B_OB = [0.002; -0.03; 0.013];  % Initial rotational velocity
eulAng = [130;50;-90];                     % Initial orientation    
initLat = 0;                               % Initial latitude
initJoule = 0;                             % Always to be set to zero

% Transform initial parameters     
qua = euler2q(eulAng(1)*pi/180, eulAng(2)*pi/180, eulAng(3)*pi/180);   
S_eps = [ 0, -qua(4), qua(3);
    qua(4), 0, -qua(2);
    -qua(3), qua(2), 0];
R_O_B = eye(3) + 2*qua(1)*S_eps + 2*S_eps^2;
R_B_O = R_O_B';
w_B_IB = w_B_OB + R_B_O*[0; -w_o; 0];

% State vector x = [wx wy wz eta eps1 eps2 eps3 latitude totalJoule];
xInit = [w_B_IB;qua;initLat;initJoule];

% ODE45 simulaton settings
options = odeset('MaxStep',0.1,'OutputFcn',@outFcn,'Refine',1);
%options = odeset('MaxStep',0.1,'OutputFcn',@outFcn,'RelTol',1e-5,'AbsTol',1e-5,'Refine',1);

% Simulate system
[tout,yout] = ode45(@(t,x)nonlinearSatellite(t,x,P),tSpan,xInit,options);

%% Angular velocity w^B_OB
w_B_OB = zeros(length(tout),3);
for i=1:size(tout)
    S_eps = [ 0, -yout(i,7), yout(i,6);
        yout(i,7), 0, -yout(i,5);
        -yout(i,6), yout(i,5), 0];
    R_O_B = eye(3) + 2*yout(i,4)*S_eps + 2*S_eps^2;
    R_B_O = R_O_B';
    
    w_B_OB(i,1:3) = yout(i,1:3)'-R_B_O*[0;-P.w_o;0];
end

figure(1)
subplot(3,1,1); hold on; grid on;
title('Angular Velocities, \omega^B_{OB}')
ylabel('x_{comp} [rad/sec]')
xlabel('Time [orbits]')
plot(tout*scalePlot,w_B_OB(:,1)) %,'k','LineWidth',2)

subplot(3,1,2); hold on; grid on;
ylabel('y_{comp} [rad/sec]')
xlabel('Time [orbits]')
plot(tout*scalePlot,w_B_OB(:,2)) %,'k','LineWidth',2)

subplot(3,1,3); hold on; grid on;
ylabel('z_{comp} [rad/sec]')
xlabel('Time [orbits]')
plot(tout*scalePlot,w_B_OB(:,3)) %,'k','LineWidth',2)

%% Euler Angles
for i=1:size(tout)
    [phi(i),theta(i),psi(i)] = q2euler(yout(i,4:7)/norm(yout(i,4:7)));
    phi(i) = phi(i)*180/pi;
    theta(i) = theta(i)*180/pi;
    psi(i) = psi(i)*180/pi;
end

figure(2)
subplot(3,1,1); hold on; grid on;
title('Euler Angeles')
ylabel('Roll (x) [degrees]');
xlabel('Time [orbits]')
plot(tout*scalePlot,phi) %,'k','LineWidth',2)

subplot(3,1,2); hold on; grid on;
ylabel('Pitch (y) [degrees]');
xlabel('Time [orbits]')
plot(tout*scalePlot,theta) %,'k','LineWidth',2)

subplot(3,1,3); hold on; grid on;
ylabel('Yaw (z) [degrees]');
xlabel('Time [orbits]')
plot(tout*scalePlot,psi) %,'k','LineWidth',2)

%% Quaternions
figure(3)
subplot(4,1,1); hold on; grid on;
title('Quaternions')
ylabel('\eta');
xlabel('Time [orbits]')
plot(tout*scalePlot,yout(:,4)) %,'k','LineWidth',2)
axis([0 numOfOrbits -1.1 1.1]);

subplot(4,1,2); hold on; grid on;
ylabel('\epsilon_1');
xlabel('Time [orbits]')
plot(tout*scalePlot,yout(:,5)) %,'k','LineWidth',2)

subplot(4,1,3); hold on; grid on;
ylabel('\epsilon_2');
xlabel('Time [orbits]')
plot(tout*scalePlot,yout(:,6)) %,'k','LineWidth',2)

subplot(4,1,4); hold on; grid on;
ylabel('\epsilon_3');
xlabel('Time [orbits]')
plot(tout*scalePlot,yout(:,7)) %,'k','LineWidth',2)
 
%% Moments generated by the coils
figure(4)
subplot(3,1,1); hold on; grid on;
title('Magnetic Moment, m^b')
ylabel('x^+-face [Am^2] ')
xlabel('Time [orbits]')
plot(VAR.t*scalePlot,VAR.moment(:,1)) %,'k','LineWidth',2)

subplot(3,1,2); hold on; grid on;
ylabel('y^+-face [Am^2] ')
xlabel('Time [orbits]')
plot(VAR.t*scalePlot,VAR.moment(:,2)) %,'k','LineWidth',2)

subplot(3,1,3); hold on; grid on;
ylabel('z^+-face [Am^2] ')
xlabel('Time [orbits]')
plot(VAR.t*scalePlot,VAR.moment(:,3)) %,'k','LineWidth',2)

%% Torques generated by the coils
figure(5)
subplot(3,1,1); hold on; grid on;
title('Torque, \tau^b')
ylabel('x-axis [Nm]');
xlabel('Time [orbits]')
plot(VAR.t*scalePlot,VAR.torque(:,1)) %,'k','LineWidth',2)

subplot(3,1,2); hold on; grid on;
ylabel('y-axis [Nm]');
xlabel('Time [orbits]')
plot(VAR.t*scalePlot,VAR.torque(:,2)) %,'k','LineWidth',2)

subplot(3,1,3); hold on; grid on;
ylabel('z-axis [Nm]');
xlabel('Time [orbits]')
plot(VAR.t*scalePlot,VAR.torque(:,3)) %,'k','LineWidth',2)