%**************************************************************************
% File IGRF.m
%
% Algorithm for finding the magnetic field vectors from the IGRF-11
% model for a given latitude LAT and longitude LON in radians, with
% order n and degree m. The output is the magnetic field vector
% represented in the orbit frame.
%
% Created by Raymond Kristiansen, 2000
% Edited by Zdenko Tudor, 2011
%**************************************************************************
function B = IGRF(n, m, lat, lon, P)
% Load IGRF11 coefficient variables
load('IGRF11_data')

i = n;
j = m;

% Fix latitude in case of possible singularity
mt = mod(lat,2*pi);
tol = 1e-9;
if (mt>(pi/2)-tol && mt<(pi/2)+tol)
    lat=(pi/2)+tol;
elseif(mt>(3*pi/2)-tol && mt<(3*pi/2)+tol)
    lat=(3*pi/2)+tol;
end

% Defining constants
Re = P.Re;
Rc = P.Rc;

% Calculating colatitude and longitude in radians
theta = (pi/2)-lat;
phi = lon;

% Zero offset
O = 1;

% Defining temporary variables
Bt2 = 0; Bp2 = 0; Br2 = 0;

% Calculating Legendre polynominals
[P, dP, S]=Pfunk(n, m, theta);

% Calculating field vectors
for n=1:i
    Bt1 = 0;
    Bp1 = 0;
    Br1 = 0;
    for m=0:j
        Bt1 = Bt1 + (S(O+n,O+m)*g_data(O+n,O+m)*cos(m*phi)+S(O+n,O+m)*...
            h_data(O+n, O+m)*sin(m*phi))*dP(O+n,O+m);
        Bp1 = Bp1 + (m*S(O+n,O+m)*h_data(O+n, O+m)*cos(m*phi)-m*...
            S(O+n,O+m)*g_data(O+n, O+m)*sin(m*phi))*P(O+n,O+m);
        Br1 = Br1 + (S(O+n,O+m)*g_data(O+n,O+m)*cos(m*phi)+S(O+n,O+m)*...
            h_data(O+n, O+m)*sin(m*phi))*P(O+n,O+m);
    end
    Bt2 = Bt2 + ((Re/Rc)^(n+2))*Bt1;
    Bp2 = Bp2 + ((Re/Rc)^(n+2))*Bp1;
    Br2 = Br2 + (n+1)*((Re/Rc)^(n+2))*Br1;
end

% In Cartesian coordinates
eps = 0;
X = Bt2*cos(eps)-Br2*sin(eps);
Y = -Bp2/(sin(theta));
Z = -Bt2*sin(eps)-Br2*cos(eps);

B = [X;Y;Z]*1e-9;
end

%**************************************************************************
%
% Algorithm for calculating the associated Legendre polynominals
% for the given order n, degree m and co-latitude theta. The output
% is the Legendre polynominal P, its partial derivative dP and
% the Schmidt normalization matrix S.
%
% Created by Raymond Kristiansen, 2000
% Edited by Zdenko Tudor, 2011
%**************************************************************************
function [P, dP, S] = Pfunk(n, m, theta)
O = 1;  % Zero index offset

% Defining zero matrices
P = ones(n+O, m+O);
dP = zeros(n+O, m+O);
S = ones(n+O, m+O);
i = n;
j = m;

% Calculating S matrix
for n = 0:1:i
    for m = 0:1:j
        if n > 0
            if m == 0
                S(O+n, O+m) = S(n-1+O, 0+O)*(2*n-1)/n;
            else
                if m == 1
                    deltaFun = 1;
                else
                    deltaFun = 0;
                end
                S(n+O, m+O) = S(n+O, m-1+O)*sqrt(((n-m+1)*(deltaFun+1))/(n+m));
            end
        end
    end
end

% Calculates the Legendre polynominal P and its partial derivative dP
for n = 0:1:i
    for m = 0:1:j
        if n==1
            dP(n+O, m+O) = cos(theta)*dP(n-1+O, m+O)-sin(theta)*P(n-1+O,m+O);
            P(n+O, m+O) = cos(theta)*P(n-1+O, m+O);   
        elseif n>1
            K=((n-1)^2-m^2)/((2*n-1)*(2*n-3));
            dP(n+O, m+O) = cos(theta)*dP(n-1+O, m+O)-sin(theta)*P(n-1+O,m+O)-...
                K*dP(n-2+O, m+O);
            P(n+O, m+O) = cos(theta)*P(n-1+O, m+O) - K*P(n-2+O, m+O);
        end
        if n==m && n>0
            P(n+O, m+O) = sin(theta)*P(n-1+O, n-1+O);
            dP(n+O, m+O) = sin(theta)*dP(n-1+O, n-1+O)+cos(theta)*P(n-1+O, n-1+O);
        end
    end
end

end
