%**************************************************************************
% File nonlinearSatellite.m
%
% xDot = nonlinearSatellite(T,X,P)
% 
% Simulates the earth-satellite system. To be used with an ordinary
% differential equation solver such as MATLAB's Runge Kutta(4,5) method,
% ODE45. 
% The input:
% T - time vector (implicit)
% X - state vector, initial state has to be provided. The state vector is
% of the form X = [W^B_IB' Q' LAMBDA JOULE]. W^B_IB is a 3x1 vector
% containing the three rotational velocity components. Q is the quaternion
% vector of the form Q = [eta eps1 eps2 eps3]. LAMBDA is the latitude of
% the satellite and JOULE is only a testing paramemter which will return
% the total Joule consumption of the satellite during the simulation.
% Initial JOULE should always be set to zero.
% P - struct containg necessary parameters (w_o, I, Rc, coil data, voltage
% and current ratings).
%
% Example:
% [TOUT,YOUT] = ode45(@(t,x)nonlinearSatellite(t,x,P),tSpan,xInit,options)
% this produces:
% TOUT - time vector for the integration
% YOUT - Integrated xDot vector for each timestep size(YOUT)=length(TOUT)x9
%
% Written by Zdenko Tudor, 2011. Edited by Marius Westgaard, 2014
%**************************************************************************

function xDot = nonlinearSatellite(t,x,P)
global R_B_O;
global tmpVAR;
global simulation_time;
global control_output_found;
global control_output;
global B_dot_estimate_prev;
global B_B_prev;

% Load necessary parameters from P struct.
w_o = P.w_o;
I = P.I;
J = P.J;
rho_a = P.rho_a;
A_drag = P.A_drag;
c_p_x = P.c_p_x;
vel = P.vel;

m_B_x_max = P.ix_max*P.Ax*P.Nx;
m_B_y_max = P.iy_max*P.Ay*P.Ny;
m_B_z_max = P.iz_max*P.Az*P.Nz;

% Normalization of quartenions
x(4:7) = x(4:7)./norm(x(4:7));

% State vector x = [];
w_B_IB = x(1:3);              % Angular velocity vector
%w_B_IB_noise = w_B_IB + 0.0001*randn(size(w_B_IB));       
w_B_IB_noise = w_B_IB + 0.00002*sin(0.5*t)*[1; 1; 1];   
eta = x(4);                 % Euler parameter eta
eps = x(5:7);               % Euler parameter epsilon
lat = x(8);                 % Lambda
q = x(4:7);

% Epsilon crossmatrix, skew-symmetric
S_eps = [ 0, -eps(3), eps(2);   
          eps(3), 0, -eps(1);
         -eps(2), eps(1), 0];

% Rotation matrices
R_O_B = eye(3) + 2*eta*S_eps + 2*S_eps^2;
R_B_O = R_O_B';

% Angular velocities of frames relative each other
w_O_IO = [0;-w_o;0];
w_B_OB = w_B_IB-R_B_O*w_O_IO;

% Geomagnetic field in Orbit frame and in Body frame
[B_O] = IGRF(10,10,lat,0,P);         % IGRF model
B_B = R_B_O*B_O;

% Choose controller
tic
[tau_m, m_B, W] = pointing_controller(P, w_B_IB_noise, B_B, q);
%[tau_m, m_B, W, B_dot_estimate] = detumbling_controller(P, B_B, B_B_prev, B_dot_estimate_prev);
% if (t>P.orbitPeriod)
%     if (B_B_prev < 5)
%         w_B_IB = w_B_IB + 0.002*randn(size(w_B_IB));
%     end
%     [tau_m, m_B, W] = tumbling_controller(P, w_B_IB, B_B, q);
% else
%     tau_m = [0; 0; 0];
%     m_B = [0; 0; 0];
%     W = 0;
% end
toc

% Set estimation period
estimation_time = 0.5;

% Actuators are being used but with a constant gain
if (t - simulation_time > estimation_time && t - simulation_time < 2*estimation_time && control_output_found == 0)
    control_output = m_B;
    control_output_found = 1;
    
    % Used for tumbling testing
%     if (t>P.orbitPeriod) 
%         B_B_prev = B_B_prev + 1;
%     end
    
    % Used for detumbling testing
    %B_B_prev = B_B;
    %B_dot_estimate_prev = B_dot_estimate;
    %m_B = [0; 0; 0];
end

if (t - simulation_time > estimation_time && t - simulation_time < ((abs(control_output(1))/m_B_x_max)*estimation_time + estimation_time) && control_output_found == 1)
    m_B(1) = m_B_x_max*sign(control_output(1));
else
    m_B(1) = 0;     
end

if (t - simulation_time > estimation_time && t - simulation_time < ((abs(control_output(2))/m_B_y_max)*estimation_time + estimation_time) && control_output_found == 1)
    m_B(2) = m_B_y_max*sign(control_output(2));
else
    m_B(2) = 0; 
end

if (t - simulation_time > estimation_time && t - simulation_time < ((abs(control_output(3))/m_B_z_max)*estimation_time + estimation_time) && control_output_found == 1)
    m_B(3) = m_B_z_max*sign(control_output(3));
else
   m_B(3) = 0; 
end

% Start timer from zero and repeat the process
if(t - simulation_time >= 2*estimation_time && control_output_found == 1)
    simulation_time = t;
    control_output_found = 0;
    m_B = [0; 0; 0];
end

tau_m = cross(m_B,B_B);

%-----Storing variables used for testing purposes only-----Start-----
tmpVAR.torque = tau_m;
tmpVAR.moment=m_B;
tmpVAR.W = W;
xDot(9,1) = W;
%-----Storing variables used for testing purposes only------End------

% Gravitational torque
c3 = R_B_O(:,3);
tau_g = 3*w_o^2*cross(c3,I*c3);

% Aerodynamic drag torque
Vr = R_B_O*[-1 0 0]';
Vr_x = [    0  -Vr(3)   Vr(2)
      Vr(3)     0   -Vr(1)
     -Vr(2)   Vr(1)     0 ];
tau_a = rho_a*vel*(vel*A_drag*c_p_x*Vr - (I + Vr_x*J)*w_B_OB);

% Total disturbance torques
tau_dist = tau_g; % + tau_a;

% Angular acceleration
wDot_B_IB=I\(tau_m+tau_dist-cross(w_B_IB,I*w_B_IB));

% Eta
etaDot = -0.5*eps'*w_B_OB;

% Epsilon
epsDot = 0.5*(eta*eye(3)+S_eps)*w_B_OB;

% Variable assignment
xDot(1:3,1) = wDot_B_IB;
xDot(4,1) = etaDot;     %qDot(1);
xDot(5:7,1) = epsDot;   %q(2:4);
xDot(8,1) = w_o;

end


