%**************************************************************************
% File parameters.m
%
% P = parameters()
%
% Creates a parameter struct P, which is used to pass the parameters
% between different functions.
%
% Written by Zdenko Tudor, 2011. Edited by Marius Westgaard, 2014
%**************************************************************************

function P = parameters()
% Coil data
% Battery voltage
V = 3.3;

% Number of coil windings (turns)
Nx = 155;
Ny = 155;
Nz = 238;

% [m^2] Coil area
Ax = 0.08*0.18;
Ay = 0.08*0.18;
Az = 0.08^2;

% [Ohm] Coil resistance
Rx = 27.2;
Ry = 27.2;
Rz = 18.5;

% [A] Maximum current through coils
ix_max = V/Rx;
iy_max = V/Ry;
iz_max = V/Rz;

% Satellite dimensions
x = 0.1;                    % X-axis length
y = 0.1;                    % Y-axis length
z = 0.2;                    % Z-axis length

% Other constants
m = 2.6;                    % [kg] Satellite mass

Re = 6371.2e3;              % [m] Earth radius
Rs = 600e3;                 % [m] Satellite altitude
Rc = Re+Rs;                 % [m] Distance from earth center to satellite

G = 6.67428e-11;            % Earth gravitational constant
M = 5.972e24;               % Earth mass

% Calculated constants
Ix = (m*0.9/12)*(y^2+z^2);  % X-axis inertia
Iy = (m/12)*(x^2+z^2);      % Y-axis inertia
Iz = (m/12)*(x^2+y^2);      % Z-axis inertia
I = diag([Ix Iy Iz]);       % Inertia matrix

w_o = sqrt(G*M/Rc^3);       % Satellite angular velocity relative Earth

orbitPeriod = (2*pi)/(w_o); % For defining simulation length
scalePlot = 1/orbitPeriod;  % For x-axes scaling when plotting

% Aerodynamic drag constants
jx = 0.03;                  % X-axis length for centre of pressure
jy = 0.03;                  % Y-axis length for centre of pressure
jz = 0.083;                 % Z-axis length for centre of pressure
Jx = (m/12)*(jy^2+jz^2);    % X-axis inertia for centre of pressure
Jy = (m/12)*(jx^2+jz^2);    % Y-axis inertia for centre of pressure
Jz = (m/12)*(jx^2+jy^2);    % Z-axis inertia for centre of pressure
J = diag([Jx Jy Jz]);       % Inertia matrix for centre of pressure

A_drag = z*sqrt(x^2 + y^2); % Maximum area for calc disturbance drags

vel = 2*pi*Rc/orbitPeriod;  % Velocity

rho_a = 4.89e-13;           % Atmospheric density

c_p = [0.02 0.02 0.02]';    % Relative location of centre of pressure
c_p_x = [    0  -c_p(3)   c_p(2)
      c_p(3)     0   -c_p(1)
     -c_p(2)   c_p(1)     0 ];

% Store variables
P.w_o = w_o;
P.V = V; P.I = I; P.J = J;
P.Nx = Nx; P.Ny = Ny; P.Nz = Nz;
P.Ax = Ax; P.Ay = Ay; P.Az = Az;
P.Rx = Rx; P.Ry = Ry; P.Rz = Rz;
P.ix_max = ix_max; P.iy_max = iy_max; P.iz_max = iz_max;
P.Re = Re; P.Rc = Rc;
P.orbitPeriod = orbitPeriod; P.scalePlot = scalePlot;
P.A_drag = A_drag; P.vel = vel;
P.rho_a = rho_a; P.c_p_x = c_p_x;
P.Ix = Ix; P.Iy = Iy; P.Iz = Iz;
end